// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    kw,
    module::transform_body::{
        expand_match::{parse_enum_path, EnumPath},
        ExprOptions, Visitor,
    },
    HdlAttr,
};
use quote::{format_ident, quote_spanned};
use std::mem;
use syn::{
    parse_quote_spanned, punctuated::Punctuated, spanned::Spanned, token::Paren, Expr, ExprArray,
    ExprCall, ExprGroup, ExprMethodCall, ExprParen, ExprPath, ExprRepeat, ExprStruct, ExprTuple,
    FieldValue, Token, TypePath,
};

impl Visitor<'_> {
    pub(crate) fn process_hdl_array(
        &mut self,
        hdl_attr: HdlAttr<ExprOptions, kw::hdl>,
        mut expr_array: ExprArray,
    ) -> Expr {
        let ExprOptions { sim } = hdl_attr.body;
        let span = hdl_attr.kw.span;
        if sim.is_some() {
            for elem in &mut expr_array.elems {
                *elem = parse_quote_spanned! {elem.span()=>
                    ::fayalite::sim::value::ToSimValue::to_sim_value(&(#elem))
                };
            }
            parse_quote_spanned! {span=>
                ::fayalite::sim::value::ToSimValue::into_sim_value(#expr_array)
            }
        } else {
            for elem in &mut expr_array.elems {
                *elem = parse_quote_spanned! {elem.span()=>
                    ::fayalite::expr::ToExpr::to_expr(&(#elem))
                };
            }
            parse_quote_spanned! {span=>
                ::fayalite::expr::ToExpr::to_expr(&#expr_array)
            }
        }
    }
    pub(crate) fn process_hdl_repeat(
        &mut self,
        hdl_attr: HdlAttr<ExprOptions, kw::hdl>,
        mut expr_repeat: ExprRepeat,
    ) -> Expr {
        let repeated_value = &expr_repeat.expr;
        let ExprOptions { sim } = hdl_attr.body;
        let span = hdl_attr.kw.span;
        if sim.is_some() {
            *expr_repeat.expr = parse_quote_spanned! {repeated_value.span()=>
                ::fayalite::sim::value::ToSimValue::to_sim_value(&(#repeated_value))
            };
            parse_quote_spanned! {span=>
                ::fayalite::sim::value::ToSimValue::into_sim_value(#expr_repeat)
            }
        } else {
            *expr_repeat.expr = parse_quote_spanned! {repeated_value.span()=>
                ::fayalite::expr::ToExpr::to_expr(&(#repeated_value))
            };
            parse_quote_spanned! {span=>
                ::fayalite::expr::ToExpr::to_expr(&#expr_repeat)
            }
        }
    }
    pub(crate) fn process_hdl_struct(
        &mut self,
        hdl_attr: HdlAttr<ExprOptions, kw::hdl>,
        mut expr_struct: ExprStruct,
    ) -> Expr {
        let name_span = expr_struct.path.segments.last().unwrap().ident.span();
        let ExprOptions { sim } = hdl_attr.body;
        if sim.is_some() {
            let ty_path = TypePath {
                qself: expr_struct.qself.take(),
                path: expr_struct.path,
            };
            expr_struct.path = parse_quote_spanned! {name_span=>
                __SimValue::<#ty_path>
            };
            for field in &mut expr_struct.fields {
                let expr = &field.expr;
                field.expr = parse_quote_spanned! {field.member.span()=>
                    ::fayalite::sim::value::ToSimValue::to_sim_value(&(#expr))
                };
            }
            return parse_quote_spanned! {name_span=>
                {
                    type __SimValue<T> = <T as ::fayalite::ty::Type>::SimValue;
                    let value: ::fayalite::sim::value::SimValue<#ty_path> = ::fayalite::sim::value::ToSimValue::into_sim_value(#expr_struct);
                    value
                }
            };
        }
        let builder_ident = format_ident!("__builder", span = name_span);
        let empty_builder = if expr_struct.qself.is_some()
            || expr_struct
                .path
                .segments
                .iter()
                .any(|seg| !seg.arguments.is_none())
        {
            let ty = TypePath {
                qself: expr_struct.qself,
                path: expr_struct.path,
            };
            let builder_ty = quote_spanned! {name_span=>
                <#ty as ::fayalite::bundle::BundleType>::Builder
            };
            quote_spanned! {name_span=>
                <#builder_ty as ::fayalite::__std::default::Default>::default()
            }
        } else {
            let path = ExprPath {
                attrs: vec![],
                qself: expr_struct.qself,
                path: expr_struct.path,
            };
            quote_spanned! {name_span=>
                #path::__bundle_builder()
            }
        };
        let field_calls = Vec::from_iter(expr_struct.fields.iter().map(
            |FieldValue {
                 attrs: _,
                 member,
                 colon_token: _,
                 expr,
             }| {
                let field_fn = format_ident!("field_{}", member);
                quote_spanned! {member.span()=>
                    let #builder_ident = #builder_ident.#field_fn(#expr);
                }
            },
        ));
        parse_quote_spanned! {name_span=>
            {
                let #builder_ident = #empty_builder;
                #(#field_calls)*
                ::fayalite::expr::ToExpr::to_expr(&#builder_ident)
            }
        }
    }
    pub(crate) fn process_hdl_tuple(
        &mut self,
        hdl_attr: HdlAttr<ExprOptions, kw::hdl>,
        mut expr_tuple: ExprTuple,
    ) -> Expr {
        let ExprOptions { sim } = hdl_attr.body;
        if sim.is_some() {
            for element in &mut expr_tuple.elems {
                *element = parse_quote_spanned! {element.span()=>
                    &(#element)
                };
            }
            parse_quote_spanned! {expr_tuple.span()=>
                ::fayalite::sim::value::ToSimValue::into_sim_value(#expr_tuple)
            }
        } else {
            parse_quote_spanned! {expr_tuple.span()=>
                ::fayalite::expr::ToExpr::to_expr(&#expr_tuple)
            }
        }
    }
    pub(crate) fn process_hdl_call(
        &mut self,
        hdl_attr: HdlAttr<ExprOptions, kw::hdl>,
        mut expr_call: ExprCall,
    ) -> Expr {
        let span = hdl_attr.kw.span;
        let mut func = &mut *expr_call.func;
        let EnumPath {
            variant_path: _,
            enum_path,
            variant_name,
        } = loop {
            match func {
                Expr::Group(ExprGroup { expr, .. }) | Expr::Paren(ExprParen { expr, .. }) => {
                    func = &mut **expr;
                }
                Expr::Path(_) => {
                    let Expr::Path(ExprPath { attrs, qself, path }) =
                        mem::replace(func, Expr::PLACEHOLDER)
                    else {
                        unreachable!();
                    };
                    match parse_enum_path(TypePath { qself, path }) {
                        Ok(path) => break path,
                        Err(path) => {
                            self.errors.error(&path, "unsupported enum variant path");
                            let TypePath { qself, path } = path;
                            *func = ExprPath { attrs, qself, path }.into();
                            return expr_call.into();
                        }
                    }
                }
                _ => {
                    self.errors.error(
                        &expr_call.func,
                        "#[hdl] function call -- function must be a possibly-parenthesized path",
                    );
                    return expr_call.into();
                }
            }
        };
        self.process_hdl_method_call(
            hdl_attr,
            ExprMethodCall {
                attrs: expr_call.attrs,
                receiver: parse_quote_spanned! {span=>
                    <#enum_path as ::fayalite::ty::StaticType>::TYPE
                },
                dot_token: Token![.](span),
                method: variant_name,
                turbofish: None,
                paren_token: expr_call.paren_token,
                args: expr_call.args,
            },
        )
    }
    pub(crate) fn process_hdl_method_call(
        &mut self,
        hdl_attr: HdlAttr<ExprOptions, kw::hdl>,
        mut expr_method_call: ExprMethodCall,
    ) -> Expr {
        let ExprOptions { sim } = hdl_attr.body;
        let span = hdl_attr.kw.span;
        // remove any number of groups and up to one paren
        let mut receiver = &mut *expr_method_call.receiver;
        let mut has_group = false;
        let receiver = loop {
            match receiver {
                Expr::Group(ExprGroup { expr, .. }) => {
                    has_group = true;
                    receiver = expr;
                }
                Expr::Paren(ExprParen { expr, .. }) => break &mut **expr,
                receiver @ Expr::Path(_) => break receiver,
                _ => {
                    if !has_group {
                        self.errors.error(
                            &expr_method_call.receiver,
                            "#[hdl] on a method call needs parenthesized receiver",
                        );
                    }
                    break &mut *expr_method_call.receiver;
                }
            }
        };
        let func = if sim.is_some() {
            parse_quote_spanned! {span=>
                ::fayalite::enum_::enum_type_to_sim_builder
            }
        } else {
            parse_quote_spanned! {span=>
                ::fayalite::enum_::assert_is_enum_type
            }
        };
        *expr_method_call.receiver = ExprCall {
            attrs: vec![],
            func,
            paren_token: Paren(span),
            args: Punctuated::from_iter([mem::replace(receiver, Expr::PLACEHOLDER)]),
        }
        .into();
        expr_method_call.into()
    }
}
