// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    expr::{
        ops::{ArrayLiteral, BundleLiteral, ExprPartialEq},
        CastToBits, Expr, ReduceBits, ToExpr,
    },
    int::{Bool, DynSize},
    intern::{Intern, Interned},
    sim::value::{SimValue, SimValuePartialEq, ToSimValue, ToSimValueWithType},
    source_location::SourceLocation,
    ty::{
        impl_match_variant_as_self, CanonicalType, MatchVariantWithoutScope, OpaqueSimValue,
        StaticType, Type, TypeProperties, TypeWithDeref,
    },
    util::HashMap,
};
use bitvec::{slice::BitSlice, vec::BitVec};
use serde::{Deserialize, Serialize};
use std::{fmt, marker::PhantomData};

#[derive(Copy, Clone, Debug, Hash, PartialEq, Eq, Serialize, Deserialize)]
pub struct BundleField {
    pub name: Interned<str>,
    pub flipped: bool,
    pub ty: CanonicalType,
}

impl BundleField {
    pub fn fmt_debug_in_struct(self, field_offset: usize) -> FmtDebugInStruct {
        FmtDebugInStruct {
            field: self,
            field_offset,
        }
    }
}

#[derive(Copy, Clone)]
pub struct FmtDebugInStruct {
    field: BundleField,
    field_offset: usize,
}

impl fmt::Debug for FmtDebugInStruct {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self {
            field: BundleField { name, flipped, ty },
            field_offset,
        } = *self;
        if flipped {
            write!(f, "#[hdl(flip)] ")?;
        }
        if f.alternate() {
            writeln!(f, "/* offset = {field_offset} */")?;
        }
        write!(f, "{name}: ")?;
        ty.fmt(f)
    }
}

impl fmt::Display for FmtDebugInStruct {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        fmt::Debug::fmt(self, f)
    }
}

#[derive(Clone, Eq)]
struct BundleImpl {
    fields: Interned<[BundleField]>,
    name_indexes: HashMap<Interned<str>, usize>,
    field_offsets: Interned<[usize]>,
    type_properties: TypeProperties,
}

impl std::hash::Hash for BundleImpl {
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        self.fields.hash(state);
    }
}

impl PartialEq for BundleImpl {
    fn eq(&self, other: &Self) -> bool {
        self.fields == other.fields
    }
}

impl std::fmt::Debug for BundleImpl {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str("Bundle ")?;
        f.debug_set()
            .entries(
                self.fields
                    .iter()
                    .enumerate()
                    .map(|(index, field)| field.fmt_debug_in_struct(self.field_offsets[index])),
            )
            .finish()
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash)]
pub struct Bundle(Interned<BundleImpl>);

impl std::fmt::Debug for Bundle {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.0.fmt(f)
    }
}

#[derive(Clone)]
pub struct BundleTypePropertiesBuilder(TypeProperties);

impl BundleTypePropertiesBuilder {
    #[must_use]
    pub const fn new() -> Self {
        Self(TypeProperties {
            is_passive: true,
            is_storable: true,
            is_castable_from_bits: true,
            bit_width: 0,
        })
    }
    pub const fn clone(&self) -> Self {
        Self(self.0)
    }
    #[must_use]
    pub const fn field(self, flipped: bool, field_props: TypeProperties) -> Self {
        let Some(bit_width) = self.0.bit_width.checked_add(field_props.bit_width) else {
            panic!("bundle is too big: bit-width overflowed");
        };
        if flipped {
            Self(TypeProperties {
                is_passive: false,
                is_storable: false,
                is_castable_from_bits: false,
                bit_width,
            })
        } else {
            Self(TypeProperties {
                is_passive: self.0.is_passive & field_props.is_passive,
                is_storable: self.0.is_storable & field_props.is_storable,
                is_castable_from_bits: self.0.is_castable_from_bits
                    & field_props.is_castable_from_bits,
                bit_width,
            })
        }
    }
    pub const fn finish(self) -> TypeProperties {
        self.0
    }
}

impl Default for BundleTypePropertiesBuilder {
    fn default() -> Self {
        Self::new()
    }
}

impl Bundle {
    #[track_caller]
    pub fn new(fields: Interned<[BundleField]>) -> Self {
        let mut name_indexes = HashMap::with_capacity_and_hasher(fields.len(), Default::default());
        let mut field_offsets = Vec::with_capacity(fields.len());
        let mut type_props_builder = BundleTypePropertiesBuilder::new();
        for (index, &BundleField { name, flipped, ty }) in fields.iter().enumerate() {
            if let Some(old_index) = name_indexes.insert(name, index) {
                panic!("duplicate field name {name:?}: at both index {old_index} and {index}");
            }
            field_offsets.push(type_props_builder.0.bit_width);
            type_props_builder = type_props_builder.field(flipped, ty.type_properties());
        }
        Self(Intern::intern_sized(BundleImpl {
            fields,
            name_indexes,
            field_offsets: Intern::intern_owned(field_offsets),
            type_properties: type_props_builder.finish(),
        }))
    }
    pub fn name_indexes(&self) -> &HashMap<Interned<str>, usize> {
        &self.0.name_indexes
    }
    pub fn field_by_name(&self, name: Interned<str>) -> Option<BundleField> {
        Some(self.0.fields[*self.0.name_indexes.get(&name)?])
    }
    pub fn field_offsets(self) -> Interned<[usize]> {
        self.0.field_offsets
    }
    pub fn type_properties(self) -> TypeProperties {
        self.0.type_properties
    }
    pub fn can_connect(self, rhs: Self) -> bool {
        if self.0.fields.len() != rhs.0.fields.len() {
            return false;
        }
        for (
            &BundleField {
                name: lhs_name,
                flipped: lhs_flipped,
                ty: lhs_ty,
            },
            &BundleField {
                name: rhs_name,
                flipped: rhs_flipped,
                ty: rhs_ty,
            },
        ) in self.0.fields.iter().zip(rhs.0.fields.iter())
        {
            if lhs_name != rhs_name || lhs_flipped != rhs_flipped || !lhs_ty.can_connect(rhs_ty) {
                return false;
            }
        }
        true
    }
}

impl Type for Bundle {
    type BaseType = Bundle;
    type MaskType = Bundle;
    type SimValue = OpaqueSimValue;
    impl_match_variant_as_self!();
    fn mask_type(&self) -> Self::MaskType {
        Self::new(Interned::from_iter(self.0.fields.into_iter().map(
            |BundleField { name, flipped, ty }| BundleField {
                name,
                flipped,
                ty: ty.mask_type(),
            },
        )))
    }
    fn canonical(&self) -> CanonicalType {
        CanonicalType::Bundle(*self)
    }
    #[track_caller]
    fn from_canonical(canonical_type: CanonicalType) -> Self {
        let CanonicalType::Bundle(bundle) = canonical_type else {
            panic!("expected bundle");
        };
        bundle
    }
    fn source_location() -> SourceLocation {
        SourceLocation::builtin()
    }
    fn sim_value_from_bits(&self, bits: &BitSlice) -> Self::SimValue {
        assert_eq!(bits.len(), self.type_properties().bit_width);
        OpaqueSimValue::from_bitslice(bits)
    }
    fn sim_value_clone_from_bits(&self, value: &mut Self::SimValue, bits: &BitSlice) {
        assert_eq!(bits.len(), self.type_properties().bit_width);
        assert_eq!(value.bit_width(), self.type_properties().bit_width);
        value.bits_mut().bits_mut().copy_from_bitslice(bits);
    }
    fn sim_value_to_bits(&self, value: &Self::SimValue, bits: &mut BitSlice) {
        assert_eq!(bits.len(), self.type_properties().bit_width);
        assert_eq!(value.bit_width(), self.type_properties().bit_width);
        bits.copy_from_bitslice(value.bits().bits());
    }
}

pub trait BundleType: Type<BaseType = Bundle> {
    type Builder: Default;
    type FilledBuilder: ToExpr<Type = Self>;
    fn fields(&self) -> Interned<[BundleField]>;
}

pub struct BundleSimValueFromBits<'a> {
    fields: std::slice::Iter<'static, BundleField>,
    bits: &'a BitSlice,
}

impl<'a> BundleSimValueFromBits<'a> {
    #[track_caller]
    pub fn new<T: BundleType>(bundle_ty: T, bits: &'a BitSlice) -> Self {
        let fields = bundle_ty.fields();
        assert_eq!(
            bits.len(),
            fields
                .iter()
                .map(|BundleField { ty, .. }| ty.bit_width())
                .sum::<usize>()
        );
        Self {
            fields: Interned::into_inner(fields).iter(),
            bits,
        }
    }
    #[track_caller]
    fn field_ty_and_bits<T: Type>(&mut self) -> (T, &'a BitSlice) {
        let Some(&BundleField {
            name: _,
            flipped: _,
            ty,
        }) = self.fields.next()
        else {
            panic!("tried to read too many fields from BundleSimValueFromBits");
        };
        let (field_bits, rest) = self.bits.split_at(ty.bit_width());
        self.bits = rest;
        (T::from_canonical(ty), field_bits)
    }
    #[track_caller]
    pub fn field_from_bits<T: Type>(&mut self) -> SimValue<T> {
        let (field_ty, field_bits) = self.field_ty_and_bits::<T>();
        SimValue::from_bitslice(field_ty, field_bits)
    }
    #[track_caller]
    pub fn field_clone_from_bits<T: Type>(&mut self, field_value: &mut SimValue<T>) {
        let (field_ty, field_bits) = self.field_ty_and_bits::<T>();
        assert_eq!(field_ty, SimValue::ty(field_value));
        SimValue::bits_mut(field_value)
            .bits_mut()
            .copy_from_bitslice(field_bits);
    }
}

pub struct BundleSimValueToBits<'a> {
    fields: std::slice::Iter<'static, BundleField>,
    bits: &'a mut BitSlice,
}

impl<'a> BundleSimValueToBits<'a> {
    #[track_caller]
    pub fn new<T: BundleType>(bundle_ty: T, bits: &'a mut BitSlice) -> Self {
        let fields = bundle_ty.fields();
        assert_eq!(
            bits.len(),
            fields
                .iter()
                .map(|BundleField { ty, .. }| ty.bit_width())
                .sum::<usize>()
        );
        Self {
            fields: Interned::into_inner(fields).iter(),
            bits,
        }
    }
    #[track_caller]
    pub fn field_to_bits<T: Type>(&mut self, field_value: &SimValue<T>) {
        let Some(&BundleField {
            name: _,
            flipped: _,
            ty,
        }) = self.fields.next()
        else {
            panic!("tried to read too many fields from BundleSimValueFromBits");
        };
        assert_eq!(T::from_canonical(ty), SimValue::ty(field_value));
        self.bits[..ty.bit_width()].copy_from_bitslice(SimValue::bits(field_value).bits());
        self.bits = &mut std::mem::take(&mut self.bits)[ty.bit_width()..];
    }
}

#[derive(Default)]
pub struct NoBuilder;

pub struct Unfilled<T: Type>(PhantomData<T>);

impl<T: Type> Default for Unfilled<T> {
    fn default() -> Self {
        Self(PhantomData)
    }
}

impl BundleType for Bundle {
    type Builder = NoBuilder;
    type FilledBuilder = Expr<Bundle>;
    fn fields(&self) -> Interned<[BundleField]> {
        self.0.fields
    }
}

#[derive(Default)]
pub struct TupleBuilder<T>(T);

macro_rules! impl_tuple_builder_fields {
    (
        @impl
        {
        }
        [
            $({
                #[type_var($head_type_var:ident)]
                #[field($head_field:ident)]
                #[var($head_var:ident)]
            })*
        ]
        {
            #[type_var($cur_type_var:ident)]
            #[field($cur_field:ident)]
            #[var($cur_var:ident)]
        }
        [
            $({
                #[type_var($tail_type_var:ident)]
                #[field($tail_field:ident)]
                #[var($tail_var:ident)]
            })*
        ]
    ) => {
        impl<
                $($head_type_var,)*
                $cur_type_var: Type,
                $($tail_type_var,)*
            > TupleBuilder<(
                $($head_type_var,)*
                Unfilled<$cur_type_var>,
                $($tail_type_var,)*
            )>
        {
            pub fn $cur_field(self, $cur_var: impl ToExpr<Type = $cur_type_var>) -> TupleBuilder<(
                    $($head_type_var,)*
                    Expr<$cur_type_var>,
                    $($tail_type_var,)*
                )>
            {
                let ($($head_var,)* _, $($tail_var,)*) = self.0;
                TupleBuilder(($($head_var,)* $cur_var.to_expr(), $($tail_var,)*))
            }
        }
    };
    ($global:tt [$($head:tt)*] $cur:tt [$next:tt $($tail:tt)*]) => {
        impl_tuple_builder_fields!(@impl $global [$($head)*] $cur [$next $($tail)*]);
        impl_tuple_builder_fields!($global [$($head)* $cur] $next [$($tail)*]);
    };
    ($global:tt [$($head:tt)*] $cur:tt []) => {
        impl_tuple_builder_fields!(@impl $global [$($head)*] $cur []);
    };
    ($global:tt [$cur:tt $($tail:tt)*]) => {
        impl_tuple_builder_fields!($global [] $cur [$($tail)*]);
    };
    ($global:tt []) => {};
}

macro_rules! impl_tuples {
    (
        [$({
            #[
                num = $num:tt,
                field = $field:ident,
                ty = $ty_var:ident: $Ty:ident,
                lhs = $lhs_var:ident: $Lhs:ident,
                rhs = $rhs_var:ident: $Rhs:ident
            ]
            $var:ident: $T:ident
        })*]
        []
    ) => {
        impl_tuple_builder_fields! {
            {}
            [$({
                #[type_var($T)]
                #[field($field)]
                #[var($var)]
            })*]
        }
        impl<$($T: Type,)*> Type for ($($T,)*) {
            type BaseType = Bundle;
            type MaskType = ($($T::MaskType,)*);
            type SimValue = ($(SimValue<$T>,)*);
            type MatchVariant = ($(Expr<$T>,)*);
            type MatchActiveScope = ();
            type MatchVariantAndInactiveScope = MatchVariantWithoutScope<Self::MatchVariant>;
            type MatchVariantsIter = std::iter::Once<Self::MatchVariantAndInactiveScope>;
            fn match_variants(
                this: Expr<Self>,
                source_location: SourceLocation,
            ) -> Self::MatchVariantsIter {
                let _ = this;
                let _ = source_location;
                std::iter::once(MatchVariantWithoutScope(($(Expr::field(this, stringify!($num)),)*)))
            }
            fn mask_type(&self) -> Self::MaskType {
                #![allow(clippy::unused_unit)]
                let ($($var,)*) = self;
                ($($var.mask_type(),)*)
            }
            fn canonical(&self) -> CanonicalType {
                Bundle::new(self.fields()).canonical()
            }
            #[track_caller]
            fn from_canonical(canonical_type: CanonicalType) -> Self {
                #![allow(clippy::unused_unit)]
                let CanonicalType::Bundle(bundle) = canonical_type else {
                    panic!("expected bundle");
                };
                let [$($var,)*] = *bundle.fields() else {
                    panic!("bundle has wrong number of fields");
                };
                $(let BundleField { name, flipped, ty } = $var;
                assert_eq!(&*name, stringify!($num));
                assert!(!flipped);
                let $var = $T::from_canonical(ty);)*
                ($($var,)*)
            }
            fn source_location() -> SourceLocation {
                SourceLocation::builtin()
            }
            fn sim_value_from_bits(&self, bits: &BitSlice) -> Self::SimValue {
                #![allow(unused_mut, unused_variables)]
                let mut v = BundleSimValueFromBits::new(*self, bits);
                $(let $var = v.field_from_bits();)*
                ($($var,)*)
            }
            fn sim_value_clone_from_bits(&self, value: &mut Self::SimValue, bits: &BitSlice) {
                #![allow(unused_mut, unused_variables)]
                let mut v = BundleSimValueFromBits::new(*self, bits);
                let ($($var,)*) = value;
                $(v.field_clone_from_bits($var);)*
            }
            fn sim_value_to_bits(&self, value: &Self::SimValue, bits: &mut BitSlice) {
                #![allow(unused_mut, unused_variables)]
                let mut v = BundleSimValueToBits::new(*self, bits);
                let ($($var,)*) = value;
                $(v.field_to_bits($var);)*
            }
        }
        impl<$($T: Type,)*> BundleType for ($($T,)*) {
            type Builder = TupleBuilder<($(Unfilled<$T>,)*)>;
            type FilledBuilder = TupleBuilder<($(Expr<$T>,)*)>;
            fn fields(&self) -> Interned<[BundleField]> {
                let ($($var,)*) = self;
                [$(BundleField { name: stringify!($num).intern(), flipped: false, ty: $var.canonical() }),*][..].intern()
            }
        }
        impl<$($T: Type,)*> TypeWithDeref for ($($T,)*) {
            fn expr_deref(this: &Expr<Self>) -> &Self::MatchVariant {
                let _ = this;
                Interned::into_inner(($(Expr::field(*this, stringify!($num)),)*).intern_sized())
            }
        }
        impl<$($T: StaticType,)*> StaticType for ($($T,)*) {
            const TYPE: Self = ($($T::TYPE,)*);
            const MASK_TYPE: Self::MaskType = ($($T::MASK_TYPE,)*);
            const TYPE_PROPERTIES: TypeProperties = {
                let builder = BundleTypePropertiesBuilder::new();
                $(let builder = builder.field(false, $T::TYPE_PROPERTIES);)*
                builder.finish()
            };
            const MASK_TYPE_PROPERTIES: TypeProperties = {
                let builder = BundleTypePropertiesBuilder::new();
                $(let builder = builder.field(false, $T::MASK_TYPE_PROPERTIES);)*
                builder.finish()
            };
        }
        impl<$($T: ToExpr,)*> ToExpr for ($($T,)*) {
            type Type = ($($T::Type,)*);

            fn to_expr(&self) -> Expr<Self::Type> {
                let ($($var,)*) = self;
                $(let $var = $var.to_expr();)*
                let ty = ($(Expr::ty($var),)*);
                let field_values = [$(Expr::canonical($var)),*];
                BundleLiteral::new(ty, field_values[..].intern()).to_expr()
            }
        }
        impl<$($T: Type,)*> ToExpr for TupleBuilder<($(Expr<$T>,)*)> {
            type Type = ($($T,)*);

            fn to_expr(&self) -> Expr<Self::Type> {
                let ($($var,)*) = self.0;
                let ty = ($(Expr::ty($var),)*);
                let field_values = [$(Expr::canonical($var)),*];
                BundleLiteral::new(ty, field_values[..].intern()).to_expr()
            }
        }
        impl<$($T: ToSimValueWithType<CanonicalType>,)*> ToSimValueWithType<CanonicalType> for ($($T,)*) {
            #[track_caller]
            fn to_sim_value_with_type(&self, ty: CanonicalType) -> SimValue<CanonicalType> {
                SimValue::into_canonical(ToSimValueWithType::<Bundle>::to_sim_value_with_type(self, Bundle::from_canonical(ty)))
            }
            #[track_caller]
            fn into_sim_value_with_type(self, ty: CanonicalType) -> SimValue<CanonicalType>
            {
                SimValue::into_canonical(ToSimValueWithType::<Bundle>::into_sim_value_with_type(self, Bundle::from_canonical(ty)))
            }
        }
        impl<$($T: ToSimValueWithType<CanonicalType>,)*> ToSimValueWithType<Bundle> for ($($T,)*) {
            #[track_caller]
            fn to_sim_value_with_type(&self, ty: Bundle) -> SimValue<Bundle> {
                let ($($var,)*) = self;
                let [$($ty_var,)*] = *ty.fields() else {
                    panic!("bundle has wrong number of fields");
                };
                $(let $var = $var.to_sim_value_with_type($ty_var.ty);)*
                ToSimValueWithType::into_sim_value_with_type(($($var,)*), ty)
            }
            #[track_caller]
            fn into_sim_value_with_type(self, ty: Bundle) -> SimValue<Bundle> {
                #![allow(unused_mut)]
                #![allow(clippy::unused_unit)]
                let ($($var,)*) = self;
                let [$($ty_var,)*] = *ty.fields() else {
                    panic!("bundle has wrong number of fields");
                };
                let mut bits = BitVec::new();
                $(let $var = $var.into_sim_value_with_type($ty_var.ty);
                assert_eq!(SimValue::ty(&$var), $ty_var.ty);
                bits.extend_from_bitslice(SimValue::bits(&$var).bits());
                )*
                bits.into_sim_value_with_type(ty)
            }
        }
        impl<$($T: ToSimValueWithType<$Ty>, $Ty: Type,)*> ToSimValueWithType<($($Ty,)*)> for ($($T,)*) {
            #[track_caller]
            fn to_sim_value_with_type(&self, ty: ($($Ty,)*)) -> SimValue<($($Ty,)*)> {
                let ($($var,)*) = self;
                let ($($ty_var,)*) = ty;
                $(let $var = $var.to_sim_value_with_type($ty_var);)*
                SimValue::from_value(ty, ($($var,)*))
            }
            #[track_caller]
            fn into_sim_value_with_type(self, ty: ($($Ty,)*)) -> SimValue<($($Ty,)*)> {
                let ($($var,)*) = self;
                let ($($ty_var,)*) = ty;
                $(let $var = $var.into_sim_value_with_type($ty_var);)*
                SimValue::from_value(ty, ($($var,)*))
            }
        }
        impl<$($T: ToSimValue,)*> ToSimValue for ($($T,)*) {
            type Type = ($($T::Type,)*);
            #[track_caller]
            fn to_sim_value(&self) -> SimValue<Self::Type> {
                let ($($var,)*) = self;
                $(let $var = $var.to_sim_value();)*
                SimValue::from_value(($(SimValue::ty(&$var),)*), ($($var,)*))
            }
            #[track_caller]
            fn into_sim_value(self) -> SimValue<Self::Type> {
                let ($($var,)*) = self;
                $(let $var = $var.to_sim_value();)*
                SimValue::from_value(($(SimValue::ty(&$var),)*), ($($var,)*))
            }
        }
        impl<$($Lhs: Type + ExprPartialEq<$Rhs>, $Rhs: Type,)*> ExprPartialEq<($($Rhs,)*)> for ($($Lhs,)*) {
            fn cmp_eq(lhs: Expr<Self>, rhs: Expr<($($Rhs,)*)>) -> Expr<Bool> {
                let ($($lhs_var,)*) = *lhs;
                let ($($rhs_var,)*) = *rhs;
                ArrayLiteral::<Bool, DynSize>::new(
                    Bool,
                    FromIterator::from_iter([$(Expr::canonical(ExprPartialEq::cmp_eq($lhs_var, $rhs_var)),)*]),
                )
                .cast_to_bits()
                .all_one_bits()
            }

            fn cmp_ne(lhs: Expr<Self>, rhs: Expr<($($Rhs,)*)>) -> Expr<Bool> {
                let ($($lhs_var,)*) = *lhs;
                let ($($rhs_var,)*) = *rhs;
                ArrayLiteral::<Bool, DynSize>::new(
                    Bool,
                    FromIterator::from_iter([$(Expr::canonical(ExprPartialEq::cmp_ne($lhs_var, $rhs_var)),)*]),
                )
                .cast_to_bits()
                .any_one_bits()
            }
        }
        impl<$($Lhs: SimValuePartialEq<$Rhs>, $Rhs: Type,)*> SimValuePartialEq<($($Rhs,)*)> for ($($Lhs,)*) {
            fn sim_value_eq(lhs: &SimValue<Self>, rhs: &SimValue<($($Rhs,)*)>) -> bool {
                let ($($lhs_var,)*) = &**lhs;
                let ($($rhs_var,)*) = &**rhs;
                let retval = true;
                $(let retval = retval && $lhs_var == $rhs_var;)*
                retval
            }
        }
    };
    ([$($lhs:tt)*] [$rhs_first:tt $($rhs:tt)*]) => {
        impl_tuples!([$($lhs)*] []);
        impl_tuples!([$($lhs)* $rhs_first] [$($rhs)*]);
    };
}

impl_tuples! {
    [] [
        {#[num = 0, field = field_0, ty = ty0: Ty0, lhs = lhs0: Lhs0, rhs = rhs0: Rhs0] v0: T0}
        {#[num = 1, field = field_1, ty = ty1: Ty1, lhs = lhs1: Lhs1, rhs = rhs1: Rhs1] v1: T1}
        {#[num = 2, field = field_2, ty = ty2: Ty2, lhs = lhs2: Lhs2, rhs = rhs2: Rhs2] v2: T2}
        {#[num = 3, field = field_3, ty = ty3: Ty3, lhs = lhs3: Lhs3, rhs = rhs3: Rhs3] v3: T3}
        {#[num = 4, field = field_4, ty = ty4: Ty4, lhs = lhs4: Lhs4, rhs = rhs4: Rhs4] v4: T4}
        {#[num = 5, field = field_5, ty = ty5: Ty5, lhs = lhs5: Lhs5, rhs = rhs5: Rhs5] v5: T5}
        {#[num = 6, field = field_6, ty = ty6: Ty6, lhs = lhs6: Lhs6, rhs = rhs6: Rhs6] v6: T6}
        {#[num = 7, field = field_7, ty = ty7: Ty7, lhs = lhs7: Lhs7, rhs = rhs7: Rhs7] v7: T7}
        {#[num = 8, field = field_8, ty = ty8: Ty8, lhs = lhs8: Lhs8, rhs = rhs8: Rhs8] v8: T8}
        {#[num = 9, field = field_9, ty = ty9: Ty9, lhs = lhs9: Lhs9, rhs = rhs9: Rhs9] v9: T9}
        {#[num = 10, field = field_10, ty = ty10: Ty10, lhs = lhs10: Lhs10, rhs = rhs10: Rhs10] v10: T10}
        {#[num = 11, field = field_11, ty = ty11: Ty11, lhs = lhs11: Lhs11, rhs = rhs11: Rhs11] v11: T11}
    ]
}

impl<T: ?Sized + Send + Sync + 'static> Type for PhantomData<T> {
    type BaseType = Bundle;
    type MaskType = ();
    type SimValue = PhantomData<T>;
    type MatchVariant = PhantomData<T>;
    type MatchActiveScope = ();
    type MatchVariantAndInactiveScope = MatchVariantWithoutScope<Self::MatchVariant>;
    type MatchVariantsIter = std::iter::Once<Self::MatchVariantAndInactiveScope>;
    fn match_variants(
        this: Expr<Self>,
        source_location: SourceLocation,
    ) -> Self::MatchVariantsIter {
        let _ = this;
        let _ = source_location;
        std::iter::once(MatchVariantWithoutScope(PhantomData))
    }
    fn mask_type(&self) -> Self::MaskType {
        ()
    }
    fn canonical(&self) -> CanonicalType {
        Bundle::new(self.fields()).canonical()
    }
    #[track_caller]
    fn from_canonical(canonical_type: CanonicalType) -> Self {
        let CanonicalType::Bundle(bundle) = canonical_type else {
            panic!("expected bundle");
        };
        assert!(
            bundle.fields().is_empty(),
            "bundle has wrong number of fields"
        );
        PhantomData
    }
    fn source_location() -> SourceLocation {
        SourceLocation::builtin()
    }
    fn sim_value_from_bits(&self, bits: &BitSlice) -> Self::SimValue {
        assert!(bits.is_empty());
        *self
    }
    fn sim_value_clone_from_bits(&self, _value: &mut Self::SimValue, bits: &BitSlice) {
        assert!(bits.is_empty());
    }
    fn sim_value_to_bits(&self, _value: &Self::SimValue, bits: &mut BitSlice) {
        assert!(bits.is_empty());
    }
}

pub struct PhantomDataBuilder<T: ?Sized + Send + Sync + 'static>(PhantomData<T>);

impl<T: ?Sized + Send + Sync + 'static> Default for PhantomDataBuilder<T> {
    fn default() -> Self {
        Self(PhantomData)
    }
}

impl<T: ?Sized + Send + Sync + 'static> ToExpr for PhantomDataBuilder<T> {
    type Type = PhantomData<T>;

    fn to_expr(&self) -> Expr<Self::Type> {
        PhantomData.to_expr()
    }
}

impl<T: ?Sized + Send + Sync + 'static> BundleType for PhantomData<T> {
    type Builder = PhantomDataBuilder<T>;
    type FilledBuilder = PhantomDataBuilder<T>;
    fn fields(&self) -> Interned<[BundleField]> {
        Interned::default()
    }
}

impl<T: ?Sized + Send + Sync + 'static> TypeWithDeref for PhantomData<T> {
    fn expr_deref(_this: &Expr<Self>) -> &Self::MatchVariant {
        &PhantomData
    }
}

impl<T: ?Sized + Send + Sync + 'static> StaticType for PhantomData<T> {
    const TYPE: Self = PhantomData;
    const MASK_TYPE: Self::MaskType = ();
    const TYPE_PROPERTIES: TypeProperties = <()>::TYPE_PROPERTIES;
    const MASK_TYPE_PROPERTIES: TypeProperties = <()>::TYPE_PROPERTIES;
}

impl<T: ?Sized + Send + Sync + 'static> ToExpr for PhantomData<T> {
    type Type = PhantomData<T>;

    fn to_expr(&self) -> Expr<Self::Type> {
        BundleLiteral::new(PhantomData, Interned::default()).to_expr()
    }
}

impl<T: ?Sized + Send + Sync + 'static> ToSimValue for PhantomData<T> {
    type Type = PhantomData<T>;

    #[track_caller]
    fn to_sim_value(&self) -> SimValue<Self> {
        SimValue::from_value(*self, *self)
    }
}

impl<T: ?Sized + Send + Sync + 'static> ToSimValueWithType<Self> for PhantomData<T> {
    #[track_caller]
    fn to_sim_value_with_type(&self, ty: Self) -> SimValue<Self> {
        SimValue::from_value(ty, *self)
    }
}

impl<T: ?Sized> ToSimValueWithType<Bundle> for PhantomData<T> {
    #[track_caller]
    fn to_sim_value_with_type(&self, ty: Bundle) -> SimValue<Bundle> {
        assert!(ty.fields().is_empty());
        ToSimValueWithType::into_sim_value_with_type(BitVec::new(), ty)
    }
}

impl<T: ?Sized> ToSimValueWithType<CanonicalType> for PhantomData<T> {
    #[track_caller]
    fn to_sim_value_with_type(&self, ty: CanonicalType) -> SimValue<CanonicalType> {
        let ty = Bundle::from_canonical(ty);
        assert!(ty.fields().is_empty());
        SimValue::into_canonical(ToSimValueWithType::into_sim_value_with_type(
            BitVec::new(),
            ty,
        ))
    }
}
