// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use std::{
    any::{Any, TypeId},
    hash::{BuildHasher, Hasher},
    sync::RwLock,
};

struct TypeIdHasher(u64);

// assumes TypeId has at least 64 bits that is a good hash
impl Hasher for TypeIdHasher {
    #[inline(always)]
    fn finish(&self) -> u64 {
        self.0
    }

    #[inline(always)]
    fn write(&mut self, bytes: &[u8]) {
        for &byte in bytes {
            self.write_u8(byte);
        }
    }

    #[inline(always)]
    fn write_u8(&mut self, i: u8) {
        self.0 = self.0.rotate_left(8);
        self.0 ^= i as u64;
    }

    #[inline(always)]
    fn write_u16(&mut self, i: u16) {
        self.0 = self.0.rotate_left(16);
        self.0 ^= i as u64;
    }

    #[inline(always)]
    fn write_u32(&mut self, i: u32) {
        self.0 = self.0.rotate_left(32);
        self.0 ^= i as u64;
    }

    #[inline(always)]
    fn write_u64(&mut self, i: u64) {
        self.0 ^= i;
    }

    #[inline(always)]
    fn write_u128(&mut self, i: u128) {
        self.write_u64(i as u64);
        self.write_u64((i >> 64) as u64);
    }

    #[inline(always)]
    fn write_usize(&mut self, i: usize) {
        match usize::BITS {
            128 => self.write_u128(i as u128),
            64 => self.write_u64(i as u64),
            32 => self.write_u32(i as u32),
            16 => self.write_u16(i as u16),
            _ => self.write(&i.to_ne_bytes()),
        }
    }
}

struct TypeIdBuildHasher;

impl BuildHasher for TypeIdBuildHasher {
    type Hasher = TypeIdHasher;

    fn build_hasher(&self) -> Self::Hasher {
        TypeIdHasher(0)
    }
}

pub(crate) struct TypeIdMap(
    RwLock<hashbrown::HashMap<TypeId, &'static (dyn Any + Send + Sync), TypeIdBuildHasher>>,
);

impl TypeIdMap {
    pub(crate) const fn new() -> Self {
        Self(RwLock::new(hashbrown::HashMap::with_hasher(
            TypeIdBuildHasher,
        )))
    }
    #[cold]
    fn insert_slow(
        &self,
        type_id: TypeId,
        make: fn() -> Box<dyn Any + Sync + Send>,
    ) -> &'static (dyn Any + Sync + Send) {
        let value = Box::leak(make());
        let mut write_guard = self.0.write().unwrap();
        *write_guard.entry(type_id).or_insert(value)
    }
    pub(crate) fn get_or_insert_default<T: Sized + Any + Send + Sync + Default>(&self) -> &T {
        let type_id = TypeId::of::<T>();
        let read_guard = self.0.read().unwrap();
        let retval = read_guard.get(&type_id).map(|v| *v);
        drop(read_guard);
        let retval = match retval {
            Some(retval) => retval,
            None => self.insert_slow(type_id, move || Box::new(T::default())),
        };
        retval.downcast_ref().expect("known to have correct TypeId")
    }
}

impl Default for TypeIdMap {
    fn default() -> Self {
        Self::new()
    }
}
