// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use crate::{
    clock::Clock,
    expr::{ops, Expr, ToExpr},
    int::{Bool, SInt, UInt},
    source_location::SourceLocation,
    ty::{impl_match_variant_as_self, CanonicalType, StaticType, Type, TypeProperties},
};
use bitvec::slice::BitSlice;

mod sealed {
    pub trait ResetTypeSealed {}
}

pub trait ResetType:
    StaticType<MaskType = Bool>
    + sealed::ResetTypeSealed
    + ops::ExprCastTo<Bool>
    + ops::ExprCastTo<Reset>
    + ops::ExprCastTo<SyncReset>
    + ops::ExprCastTo<AsyncReset>
    + ops::ExprCastTo<Clock>
    + ops::ExprCastTo<UInt<1>>
    + ops::ExprCastTo<SInt<1>>
    + ops::ExprCastTo<UInt>
    + ops::ExprCastTo<SInt>
{
    fn dispatch<D: ResetTypeDispatch>(input: D::Input<Self>, dispatch: D) -> D::Output<Self>;
}

pub trait ResetTypeDispatch: Sized {
    type Input<T: ResetType>;
    type Output<T: ResetType>;

    fn reset(self, input: Self::Input<Reset>) -> Self::Output<Reset>;
    fn sync_reset(self, input: Self::Input<SyncReset>) -> Self::Output<SyncReset>;
    fn async_reset(self, input: Self::Input<AsyncReset>) -> Self::Output<AsyncReset>;
}

macro_rules! reset_type {
    ($name:ident, $(#[$impl_trait:ident])? $Trait:ident::$trait_fn:ident, $is_castable_from_bits:literal, $dispatch_fn:ident) => {
        #[derive(Copy, Clone, Eq, PartialEq, Hash, Debug, Default)]
        pub struct $name;

        impl Type for $name {
            type BaseType = $name;
            type MaskType = Bool;
            type SimValue = bool;

            impl_match_variant_as_self!();

            fn mask_type(&self) -> Self::MaskType {
                Bool
            }

            fn canonical(&self) -> CanonicalType {
                CanonicalType::$name(*self)
            }

            fn source_location() -> SourceLocation {
                SourceLocation::builtin()
            }

            fn from_canonical(canonical_type: CanonicalType) -> Self {
                let CanonicalType::$name(retval) = canonical_type else {
                    panic!("expected {}", stringify!($name));
                };
                retval
            }

            fn sim_value_from_bits(&self, bits: &BitSlice) -> Self::SimValue {
                assert_eq!(bits.len(), 1);
                bits[0]
            }

            fn sim_value_clone_from_bits(&self, value: &mut Self::SimValue, bits: &BitSlice) {
                assert_eq!(bits.len(), 1);
                *value = bits[0];
            }

            fn sim_value_to_bits(&self, value: &Self::SimValue, bits: &mut BitSlice) {
                assert_eq!(bits.len(), 1);
                bits.set(0, *value);
            }
        }

        impl $name {
            pub fn type_properties(self) -> TypeProperties {
                Self::TYPE_PROPERTIES
            }
            pub fn can_connect(self, _rhs: Self) -> bool {
                true
            }
        }

        impl StaticType for $name {
            const TYPE: Self = Self;
            const MASK_TYPE: Self::MaskType = Bool;
            const TYPE_PROPERTIES: TypeProperties = TypeProperties {
                is_passive: true,
                is_storable: false,
                is_castable_from_bits: $is_castable_from_bits,
                bit_width: 1,
            };
            const MASK_TYPE_PROPERTIES: TypeProperties = Bool::TYPE_PROPERTIES;
        }

        impl sealed::ResetTypeSealed for $name {}

        impl ResetType for $name {
            fn dispatch<D: ResetTypeDispatch>(
                input: D::Input<Self>,
                dispatch: D,
            ) -> D::Output<Self> {
                dispatch.$dispatch_fn(input)
            }
        }

        pub trait $Trait {
            fn $trait_fn(&self) -> Expr<$name>;
        }

        impl<T: ?Sized + $Trait> $Trait for &'_ T {
            fn $trait_fn(&self) -> Expr<$name> {
                (**self).$trait_fn()
            }
        }

        impl<T: ?Sized + $Trait> $Trait for &'_ mut T {
            fn $trait_fn(&self) -> Expr<$name> {
                (**self).$trait_fn()
            }
        }

        impl<T: ?Sized + $Trait> $Trait for Box<T> {
            fn $trait_fn(&self) -> Expr<$name> {
                (**self).$trait_fn()
            }
        }

        $($impl_trait $Trait for Expr<$name> {
            fn $trait_fn(&self) -> Expr<$name> {
                *self
            }
        })?
    };
}

reset_type!(AsyncReset, #[impl] ToAsyncReset::to_async_reset, true, async_reset);
reset_type!(SyncReset, #[impl] ToSyncReset::to_sync_reset, true, sync_reset);
reset_type!(
    Reset,
    ToReset::to_reset,
    false, // Reset is not castable from bits because we don't know if it's async or sync
    reset
);

impl ToSyncReset for bool {
    fn to_sync_reset(&self) -> Expr<SyncReset> {
        self.to_expr().to_sync_reset()
    }
}

impl ToAsyncReset for bool {
    fn to_async_reset(&self) -> Expr<AsyncReset> {
        self.to_expr().to_async_reset()
    }
}
