// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use std::{
    fmt,
    ops::{Add, AddAssign, Sub, SubAssign},
    time::Duration,
};

#[derive(Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash, Default)]
pub struct SimInstant {
    time_since_start: SimDuration,
}

impl SimInstant {
    pub const fn checked_add(self, duration: SimDuration) -> Option<Self> {
        let Some(time_since_start) = self.time_since_start.checked_add(duration) else {
            return None;
        };
        Some(SimInstant { time_since_start })
    }
    pub const fn checked_duration_since(self, earlier: Self) -> Option<SimDuration> {
        self.time_since_start.checked_sub(earlier.time_since_start)
    }
    pub const fn checked_sub(self, duration: SimDuration) -> Option<Self> {
        let Some(time_since_start) = self.time_since_start.checked_sub(duration) else {
            return None;
        };
        Some(SimInstant { time_since_start })
    }
    #[track_caller]
    pub const fn duration_since(self, earlier: Self) -> SimDuration {
        let Some(retval) = self.checked_duration_since(earlier) else {
            panic!(
                "tried to compute the duration since a later time -- durations can't be negative"
            );
        };
        retval
    }
    pub const fn saturating_duration_since(self, earlier: Self) -> SimDuration {
        let Some(retval) = self.checked_duration_since(earlier) else {
            return SimDuration::ZERO;
        };
        retval
    }
}

impl Add<SimDuration> for SimInstant {
    type Output = SimInstant;

    #[track_caller]
    fn add(mut self, rhs: SimDuration) -> Self::Output {
        self += rhs;
        self
    }
}

impl AddAssign<SimDuration> for SimInstant {
    #[track_caller]
    fn add_assign(&mut self, rhs: SimDuration) {
        self.time_since_start += rhs;
    }
}

impl Add<SimInstant> for SimDuration {
    type Output = SimInstant;

    #[track_caller]
    fn add(self, rhs: SimInstant) -> Self::Output {
        rhs.add(self)
    }
}

impl Sub for SimInstant {
    type Output = SimDuration;

    #[track_caller]
    fn sub(self, rhs: SimInstant) -> Self::Output {
        self.duration_since(rhs)
    }
}

impl Sub<SimDuration> for SimInstant {
    type Output = SimInstant;

    #[track_caller]
    fn sub(self, rhs: SimDuration) -> Self::Output {
        let Some(retval) = self.checked_sub(rhs) else {
            panic!("SimInstant underflow");
        };
        retval
    }
}

impl SubAssign<SimDuration> for SimInstant {
    #[track_caller]
    fn sub_assign(&mut self, rhs: SimDuration) {
        *self = *self - rhs;
    }
}

impl SimInstant {
    pub const START: SimInstant = SimInstant {
        time_since_start: SimDuration::ZERO,
    };
}

impl fmt::Debug for SimInstant {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.time_since_start.fmt(f)
    }
}

#[derive(Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash, Default)]
pub struct SimDuration {
    attos: u128,
}

impl AddAssign for SimDuration {
    #[track_caller]
    fn add_assign(&mut self, rhs: SimDuration) {
        *self = *self + rhs;
    }
}

impl Add for SimDuration {
    type Output = SimDuration;

    #[track_caller]
    fn add(self, rhs: SimDuration) -> Self::Output {
        SimDuration {
            attos: self
                .attos
                .checked_add(rhs.attos)
                .expect("overflow adding durations"),
        }
    }
}

impl Sub for SimDuration {
    type Output = Self;

    #[track_caller]
    fn sub(self, rhs: Self) -> Self::Output {
        SimDuration {
            attos: self
                .attos
                .checked_add(rhs.attos)
                .expect("underflow subtracting durations -- durations can't be negative"),
        }
    }
}

impl SubAssign for SimDuration {
    #[track_caller]
    fn sub_assign(&mut self, rhs: Self) {
        *self = *self - rhs;
    }
}

#[derive(Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash, Debug, Default)]
pub struct SimDurationParts {
    pub attos: u16,
    pub femtos: u16,
    pub picos: u16,
    pub nanos: u16,
    pub micros: u16,
    pub millis: u16,
    pub secs: u128,
}

macro_rules! impl_duration_units {
    (
        $(
            #[unit_const = $UNIT:ident, from_units = $from_units:ident, as_units = $as_units:ident, units = $units:ident, suffix = $suffix:literal]
            const $log10_units_per_sec:ident: u32 = $log10_units_per_sec_value:expr;
        )*
    ) => {
        impl SimDuration {
            $(
                const $log10_units_per_sec: u32 = $log10_units_per_sec_value;
                pub const fn $from_units($units: u128) -> Self {
                    Self::from_units_helper::<{ Self::$log10_units_per_sec }>($units)
                }
                pub const fn $as_units(self) -> u128 {
                    self.attos / const { 10u128.pow(Self::LOG10_ATTOS_PER_SEC - Self::$log10_units_per_sec) }
                }
            )*
            pub const fn to_parts(mut self) -> SimDurationParts {
                $(
                    let $units = self.attos / const { 10u128.pow(Self::LOG10_ATTOS_PER_SEC - Self::$log10_units_per_sec) };
                    self.attos %= const { 10u128.pow(Self::LOG10_ATTOS_PER_SEC - Self::$log10_units_per_sec) };
                )*
                SimDurationParts {
                    $($units: $units as _,)*
                }
            }
            pub const fn from_parts_checked(parts: SimDurationParts) -> Option<Self> {
                let attos = 0u128;
                $(
                    let Some(product) = const { 10u128.pow(Self::LOG10_ATTOS_PER_SEC - Self::$log10_units_per_sec) }.checked_mul(parts.$units as u128) else {
                        return None;
                    };
                    let Some(attos) = attos.checked_add(product) else {
                        return None;
                    };
                )*
                Some(Self {
                    attos,
                })
            }
        }

        impl fmt::Debug for SimDuration {
            fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
                let ilog10_attos = match self.attos.checked_ilog10() {
                    Some(v) => v,
                    None => Self::LOG10_ATTOS_PER_SEC,
                };
                let (suffix, int, fraction, fraction_digits) =
                    match Self::LOG10_ATTOS_PER_SEC.saturating_sub(ilog10_attos) {
                    $(
                        ..=Self::$log10_units_per_sec => {
                            let divisor = const { 10u128.pow(Self::LOG10_ATTOS_PER_SEC - Self::$log10_units_per_sec) };
                            (
                                $suffix,
                                self.attos / divisor,
                                self.attos % divisor,
                                (Self::LOG10_ATTOS_PER_SEC - Self::$log10_units_per_sec) as usize,
                            )
                        },
                    )*
                    _ => unreachable!(),
                };
                write!(f, "{int}")?;
                if fraction != 0 {
                    write!(f, ".{fraction:0fraction_digits$}")?;
                }
                write!(f, " {suffix}")
            }
        }

        #[cfg(test)]
        #[test]
        fn test_duration_debug() {
            $(
                assert_eq!(
                    format!("{:?}", SimDuration::$from_units(123)),
                    concat!("123 ", $suffix)
                );
                assert_eq!(
                    format!("{:?}", SimDuration::$from_units(1)),
                    concat!("1 ", $suffix),
                );
                let mut v = SimDuration::$from_units(1);
                if v.attos < 1 << 53 {
                    v.attos += 1;
                    assert_eq!(
                        format!("{v:?}"),
                        format!("{} {}", v.attos as f64 / 10.0f64.powf((SimDuration::LOG10_ATTOS_PER_SEC - SimDuration::$log10_units_per_sec) as f64), $suffix),
                        "1 {} + 1 as == {} as", $suffix, v.attos,
                    );
                }
            )*
        }
    };
}

impl_duration_units! {
    #[unit_const = SECOND, from_units = from_secs, as_units = as_secs, units = secs, suffix = "s"]
    const LOG10_SECS_PER_SEC: u32 = 0;
    #[unit_const = MILLISECOND, from_units = from_millis, as_units = as_millis, units = millis, suffix = "ms"]
    const LOG10_MILLIS_PER_SEC: u32 = 3;
    #[unit_const = MICROSECOND, from_units = from_micros, as_units = as_micros, units = micros, suffix = "μs"]
    const LOG10_MICROS_PER_SEC: u32 = 6;
    #[unit_const = NANOSECOND, from_units = from_nanos, as_units = as_nanos, units = nanos, suffix = "ns"]
    const LOG10_NANOS_PER_SEC: u32 = 9;
    #[unit_const = PICOSECOND, from_units = from_picos, as_units = as_picos, units = picos, suffix = "ps"]
    const LOG10_PICOS_PER_SEC: u32 = 12;
    #[unit_const = FEMTOSECOND, from_units = from_femtos, as_units = as_femtos, units = femtos, suffix = "fs"]
    const LOG10_FEMTOS_PER_SEC: u32 = 15;
    #[unit_const = ATTOSECOND, from_units = from_attos, as_units = as_attos, units = attos, suffix = "as"]
    const LOG10_ATTOS_PER_SEC: u32 = 18;
}

impl SimDuration {
    const fn from_units_helper<const UNITS_PER_SEC: u32>(units: u128) -> Self {
        let Some(attos) =
            units.checked_mul(const { 10u128.pow(Self::LOG10_ATTOS_PER_SEC - UNITS_PER_SEC) })
        else {
            panic!("duration too big");
        };
        Self { attos }
    }
    pub const ZERO: SimDuration = SimDuration::from_secs(0);
    pub const fn from_parts(parts: SimDurationParts) -> Self {
        match Self::from_parts_checked(parts) {
            Some(v) => v,
            None => panic!("duration too big"),
        }
    }
    pub const fn abs_diff(self, other: Self) -> Self {
        Self {
            attos: self.attos.abs_diff(other.attos),
        }
    }
    pub const fn checked_add(self, rhs: Self) -> Option<Self> {
        let Some(attos) = self.attos.checked_add(rhs.attos) else {
            return None;
        };
        Some(Self { attos })
    }
    pub const fn checked_sub(self, rhs: Self) -> Option<Self> {
        let Some(attos) = self.attos.checked_sub(rhs.attos) else {
            return None;
        };
        Some(Self { attos })
    }
    pub const fn is_zero(self) -> bool {
        self.attos == 0
    }
    pub const fn saturating_add(self, rhs: Self) -> Self {
        Self {
            attos: self.attos.saturating_add(rhs.attos),
        }
    }
    pub const fn saturating_sub(self, rhs: Self) -> Self {
        Self {
            attos: self.attos.saturating_sub(rhs.attos),
        }
    }
    pub const fn checked_ilog10(self) -> Option<i32> {
        let Some(ilog10_attos) = self.attos.checked_ilog10() else {
            return None;
        };
        Some(ilog10_attos as i32 - Self::LOG10_ATTOS_PER_SEC as i32)
    }
    #[track_caller]
    pub const fn ilog10(self) -> i32 {
        let Some(retval) = self.checked_ilog10() else {
            panic!("tried to take the ilog10 of 0");
        };
        retval
    }
    pub const fn checked_pow10(log10: i32, underflow_is_zero: bool) -> Option<Self> {
        let Some(log10) = Self::LOG10_ATTOS_PER_SEC.checked_add_signed(log10) else {
            return if log10 < 0 && underflow_is_zero {
                Some(Self::ZERO)
            } else {
                None
            };
        };
        let Some(attos) = 10u128.checked_pow(log10) else {
            return None;
        };
        Some(Self { attos })
    }
    #[track_caller]
    pub const fn pow10(log10: i32) -> Self {
        let Some(retval) = Self::checked_pow10(log10, true) else {
            panic!("pow10 overflowed");
        };
        retval
    }
    pub const fn is_power_of_ten(self) -> bool {
        const TEN: u128 = 10;
        const NUMBER_OF_POWERS_OF_TEN: usize = {
            let mut n = 0;
            while let Some(_) = TEN.checked_pow(n as u32) {
                n += 1;
            }
            n
        };
        const POWERS_OF_TEN: [u128; NUMBER_OF_POWERS_OF_TEN] = {
            let mut retval = [0; NUMBER_OF_POWERS_OF_TEN];
            let mut i = 0;
            while i < NUMBER_OF_POWERS_OF_TEN {
                retval[i] = TEN.pow(i as u32);
                i += 1;
            }
            retval
        };
        let mut i = 0;
        while i < NUMBER_OF_POWERS_OF_TEN {
            if self.attos == POWERS_OF_TEN[i] {
                return true;
            }
            i += 1;
        }
        false
    }
}

impl From<Duration> for SimDuration {
    fn from(duration: Duration) -> Self {
        Self::from_nanos(duration.as_nanos())
    }
}
