// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    array::{Array, ArrayType},
    bundle::{Bundle, BundleType},
    clock::Clock,
    enum_::{Enum, EnumType},
    expr::{CastBitsTo, Expr, ToExpr},
    int::{Bool, IntType, KnownSize, SInt, SIntType, SIntValue, Size, UInt, UIntType, UIntValue},
    reset::{AsyncReset, Reset, SyncReset},
    ty::{CanonicalType, StaticType, Type},
    util::{
        alternating_cell::{AlternatingCell, AlternatingCellMethods},
        ConstUsize,
    },
};
use bitvec::{slice::BitSlice, vec::BitVec};
use serde::{Deserialize, Deserializer, Serialize, Serializer};
use std::{
    fmt,
    ops::{Deref, DerefMut},
    sync::Arc,
};

#[derive(Copy, Clone, Eq, PartialEq)]
enum ValidFlags {
    BothValid = 0,
    OnlyValueValid = 1,
    OnlyBitsValid = 2,
}

#[derive(Clone)]
struct SimValueInner<T: Type> {
    value: T::SimValue,
    bits: UIntValue,
    valid_flags: ValidFlags,
    ty: T,
}

impl<T: Type> SimValueInner<T> {
    fn fill_bits(&mut self) {
        match self.valid_flags {
            ValidFlags::BothValid | ValidFlags::OnlyBitsValid => {}
            ValidFlags::OnlyValueValid => {
                self.ty.sim_value_to_bits(&self.value, self.bits.bits_mut());
                self.valid_flags = ValidFlags::BothValid;
            }
        }
    }
    fn into_bits(mut self) -> UIntValue {
        self.fill_bits();
        self.bits
    }
    fn bits_mut(&mut self) -> &mut UIntValue {
        self.fill_bits();
        self.valid_flags = ValidFlags::OnlyBitsValid;
        &mut self.bits
    }
    fn fill_value(&mut self) {
        match self.valid_flags {
            ValidFlags::BothValid | ValidFlags::OnlyValueValid => {}
            ValidFlags::OnlyBitsValid => {
                self.ty
                    .sim_value_clone_from_bits(&mut self.value, self.bits.bits());
                self.valid_flags = ValidFlags::BothValid;
            }
        }
    }
    fn into_value(mut self) -> T::SimValue {
        self.fill_value();
        self.value
    }
    fn value_mut(&mut self) -> &mut T::SimValue {
        self.fill_value();
        self.valid_flags = ValidFlags::OnlyValueValid;
        &mut self.value
    }
}

impl<T: Type> AlternatingCellMethods for SimValueInner<T> {
    fn unique_to_shared(&mut self) {
        match self.valid_flags {
            ValidFlags::BothValid => return,
            ValidFlags::OnlyValueValid => {
                self.ty.sim_value_to_bits(&self.value, self.bits.bits_mut())
            }
            ValidFlags::OnlyBitsValid => self
                .ty
                .sim_value_clone_from_bits(&mut self.value, self.bits.bits()),
        }
        self.valid_flags = ValidFlags::BothValid;
    }

    fn shared_to_unique(&mut self) {}
}

#[derive(Serialize, Deserialize)]
#[serde(rename = "SimValue")]
#[serde(bound(
    serialize = "T: Type<SimValue: Serialize> + Serialize",
    deserialize = "T: Type<SimValue: Deserialize<'de>> + Deserialize<'de>"
))]
struct SerdeSimValue<'a, T: Type> {
    ty: T,
    value: std::borrow::Cow<'a, T::SimValue>,
}

impl<T: Type<SimValue: Serialize> + Serialize> Serialize for SimValue<T> {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        SerdeSimValue {
            ty: SimValue::ty(self),
            value: std::borrow::Cow::Borrowed(&*self),
        }
        .serialize(serializer)
    }
}

impl<'de, T: Type<SimValue: Deserialize<'de>> + Deserialize<'de>> Deserialize<'de> for SimValue<T> {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'de>,
    {
        let SerdeSimValue { ty, value } = SerdeSimValue::<T>::deserialize(deserializer)?;
        Ok(SimValue::from_value(ty, value.into_owned()))
    }
}

pub struct SimValue<T: Type> {
    inner: AlternatingCell<SimValueInner<T>>,
}

impl<T: Type + Clone> Clone for SimValue<T> {
    fn clone(&self) -> Self {
        Self {
            inner: AlternatingCell::new_unique(self.inner.share().clone()),
        }
    }
}

impl<T: Type> SimValue<T> {
    #[track_caller]
    pub fn from_bits(ty: T, bits: UIntValue) -> Self {
        assert_eq!(ty.canonical().bit_width(), bits.width());
        let inner = SimValueInner {
            value: ty.sim_value_from_bits(bits.bits()),
            bits,
            valid_flags: ValidFlags::BothValid,
            ty,
        };
        Self {
            inner: AlternatingCell::new_shared(inner),
        }
    }
    #[track_caller]
    pub fn from_bitslice(ty: T, bits: &BitSlice) -> Self {
        Self::from_bits(ty, UIntValue::new(Arc::new(bits.to_bitvec())))
    }
    pub fn from_value(ty: T, value: T::SimValue) -> Self {
        let inner = SimValueInner {
            bits: UIntValue::new_dyn(Arc::new(BitVec::repeat(false, ty.canonical().bit_width()))),
            value,
            valid_flags: ValidFlags::OnlyValueValid,
            ty,
        };
        Self {
            inner: AlternatingCell::new_unique(inner),
        }
    }
    pub fn ty(this: &Self) -> T {
        this.inner.share().ty
    }
    pub fn into_bits(this: Self) -> UIntValue {
        this.inner.into_inner().into_bits()
    }
    pub fn into_ty_and_bits(this: Self) -> (T, UIntValue) {
        let inner = this.inner.into_inner();
        (inner.ty, inner.into_bits())
    }
    pub fn bits(this: &Self) -> &UIntValue {
        &this.inner.share().bits
    }
    pub fn bits_mut(this: &mut Self) -> &mut UIntValue {
        this.inner.unique().bits_mut()
    }
    pub fn into_value(this: Self) -> T::SimValue {
        this.inner.into_inner().into_value()
    }
    pub fn value(this: &Self) -> &T::SimValue {
        &this.inner.share().value
    }
    pub fn value_mut(this: &mut Self) -> &mut T::SimValue {
        this.inner.unique().value_mut()
    }
    #[track_caller]
    pub fn from_canonical(v: SimValue<CanonicalType>) -> Self {
        let (ty, bits) = SimValue::into_ty_and_bits(v);
        Self::from_bits(T::from_canonical(ty), bits)
    }
    pub fn into_canonical(this: Self) -> SimValue<CanonicalType> {
        let (ty, bits) = Self::into_ty_and_bits(this);
        SimValue::from_bits(ty.canonical(), bits)
    }
    pub fn canonical(this: &Self) -> SimValue<CanonicalType> {
        SimValue::from_bits(Self::ty(this).canonical(), Self::bits(this).clone())
    }
    #[track_caller]
    pub fn from_dyn_int(v: SimValue<T::Dyn>) -> Self
    where
        T: IntType,
    {
        let (ty, bits) = SimValue::into_ty_and_bits(v);
        SimValue::from_bits(T::from_dyn_int(ty), bits)
    }
    pub fn into_dyn_int(this: Self) -> SimValue<T::Dyn>
    where
        T: IntType,
    {
        let (ty, bits) = Self::into_ty_and_bits(this);
        SimValue::from_bits(ty.as_dyn_int(), bits)
    }
    pub fn to_dyn_int(this: &Self) -> SimValue<T::Dyn>
    where
        T: IntType,
    {
        SimValue::from_bits(Self::ty(this).as_dyn_int(), Self::bits(&this).clone())
    }
    #[track_caller]
    pub fn from_bundle(v: SimValue<Bundle>) -> Self
    where
        T: BundleType,
    {
        let (ty, bits) = SimValue::into_ty_and_bits(v);
        SimValue::from_bits(T::from_canonical(CanonicalType::Bundle(ty)), bits)
    }
    pub fn into_bundle(this: Self) -> SimValue<Bundle>
    where
        T: BundleType,
    {
        let (ty, bits) = Self::into_ty_and_bits(this);
        SimValue::from_bits(Bundle::from_canonical(ty.canonical()), bits)
    }
    pub fn to_bundle(this: &Self) -> SimValue<Bundle>
    where
        T: BundleType,
    {
        SimValue::from_bits(
            Bundle::from_canonical(Self::ty(this).canonical()),
            Self::bits(&this).clone(),
        )
    }
    #[track_caller]
    pub fn from_enum(v: SimValue<Enum>) -> Self
    where
        T: EnumType,
    {
        let (ty, bits) = SimValue::into_ty_and_bits(v);
        SimValue::from_bits(T::from_canonical(CanonicalType::Enum(ty)), bits)
    }
    pub fn into_enum(this: Self) -> SimValue<Enum>
    where
        T: EnumType,
    {
        let (ty, bits) = Self::into_ty_and_bits(this);
        SimValue::from_bits(Enum::from_canonical(ty.canonical()), bits)
    }
    pub fn to_enum(this: &Self) -> SimValue<Enum>
    where
        T: EnumType,
    {
        SimValue::from_bits(
            Enum::from_canonical(Self::ty(this).canonical()),
            Self::bits(&this).clone(),
        )
    }
}

impl<T: Type> Deref for SimValue<T> {
    type Target = T::SimValue;

    fn deref(&self) -> &Self::Target {
        Self::value(self)
    }
}

impl<T: Type> DerefMut for SimValue<T> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        Self::value_mut(self)
    }
}

impl<T: Type> fmt::Debug for SimValue<T> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let inner = self.inner.share();
        f.debug_struct("SimValue")
            .field("ty", &inner.ty)
            .field("value", &inner.value)
            .finish()
    }
}

impl<T: Type> ToExpr for SimValue<T> {
    type Type = T;

    #[track_caller]
    fn to_expr(&self) -> Expr<Self::Type> {
        let inner = self.inner.share();
        inner.bits.cast_bits_to(inner.ty)
    }
}

pub trait SimValuePartialEq<T: Type = Self>: Type {
    #[track_caller]
    fn sim_value_eq(this: &SimValue<Self>, other: &SimValue<T>) -> bool;
}

impl<T: SimValuePartialEq<U>, U: Type> PartialEq<SimValue<U>> for SimValue<T> {
    #[track_caller]
    fn eq(&self, other: &SimValue<U>) -> bool {
        T::sim_value_eq(self, other)
    }
}

impl<Width: Size> SimValuePartialEq<Self> for UIntType<Width> {
    fn sim_value_eq(this: &SimValue<Self>, other: &SimValue<Self>) -> bool {
        **this == **other
    }
}

impl<Width: Size> SimValuePartialEq<Self> for SIntType<Width> {
    fn sim_value_eq(this: &SimValue<Self>, other: &SimValue<Self>) -> bool {
        **this == **other
    }
}

impl SimValuePartialEq<Bool> for Bool {
    fn sim_value_eq(this: &SimValue<Self>, other: &SimValue<Bool>) -> bool {
        **this == **other
    }
}

pub trait ToSimValue: ToSimValueWithType<<Self as ToSimValue>::Type> {
    type Type: Type;

    #[track_caller]
    fn to_sim_value(&self) -> SimValue<Self::Type>;
    #[track_caller]
    fn into_sim_value(self) -> SimValue<Self::Type>
    where
        Self: Sized,
    {
        self.to_sim_value()
    }
    #[track_caller]
    fn arc_into_sim_value(self: Arc<Self>) -> SimValue<Self::Type> {
        self.to_sim_value()
    }
    #[track_caller]
    fn arc_to_sim_value(self: &Arc<Self>) -> SimValue<Self::Type> {
        self.to_sim_value()
    }
}

pub trait ToSimValueWithType<T: Type> {
    #[track_caller]
    fn to_sim_value_with_type(&self, ty: T) -> SimValue<T>;
    #[track_caller]
    fn into_sim_value_with_type(self, ty: T) -> SimValue<T>
    where
        Self: Sized,
    {
        self.to_sim_value_with_type(ty)
    }
    #[track_caller]
    fn arc_into_sim_value_with_type(self: Arc<Self>, ty: T) -> SimValue<T> {
        self.to_sim_value_with_type(ty)
    }
    #[track_caller]
    fn arc_to_sim_value_with_type(self: &Arc<Self>, ty: T) -> SimValue<T> {
        self.to_sim_value_with_type(ty)
    }
}

macro_rules! forward_to_sim_value_with_type {
    ([$($generics:tt)*] $ty:ty) => {
        impl<$($generics)*> ToSimValueWithType<<Self as ToSimValue>::Type> for $ty {
            fn to_sim_value_with_type(&self, ty: <Self as ToSimValue>::Type) -> SimValue<<Self as ToSimValue>::Type> {
                let retval = Self::to_sim_value(self);
                assert_eq!(SimValue::ty(&retval), ty);
                retval
            }
            #[track_caller]
            fn into_sim_value_with_type(self, ty: <Self as ToSimValue>::Type) -> SimValue<<Self as ToSimValue>::Type>
            where
                Self: Sized,
            {
                let retval = Self::into_sim_value(self);
                assert_eq!(SimValue::ty(&retval), ty);
                retval
            }
            #[track_caller]
            fn arc_into_sim_value_with_type(self: Arc<Self>, ty: <Self as ToSimValue>::Type) -> SimValue<<Self as ToSimValue>::Type> {
                let retval = Self::arc_into_sim_value(self);
                assert_eq!(SimValue::ty(&retval), ty);
                retval
            }
            #[track_caller]
            fn arc_to_sim_value_with_type(self: &Arc<Self>, ty: <Self as ToSimValue>::Type) -> SimValue<<Self as ToSimValue>::Type> {
                let retval = Self::arc_to_sim_value(self);
                assert_eq!(SimValue::ty(&retval), ty);
                retval
            }
        }
    };
}

impl<T: Type> ToSimValue for SimValue<T> {
    type Type = T;
    fn to_sim_value(&self) -> SimValue<Self::Type> {
        self.clone()
    }

    fn into_sim_value(self) -> SimValue<Self::Type> {
        self
    }
}

forward_to_sim_value_with_type!([T: Type] SimValue<T>);

impl<T: Type> ToSimValueWithType<T> for BitVec {
    #[track_caller]
    fn to_sim_value_with_type(&self, ty: T) -> SimValue<T> {
        self.clone().into_sim_value_with_type(ty)
    }

    #[track_caller]
    fn into_sim_value_with_type(self, ty: T) -> SimValue<T> {
        Arc::new(self).arc_into_sim_value_with_type(ty)
    }

    #[track_caller]
    fn arc_into_sim_value_with_type(self: Arc<Self>, ty: T) -> SimValue<T> {
        SimValue::from_bits(ty, UIntValue::new_dyn(self))
    }

    #[track_caller]
    fn arc_to_sim_value_with_type(self: &Arc<Self>, ty: T) -> SimValue<T> {
        SimValue::from_bits(ty, UIntValue::new_dyn(self.clone()))
    }
}

impl<T: Type> ToSimValueWithType<T> for bitvec::boxed::BitBox {
    #[track_caller]
    fn to_sim_value_with_type(&self, ty: T) -> SimValue<T> {
        self.clone().into_sim_value_with_type(ty)
    }

    #[track_caller]
    fn into_sim_value_with_type(self, ty: T) -> SimValue<T> {
        self.into_bitvec().into_sim_value_with_type(ty)
    }
}

impl<T: Type> ToSimValueWithType<T> for BitSlice {
    #[track_caller]
    fn to_sim_value_with_type(&self, ty: T) -> SimValue<T> {
        self.to_bitvec().into_sim_value_with_type(ty)
    }
}

impl<This: ?Sized + ToSimValue> ToSimValue for &'_ This {
    type Type = This::Type;

    fn to_sim_value(&self) -> SimValue<Self::Type> {
        This::to_sim_value(self)
    }
}

impl<This: ?Sized + ToSimValueWithType<T>, T: Type> ToSimValueWithType<T> for &'_ This {
    fn to_sim_value_with_type(&self, ty: T) -> SimValue<T> {
        This::to_sim_value_with_type(self, ty)
    }
}

impl<This: ?Sized + ToSimValue> ToSimValue for &'_ mut This {
    type Type = This::Type;

    fn to_sim_value(&self) -> SimValue<Self::Type> {
        This::to_sim_value(self)
    }
}

impl<This: ?Sized + ToSimValueWithType<T>, T: Type> ToSimValueWithType<T> for &'_ mut This {
    fn to_sim_value_with_type(&self, ty: T) -> SimValue<T> {
        This::to_sim_value_with_type(self, ty)
    }
}

impl<This: ?Sized + ToSimValue> ToSimValue for Arc<This> {
    type Type = This::Type;

    fn to_sim_value(&self) -> SimValue<Self::Type> {
        This::arc_to_sim_value(self)
    }
    fn into_sim_value(self) -> SimValue<Self::Type> {
        This::arc_into_sim_value(self)
    }
}

impl<This: ?Sized + ToSimValueWithType<T>, T: Type> ToSimValueWithType<T> for Arc<This> {
    fn to_sim_value_with_type(&self, ty: T) -> SimValue<T> {
        This::arc_to_sim_value_with_type(self, ty)
    }
    fn into_sim_value_with_type(self, ty: T) -> SimValue<T> {
        This::arc_into_sim_value_with_type(self, ty)
    }
}

impl<This: ?Sized + ToSimValue + Send + Sync + 'static> ToSimValue
    for crate::intern::Interned<This>
{
    type Type = This::Type;
    fn to_sim_value(&self) -> SimValue<Self::Type> {
        This::to_sim_value(self)
    }
}

impl<This: ?Sized + ToSimValueWithType<T> + Send + Sync + 'static, T: Type> ToSimValueWithType<T>
    for crate::intern::Interned<This>
{
    fn to_sim_value_with_type(&self, ty: T) -> SimValue<T> {
        This::to_sim_value_with_type(self, ty)
    }
}

impl<This: ToSimValue> ToSimValue for Box<This> {
    type Type = This::Type;

    fn to_sim_value(&self) -> SimValue<Self::Type> {
        This::to_sim_value(self)
    }
    fn into_sim_value(self) -> SimValue<Self::Type> {
        This::into_sim_value(*self)
    }
}

impl<This: ToSimValueWithType<T>, T: Type> ToSimValueWithType<T> for Box<This> {
    fn to_sim_value_with_type(&self, ty: T) -> SimValue<T> {
        This::to_sim_value_with_type(self, ty)
    }
    fn into_sim_value_with_type(self, ty: T) -> SimValue<T> {
        This::into_sim_value_with_type(*self, ty)
    }
}

impl<T: Type, Len: Size> SimValue<ArrayType<T, Len>> {
    #[track_caller]
    pub fn from_array_elements<I: IntoIterator<Item: ToSimValueWithType<T>>>(
        ty: ArrayType<T, Len>,
        elements: I,
    ) -> Self {
        let element_ty = ty.element();
        let elements = Vec::from_iter(
            elements
                .into_iter()
                .map(|element| element.into_sim_value_with_type(element_ty)),
        );
        assert_eq!(elements.len(), ty.len());
        SimValue::from_value(ty, elements.try_into().ok().expect("already checked len"))
    }
}

impl<Element: ToSimValueWithType<T>, T: Type> ToSimValueWithType<Array<T>> for [Element] {
    #[track_caller]
    fn to_sim_value_with_type(&self, ty: Array<T>) -> SimValue<Array<T>> {
        SimValue::from_array_elements(ty, self)
    }
}

impl<Element: ToSimValue<Type: StaticType>> ToSimValue for [Element] {
    type Type = Array<Element::Type>;

    #[track_caller]
    fn to_sim_value(&self) -> SimValue<Self::Type> {
        SimValue::from_array_elements(ArrayType::new_dyn(StaticType::TYPE, self.len()), self)
    }
}

impl<Element: ToSimValueWithType<CanonicalType>> ToSimValueWithType<CanonicalType> for [Element] {
    #[track_caller]
    fn to_sim_value_with_type(&self, ty: CanonicalType) -> SimValue<CanonicalType> {
        SimValue::into_canonical(SimValue::from_array_elements(
            <Array>::from_canonical(ty),
            self,
        ))
    }
}

impl<Element: ToSimValueWithType<T>, T: Type, const N: usize> ToSimValueWithType<Array<T, N>>
    for [Element; N]
where
    ConstUsize<N>: KnownSize,
{
    #[track_caller]
    fn to_sim_value_with_type(&self, ty: Array<T, N>) -> SimValue<Array<T, N>> {
        SimValue::from_array_elements(ty, self)
    }
    #[track_caller]
    fn into_sim_value_with_type(self, ty: Array<T, N>) -> SimValue<Array<T, N>> {
        SimValue::from_array_elements(ty, self)
    }
}

impl<Element: ToSimValue<Type: StaticType>, const N: usize> ToSimValue for [Element; N]
where
    ConstUsize<N>: KnownSize,
{
    type Type = Array<Element::Type, N>;

    fn to_sim_value(&self) -> SimValue<Self::Type> {
        SimValue::from_array_elements(StaticType::TYPE, self)
    }

    fn into_sim_value(self) -> SimValue<Self::Type> {
        SimValue::from_array_elements(StaticType::TYPE, self)
    }
}

impl<Element: ToSimValueWithType<T>, T: Type, const N: usize> ToSimValueWithType<Array<T>>
    for [Element; N]
{
    #[track_caller]
    fn to_sim_value_with_type(&self, ty: Array<T>) -> SimValue<Array<T>> {
        SimValue::from_array_elements(ty, self)
    }
    #[track_caller]
    fn into_sim_value_with_type(self, ty: Array<T>) -> SimValue<Array<T>> {
        SimValue::from_array_elements(ty, self)
    }
}

impl<Element: ToSimValueWithType<CanonicalType>, const N: usize> ToSimValueWithType<CanonicalType>
    for [Element; N]
{
    #[track_caller]
    fn to_sim_value_with_type(&self, ty: CanonicalType) -> SimValue<CanonicalType> {
        SimValue::into_canonical(SimValue::from_array_elements(
            <Array>::from_canonical(ty),
            self,
        ))
    }
    #[track_caller]
    fn into_sim_value_with_type(self, ty: CanonicalType) -> SimValue<CanonicalType> {
        SimValue::into_canonical(SimValue::from_array_elements(
            <Array>::from_canonical(ty),
            self,
        ))
    }
}

impl<Element: ToSimValueWithType<T>, T: Type> ToSimValueWithType<Array<T>> for Vec<Element> {
    #[track_caller]
    fn to_sim_value_with_type(&self, ty: Array<T>) -> SimValue<Array<T>> {
        SimValue::from_array_elements(ty, self)
    }
    #[track_caller]
    fn into_sim_value_with_type(self, ty: Array<T>) -> SimValue<Array<T>> {
        SimValue::from_array_elements(ty, self)
    }
}

impl<Element: ToSimValue<Type: StaticType>> ToSimValue for Vec<Element> {
    type Type = Array<Element::Type>;

    fn to_sim_value(&self) -> SimValue<Self::Type> {
        SimValue::from_array_elements(ArrayType::new_dyn(StaticType::TYPE, self.len()), self)
    }

    fn into_sim_value(self) -> SimValue<Self::Type> {
        SimValue::from_array_elements(ArrayType::new_dyn(StaticType::TYPE, self.len()), self)
    }
}

impl<Element: ToSimValueWithType<CanonicalType>> ToSimValueWithType<CanonicalType>
    for Vec<Element>
{
    #[track_caller]
    fn to_sim_value_with_type(&self, ty: CanonicalType) -> SimValue<CanonicalType> {
        SimValue::into_canonical(SimValue::from_array_elements(
            <Array>::from_canonical(ty),
            self,
        ))
    }
    #[track_caller]
    fn into_sim_value_with_type(self, ty: CanonicalType) -> SimValue<CanonicalType> {
        SimValue::into_canonical(SimValue::from_array_elements(
            <Array>::from_canonical(ty),
            self,
        ))
    }
}

impl<Element: ToSimValueWithType<T>, T: Type> ToSimValueWithType<Array<T>> for Box<[Element]> {
    #[track_caller]
    fn to_sim_value_with_type(&self, ty: Array<T>) -> SimValue<Array<T>> {
        SimValue::from_array_elements(ty, self)
    }
    #[track_caller]
    fn into_sim_value_with_type(self, ty: Array<T>) -> SimValue<Array<T>> {
        SimValue::from_array_elements(ty, self)
    }
}

impl<Element: ToSimValue<Type: StaticType>> ToSimValue for Box<[Element]> {
    type Type = Array<Element::Type>;

    fn to_sim_value(&self) -> SimValue<Self::Type> {
        SimValue::from_array_elements(ArrayType::new_dyn(StaticType::TYPE, self.len()), self)
    }

    fn into_sim_value(self) -> SimValue<Self::Type> {
        SimValue::from_array_elements(ArrayType::new_dyn(StaticType::TYPE, self.len()), self)
    }
}

impl<Element: ToSimValueWithType<CanonicalType>> ToSimValueWithType<CanonicalType>
    for Box<[Element]>
{
    #[track_caller]
    fn to_sim_value_with_type(&self, ty: CanonicalType) -> SimValue<CanonicalType> {
        SimValue::into_canonical(SimValue::from_array_elements(
            <Array>::from_canonical(ty),
            self,
        ))
    }
    #[track_caller]
    fn into_sim_value_with_type(self, ty: CanonicalType) -> SimValue<CanonicalType> {
        SimValue::into_canonical(SimValue::from_array_elements(
            <Array>::from_canonical(ty),
            self,
        ))
    }
}

impl<T: Type> ToSimValue for Expr<T> {
    type Type = T;
    #[track_caller]
    fn to_sim_value(&self) -> SimValue<Self::Type> {
        SimValue::from_bitslice(
            Expr::ty(*self),
            &crate::expr::ToLiteralBits::to_literal_bits(self)
                .expect("must be a literal expression"),
        )
    }
}

forward_to_sim_value_with_type!([T: Type] Expr<T>);

macro_rules! impl_to_sim_value_for_bool_like {
    ($ty:ident) => {
        impl ToSimValueWithType<$ty> for bool {
            fn to_sim_value_with_type(&self, ty: $ty) -> SimValue<$ty> {
                SimValue::from_value(ty, *self)
            }
        }
    };
}

impl ToSimValue for bool {
    type Type = Bool;

    fn to_sim_value(&self) -> SimValue<Self::Type> {
        SimValue::from_value(Bool, *self)
    }
}

impl_to_sim_value_for_bool_like!(Bool);
impl_to_sim_value_for_bool_like!(AsyncReset);
impl_to_sim_value_for_bool_like!(SyncReset);
impl_to_sim_value_for_bool_like!(Reset);
impl_to_sim_value_for_bool_like!(Clock);

impl ToSimValueWithType<CanonicalType> for bool {
    #[track_caller]
    fn to_sim_value_with_type(&self, ty: CanonicalType) -> SimValue<CanonicalType> {
        match ty {
            CanonicalType::UInt(_)
            | CanonicalType::SInt(_)
            | CanonicalType::Array(_)
            | CanonicalType::Enum(_)
            | CanonicalType::Bundle(_)
            | CanonicalType::PhantomConst(_) => {
                panic!("can't create SimValue from bool: expected value of type: {ty:?}");
            }
            CanonicalType::Bool(_)
            | CanonicalType::AsyncReset(_)
            | CanonicalType::SyncReset(_)
            | CanonicalType::Reset(_)
            | CanonicalType::Clock(_) => {
                SimValue::from_bits(ty, UIntValue::new(Arc::new(BitVec::repeat(*self, 1))))
            }
        }
    }
}

macro_rules! impl_to_sim_value_for_primitive_int {
    ($prim:ident) => {
        impl ToSimValue for $prim {
            type Type = <$prim as ToExpr>::Type;

            #[track_caller]
            fn to_sim_value(
                &self,
            ) -> SimValue<Self::Type> {
                SimValue::from_value(StaticType::TYPE, (*self).into())
            }
        }

        forward_to_sim_value_with_type!([] $prim);

        impl ToSimValueWithType<<<$prim as ToExpr>::Type as IntType>::Dyn> for $prim {
            #[track_caller]
            fn to_sim_value_with_type(
                &self,
                ty: <<$prim as ToExpr>::Type as IntType>::Dyn,
            ) -> SimValue<<<$prim as ToExpr>::Type as IntType>::Dyn> {
                SimValue::from_value(
                    ty,
                    <<$prim as ToExpr>::Type as Type>::SimValue::from(*self).as_dyn_int(),
                )
            }
        }

        impl ToSimValueWithType<CanonicalType> for $prim {
            #[track_caller]
            fn to_sim_value_with_type(&self, ty: CanonicalType) -> SimValue<CanonicalType> {
                let ty: <<$prim as ToExpr>::Type as IntType>::Dyn = Type::from_canonical(ty);
                SimValue::into_canonical(self.to_sim_value_with_type(ty))
            }
        }
    };
}

impl_to_sim_value_for_primitive_int!(u8);
impl_to_sim_value_for_primitive_int!(u16);
impl_to_sim_value_for_primitive_int!(u32);
impl_to_sim_value_for_primitive_int!(u64);
impl_to_sim_value_for_primitive_int!(u128);
impl_to_sim_value_for_primitive_int!(usize);
impl_to_sim_value_for_primitive_int!(i8);
impl_to_sim_value_for_primitive_int!(i16);
impl_to_sim_value_for_primitive_int!(i32);
impl_to_sim_value_for_primitive_int!(i64);
impl_to_sim_value_for_primitive_int!(i128);
impl_to_sim_value_for_primitive_int!(isize);

macro_rules! impl_to_sim_value_for_int_value {
    ($IntValue:ident, $Int:ident, $IntType:ident) => {
        impl<Width: Size> ToSimValue for $IntValue<Width> {
            type Type = $IntType<Width>;

            fn to_sim_value(&self) -> SimValue<Self::Type> {
                SimValue::from_value(self.ty(), self.clone())
            }

            fn into_sim_value(self) -> SimValue<Self::Type> {
                SimValue::from_value(self.ty(), self)
            }
        }

        impl<Width: Size> ToSimValueWithType<$IntType<Width>> for $IntValue<Width> {
            fn to_sim_value_with_type(&self, ty: $IntType<Width>) -> SimValue<$IntType<Width>> {
                SimValue::from_value(ty, self.clone())
            }

            fn into_sim_value_with_type(self, ty: $IntType<Width>) -> SimValue<$IntType<Width>> {
                SimValue::from_value(ty, self)
            }
        }

        impl<Width: KnownSize> ToSimValueWithType<$Int> for $IntValue<Width> {
            fn to_sim_value_with_type(&self, ty: $Int) -> SimValue<$Int> {
                self.bits().to_sim_value_with_type(ty)
            }

            fn into_sim_value_with_type(self, ty: $Int) -> SimValue<$Int> {
                self.into_bits().into_sim_value_with_type(ty)
            }
        }

        impl<Width: Size> ToSimValueWithType<CanonicalType> for $IntValue<Width> {
            #[track_caller]
            fn to_sim_value_with_type(&self, ty: CanonicalType) -> SimValue<CanonicalType> {
                SimValue::into_canonical(
                    self.to_sim_value_with_type($IntType::<Width>::from_canonical(ty)),
                )
            }

            #[track_caller]
            fn into_sim_value_with_type(self, ty: CanonicalType) -> SimValue<CanonicalType> {
                SimValue::into_canonical(
                    self.into_sim_value_with_type($IntType::<Width>::from_canonical(ty)),
                )
            }
        }
    };
}

impl_to_sim_value_for_int_value!(UIntValue, UInt, UIntType);
impl_to_sim_value_for_int_value!(SIntValue, SInt, SIntType);
