// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    array::Array,
    bundle::{Bundle, BundleType},
    clock::Clock,
    enum_::{Enum, EnumType},
    int::{Bool, SInt, UInt},
    intern::Interned,
    phantom_const::{PhantomConstCanonicalValue, PhantomConstValue},
    prelude::PhantomConst,
    reset::{AsyncReset, Reset, SyncReset},
    ty::{BaseType, CanonicalType},
};
use serde::{Deserialize, Deserializer, Serialize, Serializer};

pub(crate) struct SerdePhantomConst<T>(pub T);

impl<T: ?Sized + PhantomConstValue> Serialize for SerdePhantomConst<Interned<T>> {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        self.0.serialize(serializer)
    }
}

impl<'de, T: ?Sized + PhantomConstValue> Deserialize<'de> for SerdePhantomConst<Interned<T>> {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'de>,
    {
        T::deserialize_value(deserializer).map(Self)
    }
}

#[derive(Serialize, Deserialize)]
#[serde(rename = "CanonicalType")]
pub(crate) enum SerdeCanonicalType<
    ArrayElement = CanonicalType,
    ThePhantomConst = SerdePhantomConst<Interned<PhantomConstCanonicalValue>>,
> {
    UInt {
        width: usize,
    },
    SInt {
        width: usize,
    },
    Bool,
    Array {
        element: ArrayElement,
        len: usize,
    },
    Enum {
        variants: Interned<[crate::enum_::EnumVariant]>,
    },
    Bundle {
        fields: Interned<[crate::bundle::BundleField]>,
    },
    AsyncReset,
    SyncReset,
    Reset,
    Clock,
    PhantomConst(ThePhantomConst),
}

impl<ArrayElement, PhantomConstInner> SerdeCanonicalType<ArrayElement, PhantomConstInner> {
    pub(crate) fn as_serde_unexpected_str(&self) -> &'static str {
        match self {
            Self::UInt { .. } => "a UInt",
            Self::SInt { .. } => "a SInt",
            Self::Bool => "a Bool",
            Self::Array { .. } => "an Array",
            Self::Enum { .. } => "an Enum",
            Self::Bundle { .. } => "a Bundle",
            Self::AsyncReset => "an AsyncReset",
            Self::SyncReset => "a SyncReset",
            Self::Reset => "a Reset",
            Self::Clock => "a Clock",
            Self::PhantomConst(_) => "a PhantomConst",
        }
    }
}

impl<T: BaseType> From<T> for SerdeCanonicalType {
    fn from(ty: T) -> Self {
        let ty: CanonicalType = ty.into();
        match ty {
            CanonicalType::UInt(ty) => Self::UInt { width: ty.width() },
            CanonicalType::SInt(ty) => Self::SInt { width: ty.width() },
            CanonicalType::Bool(Bool {}) => Self::Bool,
            CanonicalType::Array(ty) => Self::Array {
                element: ty.element(),
                len: ty.len(),
            },
            CanonicalType::Enum(ty) => Self::Enum {
                variants: ty.variants(),
            },
            CanonicalType::Bundle(ty) => Self::Bundle {
                fields: ty.fields(),
            },
            CanonicalType::AsyncReset(AsyncReset {}) => Self::AsyncReset,
            CanonicalType::SyncReset(SyncReset {}) => Self::SyncReset,
            CanonicalType::Reset(Reset {}) => Self::Reset,
            CanonicalType::Clock(Clock {}) => Self::Clock,
            CanonicalType::PhantomConst(ty) => Self::PhantomConst(SerdePhantomConst(ty.get())),
        }
    }
}

impl From<SerdeCanonicalType> for CanonicalType {
    fn from(ty: SerdeCanonicalType) -> Self {
        match ty {
            SerdeCanonicalType::UInt { width } => Self::UInt(UInt::new(width)),
            SerdeCanonicalType::SInt { width } => Self::SInt(SInt::new(width)),
            SerdeCanonicalType::Bool => Self::Bool(Bool),
            SerdeCanonicalType::Array { element, len } => Self::Array(Array::new(element, len)),
            SerdeCanonicalType::Enum { variants } => Self::Enum(Enum::new(variants)),
            SerdeCanonicalType::Bundle { fields } => Self::Bundle(Bundle::new(fields)),
            SerdeCanonicalType::AsyncReset => Self::AsyncReset(AsyncReset),
            SerdeCanonicalType::SyncReset => Self::SyncReset(SyncReset),
            SerdeCanonicalType::Reset => Self::Reset(Reset),
            SerdeCanonicalType::Clock => Self::Clock(Clock),
            SerdeCanonicalType::PhantomConst(value) => {
                Self::PhantomConst(PhantomConst::new(value.0))
            }
        }
    }
}
