// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::util::DebugAsDisplay;
use std::{
    cell::{Cell, UnsafeCell},
    fmt,
};

pub(crate) trait AlternatingCellMethods {
    fn unique_to_shared(&mut self);
    fn shared_to_unique(&mut self);
}

#[derive(Copy, Clone, Debug)]
enum State {
    Unique,
    Shared,
    Locked,
}

pub(crate) struct AlternatingCell<T: ?Sized> {
    state: Cell<State>,
    value: UnsafeCell<T>,
}

impl<T: ?Sized + fmt::Debug + AlternatingCellMethods> fmt::Debug for AlternatingCell<T> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.debug_tuple("AlternatingCell")
            .field(
                self.try_share()
                    .as_ref()
                    .map(|v| -> &dyn fmt::Debug { v })
                    .unwrap_or(&DebugAsDisplay("<...>")),
            )
            .finish()
    }
}

impl<T: ?Sized> AlternatingCell<T> {
    pub(crate) const fn new_shared(value: T) -> Self
    where
        T: Sized,
    {
        Self {
            state: Cell::new(State::Shared),
            value: UnsafeCell::new(value),
        }
    }
    pub(crate) const fn new_unique(value: T) -> Self
    where
        T: Sized,
    {
        Self {
            state: Cell::new(State::Unique),
            value: UnsafeCell::new(value),
        }
    }
    pub(crate) fn is_unique(&self) -> bool {
        matches!(self.state.get(), State::Unique)
    }
    pub(crate) fn is_shared(&self) -> bool {
        matches!(self.state.get(), State::Shared)
    }
    pub(crate) fn into_inner(self) -> T
    where
        T: Sized,
    {
        self.value.into_inner()
    }
    pub(crate) fn try_share(&self) -> Option<&T>
    where
        T: AlternatingCellMethods,
    {
        match self.state.get() {
            State::Shared => {}
            State::Unique => {
                struct Locked<'a>(&'a Cell<State>);
                impl Drop for Locked<'_> {
                    fn drop(&mut self) {
                        self.0.set(State::Shared);
                    }
                }
                self.state.set(State::Locked);
                let lock = Locked(&self.state);
                // Safety: state is Locked, so nothing else will
                // access value while calling unique_to_shared.
                unsafe { &mut *self.value.get() }.unique_to_shared();
                drop(lock);
            }
            State::Locked => return None,
        }

        // Safety: state is Shared so nothing will create any mutable
        // references until the returned reference's lifetime expires.
        Some(unsafe { &*self.value.get() })
    }
    #[track_caller]
    pub(crate) fn share(&self) -> &T
    where
        T: AlternatingCellMethods,
    {
        let Some(retval) = self.try_share() else {
            panic!("`share` called recursively");
        };
        retval
    }
    pub(crate) fn unique(&mut self) -> &mut T
    where
        T: AlternatingCellMethods,
    {
        match self.state.get() {
            State::Shared => {
                self.state.set(State::Unique);
                self.value.get_mut().shared_to_unique();
            }
            State::Unique => {}
            State::Locked => unreachable!(),
        }
        self.value.get_mut()
    }
}
