// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use serde::{
    de::{DeserializeOwned, Error, Unexpected},
    Deserialize, Deserializer, Serialize, Serializer,
};
use std::{fmt::Debug, hash::Hash};

mod sealed {
    pub trait Sealed {}
}

/// the only implementation is `ConstUsize<Self::VALUE>`
pub trait GenericConstUsize:
    sealed::Sealed
    + Copy
    + Ord
    + Hash
    + Default
    + Debug
    + 'static
    + Send
    + Sync
    + Serialize
    + DeserializeOwned
{
    const VALUE: usize;
}

#[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord, Hash, Default)]
pub struct ConstUsize<const VALUE: usize>;

impl<const VALUE: usize> Debug for ConstUsize<VALUE> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_tuple("ConstUsize").field(&Self::VALUE).finish()
    }
}

impl<const VALUE: usize> sealed::Sealed for ConstUsize<VALUE> {}

impl<const VALUE: usize> GenericConstUsize for ConstUsize<VALUE> {
    const VALUE: usize = VALUE;
}

impl<const VALUE: usize> Serialize for ConstUsize<VALUE> {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        VALUE.serialize(serializer)
    }
}

impl<'de, const VALUE: usize> Deserialize<'de> for ConstUsize<VALUE> {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'de>,
    {
        let value = usize::deserialize(deserializer)?;
        if value == VALUE {
            Ok(ConstUsize)
        } else {
            Err(D::Error::invalid_value(
                Unexpected::Unsigned(value as u64),
                &&*VALUE.to_string(),
            ))
        }
    }
}
