// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use crate::intern::{Intern, Interned};
use bitvec::{bits, order::Lsb0, slice::BitSlice, view::BitView};
use std::{
    cell::Cell,
    fmt::{self, Debug, Write},
    rc::Rc,
    sync::{Arc, OnceLock},
};

/// replacement for Iterator::eq_by which isn't stable yet
pub fn iter_eq_by<L: IntoIterator, R: IntoIterator, F: FnMut(L::Item, R::Item) -> bool>(
    l: L,
    r: R,
    mut f: F,
) -> bool {
    let mut l = l.into_iter();
    let mut r = r.into_iter();
    l.try_for_each(|l| {
        if let Some(r) = r.next() {
            f(l, r).then_some(())
        } else {
            None
        }
    })
    .is_some()
        && r.next().is_none()
}

pub struct DebugAsDisplay<T>(pub T);

impl<T: fmt::Display> fmt::Debug for DebugAsDisplay<T> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.0.fmt(f)
    }
}

pub trait MakeMutSlice {
    type Element: Clone;
    fn make_mut_slice(&mut self) -> &mut [Self::Element];
}

impl<T: Clone> MakeMutSlice for Arc<[T]> {
    type Element = T;

    fn make_mut_slice(&mut self) -> &mut [Self::Element] {
        if Arc::get_mut(self).is_none() {
            *self = Arc::<[T]>::from(&**self);
        }
        Arc::get_mut(self).unwrap()
    }
}

impl<T: Clone> MakeMutSlice for Rc<[T]> {
    type Element = T;

    fn make_mut_slice(&mut self) -> &mut [Self::Element] {
        if Rc::get_mut(self).is_none() {
            *self = Rc::<[T]>::from(&**self);
        }
        Rc::get_mut(self).unwrap()
    }
}

pub struct DebugAsRawString<'a>(pub &'a str);

impl fmt::Debug for DebugAsRawString<'_> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let pounds = self
            .0
            .split_inclusive('"')
            .skip(1)
            .map(|v| v.len() - v.trim_start_matches('#').len())
            .max()
            .map(|v| v + 1)
            .unwrap_or(0);
        f.write_char('r')?;
        for _ in 0..pounds {
            f.write_char('#')?;
        }
        f.write_char('"')?;
        f.write_str(self.0)?;
        f.write_char('"')?;
        for _ in 0..pounds {
            f.write_char('#')?;
        }
        Ok(())
    }
}

pub fn interned_bit(v: bool) -> Interned<BitSlice> {
    static RETVAL: OnceLock<[Interned<BitSlice>; 2]> = OnceLock::new();
    // use bits![V; 1] instead of bits![V] to work around https://github.com/ferrilab/bitvec/issues/271
    RETVAL.get_or_init(|| [bits![0; 1].intern(), bits![1; 1].intern()])[v as usize]
}

#[derive(Copy, Clone)]
pub struct BitSliceWriteWithBase<'a>(pub &'a BitSlice);

impl<'a> Debug for BitSliceWriteWithBase<'a> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{self:#x}")
    }
}

impl BitSliceWriteWithBase<'_> {
    fn fmt_with_base<const BITS_PER_DIGIT: usize, const UPPER_CASE: bool>(
        self,
        f: &mut fmt::Formatter<'_>,
    ) -> fmt::Result {
        let digit_count = self.0.len().div_ceil(BITS_PER_DIGIT).max(1);
        // TODO: optimize
        let mut buf = String::with_capacity(digit_count);
        for digit_index in (0..digit_count).rev() {
            let mut digit = 0;
            let src = self
                .0
                .get(digit_index * BITS_PER_DIGIT..)
                .unwrap_or_default();
            let src = src.get(..BITS_PER_DIGIT).unwrap_or(src);
            digit.view_bits_mut::<Lsb0>()[..src.len()].copy_from_bitslice(src);
            let mut ch = char::from_digit(digit as u32, 1 << BITS_PER_DIGIT).unwrap();
            if UPPER_CASE {
                ch = ch.to_ascii_uppercase();
            }
            buf.push(ch);
        }
        f.pad_integral(
            true,
            match BITS_PER_DIGIT {
                1 => "0b",
                3 => "0o",
                4 => "0x",
                _ => "",
            },
            &buf,
        )
    }
}

impl fmt::Binary for BitSliceWriteWithBase<'_> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.fmt_with_base::<1, false>(f)
    }
}

impl fmt::Octal for BitSliceWriteWithBase<'_> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.fmt_with_base::<3, false>(f)
    }
}

impl fmt::LowerHex for BitSliceWriteWithBase<'_> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.fmt_with_base::<4, false>(f)
    }
}

impl fmt::UpperHex for BitSliceWriteWithBase<'_> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.fmt_with_base::<4, true>(f)
    }
}

#[inline]
#[track_caller]
pub fn get_many_mut<T, const N: usize>(slice: &mut [T], indexes: [usize; N]) -> [&mut T; N] {
    for i in 0..N {
        for j in 0..i {
            assert!(indexes[i] != indexes[j], "duplicate index");
        }
        assert!(indexes[i] < slice.len(), "index out of bounds");
    }
    // Safety: checked that no indexes are duplicates and no indexes are out of bounds
    unsafe {
        let base = slice.as_mut_ptr(); // convert to a raw pointer before loop to avoid aliasing with &mut [T]
        std::array::from_fn(|i| &mut *base.add(indexes[i]))
    }
}

#[derive(Clone, Default)]
pub struct RcWriter(Rc<Cell<Vec<u8>>>);

impl Debug for RcWriter {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.borrow_impl(|buf| {
            f.debug_tuple("RcWriter")
                .field(&DebugAsDisplay(format_args!("b\"{}\"", buf.escape_ascii())))
                .finish()
        })
    }
}

impl RcWriter {
    fn borrow_impl<R>(&self, f: impl FnOnce(&mut Vec<u8>) -> R) -> R {
        let buf = Cell::take(&self.0);
        struct PutBackOnDrop<'a> {
            buf: Vec<u8>,
            this: &'a RcWriter,
        }
        impl Drop for PutBackOnDrop<'_> {
            fn drop(&mut self) {
                self.this.0.set(std::mem::take(&mut self.buf));
            }
        }
        let mut buf = PutBackOnDrop { buf, this: self };
        f(&mut buf.buf)
    }
    pub fn borrow<R>(&mut self, f: impl FnOnce(&mut Vec<u8>) -> R) -> R {
        self.borrow_impl(f)
    }
    pub fn take(&mut self) -> Vec<u8> {
        Cell::take(&self.0)
    }
}

impl std::io::Write for RcWriter {
    fn write(&mut self, buf: &[u8]) -> std::io::Result<usize> {
        self.borrow(|v| v.extend_from_slice(buf));
        Ok(buf.len())
    }

    fn flush(&mut self) -> std::io::Result<()> {
        Ok(())
    }
}
