// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

//! Fayalite Simulation

use crate::{
    bundle::{BundleField, BundleType},
    enum_::{EnumType, EnumVariant},
    expr::{
        ops,
        target::{
            GetTarget, Target, TargetBase, TargetPathArrayElement, TargetPathBundleField,
            TargetPathElement,
        },
        ExprEnum, Flow, ToLiteralBits,
    },
    int::{BoolOrIntType, UIntValue},
    intern::{
        Intern, Interned, InternedCompare, Memoize, PtrEqWithTypeId, SupportsPtrEqWithTypeId,
    },
    memory::PortKind,
    module::{
        transform::deduce_resets::deduce_resets, AnnotatedModuleIO, Block, ExternModuleBody, Id,
        InstantiatedModule, ModuleBody, NameId, NormalModuleBody, ScopedNameId, Stmt, StmtConnect,
        StmtDeclaration, StmtFormal, StmtIf, StmtInstance, StmtMatch, StmtReg, StmtWire,
        TargetInInstantiatedModule,
    },
    prelude::*,
    reset::{ResetType, ResetTypeDispatch},
    sim::{
        interpreter::{
            BreakAction, BreakpointsSet, Insn, InsnField, InsnFieldKind, InsnFieldType,
            InsnOrLabel, Insns, InsnsBuilding, InsnsBuildingDone, InsnsBuildingKind, Label,
            MemoryData, RunResult, SlotDebugData, SmallUInt, State, StatePartArrayIndex,
            StatePartArrayIndexed, StatePartIndex, StatePartIndexRange, StatePartKind,
            StatePartKindBigSlots, StatePartKindMemories, StatePartKindSmallSlots, StatePartLayout,
            StatePartLen, StatePartsValue, TypeArrayIndex, TypeArrayIndexes, TypeIndex,
            TypeIndexRange, TypeLayout, TypeLen, TypeParts,
        },
        time::{SimDuration, SimInstant},
        value::SimValue,
    },
    ty::StaticType,
    util::{BitSliceWriteWithBase, DebugAsDisplay, HashMap, HashSet},
};
use bitvec::{bits, order::Lsb0, slice::BitSlice, vec::BitVec, view::BitView};
use num_bigint::BigInt;
use num_traits::{Signed, Zero};
use petgraph::{
    data::FromElements,
    visit::{
        EdgeRef, GraphBase, IntoEdgeReferences, IntoNeighbors, IntoNeighborsDirected,
        IntoNodeIdentifiers, IntoNodeReferences, NodeRef, VisitMap, Visitable,
    },
};
use std::{
    any::Any,
    borrow::Cow,
    cell::RefCell,
    collections::BTreeSet,
    fmt,
    future::{Future, IntoFuture},
    hash::Hash,
    marker::PhantomData,
    mem,
    ops::IndexMut,
    pin::Pin,
    rc::Rc,
    sync::Arc,
    task::Poll,
};

mod interpreter;
pub mod time;
pub mod value;
pub mod vcd;

#[derive(Debug, PartialEq, Eq, Hash, Clone, Copy)]
enum CondBody {
    IfTrue {
        cond: CompiledValue<Bool>,
    },
    IfFalse {
        cond: CompiledValue<Bool>,
    },
    MatchArm {
        discriminant: StatePartIndex<StatePartKindSmallSlots>,
        variant_index: usize,
    },
}

#[derive(Debug, PartialEq, Eq, Hash, Clone, Copy)]
struct Cond {
    body: CondBody,
    source_location: SourceLocation,
}

#[derive(Debug, PartialEq, Eq, Hash, Clone, Copy)]
struct CompiledBundleField {
    offset: TypeIndex,
    ty: CompiledTypeLayout<CanonicalType>,
}

#[derive(Debug, PartialEq, Eq, Hash, Clone, Copy)]
enum CompiledTypeLayoutBody {
    Scalar,
    Array {
        /// debug names are ignored, use parent's layout instead
        element: Interned<CompiledTypeLayout<CanonicalType>>,
    },
    Bundle {
        /// debug names are ignored, use parent's layout instead
        fields: Interned<[CompiledBundleField]>,
    },
}

#[derive(Debug, PartialEq, Eq, Hash, Clone, Copy)]
struct CompiledTypeLayout<T: Type> {
    ty: T,
    layout: TypeLayout<InsnsBuildingDone>,
    body: CompiledTypeLayoutBody,
}

impl<T: Type> CompiledTypeLayout<T> {
    fn with_prefixed_debug_names(self, prefix: &str) -> Self {
        let Self { ty, layout, body } = self;
        Self {
            ty,
            layout: layout.with_prefixed_debug_names(prefix),
            body,
        }
    }
    fn with_anonymized_debug_info(self) -> Self {
        let Self { ty, layout, body } = self;
        Self {
            ty,
            layout: layout.with_anonymized_debug_info(),
            body,
        }
    }
    fn get(ty: T) -> Self {
        #[derive(Clone, Copy, PartialEq, Eq, Hash, Debug)]
        struct MyMemoize;
        impl Memoize for MyMemoize {
            type Input = CanonicalType;
            type InputOwned = CanonicalType;
            type Output = CompiledTypeLayout<CanonicalType>;

            fn inner(self, input: &Self::Input) -> Self::Output {
                match input {
                    CanonicalType::UInt(_)
                    | CanonicalType::SInt(_)
                    | CanonicalType::Bool(_)
                    | CanonicalType::Enum(_)
                    | CanonicalType::AsyncReset(_)
                    | CanonicalType::SyncReset(_)
                    | CanonicalType::Reset(_)
                    | CanonicalType::Clock(_) => {
                        let mut layout = TypeLayout::empty();
                        let debug_data = SlotDebugData {
                            name: Interned::default(),
                            ty: *input,
                        };
                        layout.big_slots = StatePartLayout::scalar(debug_data, ());
                        CompiledTypeLayout {
                            ty: *input,
                            layout: layout.into(),
                            body: CompiledTypeLayoutBody::Scalar,
                        }
                    }
                    CanonicalType::Array(array) => {
                        let mut layout = TypeLayout::empty();
                        let element = CompiledTypeLayout::get(array.element()).intern_sized();
                        for index in 0..array.len() {
                            layout.allocate(
                                &element
                                    .layout
                                    .with_prefixed_debug_names(&format!("[{index}]")),
                            );
                        }
                        CompiledTypeLayout {
                            ty: *input,
                            layout: layout.into(),
                            body: CompiledTypeLayoutBody::Array { element },
                        }
                    }
                    CanonicalType::PhantomConst(_) => {
                        let unit_layout = CompiledTypeLayout::get(());
                        CompiledTypeLayout {
                            ty: *input,
                            layout: unit_layout.layout,
                            body: unit_layout.body,
                        }
                    }
                    CanonicalType::Bundle(bundle) => {
                        let mut layout = TypeLayout::empty();
                        let fields = bundle
                            .fields()
                            .iter()
                            .map(
                                |BundleField {
                                     name,
                                     flipped: _,
                                     ty,
                                 }| {
                                    let ty = CompiledTypeLayout::get(*ty);
                                    let offset = layout
                                        .allocate(
                                            &ty.layout
                                                .with_prefixed_debug_names(&format!(".{name}")),
                                        )
                                        .start();
                                    CompiledBundleField { offset, ty }
                                },
                            )
                            .collect();
                        CompiledTypeLayout {
                            ty: *input,
                            layout: layout.into(),
                            body: CompiledTypeLayoutBody::Bundle { fields },
                        }
                    }
                }
            }
        }
        let CompiledTypeLayout {
            ty: _,
            layout,
            body,
        } = MyMemoize.get_owned(ty.canonical());
        Self { ty, layout, body }
    }
}

#[derive(Debug, PartialEq, Eq, Hash, Clone, Copy)]
struct CompiledValue<T: Type> {
    layout: CompiledTypeLayout<T>,
    range: TypeIndexRange,
    write: Option<(CompiledTypeLayout<T>, TypeIndexRange)>,
}

impl<T: Type> CompiledValue<T> {
    fn write(self) -> (CompiledTypeLayout<T>, TypeIndexRange) {
        self.write.unwrap_or((self.layout, self.range))
    }
    fn write_value(self) -> Self {
        let (layout, range) = self.write();
        Self {
            layout,
            range,
            write: None,
        }
    }
    fn map<U: Type>(
        self,
        mut f: impl FnMut(
            CompiledTypeLayout<T>,
            TypeIndexRange,
        ) -> (CompiledTypeLayout<U>, TypeIndexRange),
    ) -> CompiledValue<U> {
        let (layout, range) = f(self.layout, self.range);
        CompiledValue {
            layout,
            range,
            write: self.write.map(|(layout, range)| f(layout, range)),
        }
    }
    fn map_ty<U: Type>(self, mut f: impl FnMut(T) -> U) -> CompiledValue<U> {
        self.map(|CompiledTypeLayout { ty, layout, body }, range| {
            (
                CompiledTypeLayout {
                    ty: f(ty),
                    layout,
                    body,
                },
                range,
            )
        })
    }
}

impl CompiledValue<Bundle> {
    fn field_by_index(self, field_index: usize) -> CompiledValue<CanonicalType> {
        self.map(|layout, range| {
            let CompiledTypeLayout {
                ty: _,
                layout: _,
                body: CompiledTypeLayoutBody::Bundle { fields },
            } = layout
            else {
                unreachable!();
            };
            (
                fields[field_index].ty,
                range.slice(TypeIndexRange::new(
                    fields[field_index].offset,
                    fields[field_index].ty.layout.len(),
                )),
            )
        })
    }
    fn field_by_name(self, name: Interned<str>) -> CompiledValue<CanonicalType> {
        self.field_by_index(self.layout.ty.name_indexes()[&name])
    }
}

impl CompiledValue<Array> {
    fn element(self, index: usize) -> CompiledValue<CanonicalType> {
        self.map(|layout, range| {
            let CompiledTypeLayoutBody::Array { element } = layout.body else {
                unreachable!();
            };
            (*element, range.index_array(element.layout.len(), index))
        })
    }
    fn element_dyn(
        self,
        index_slot: StatePartIndex<StatePartKindSmallSlots>,
    ) -> CompiledExpr<CanonicalType> {
        CompiledExpr::from(self).element_dyn(index_slot)
    }
}

#[derive(Debug, PartialEq, Eq, Hash, Copy, Clone)]
struct CompiledExpr<T: Type> {
    static_part: CompiledValue<T>,
    indexes: TypeArrayIndexes,
}

impl<T: Type> From<CompiledValue<T>> for CompiledExpr<T> {
    fn from(static_part: CompiledValue<T>) -> Self {
        Self {
            static_part,
            indexes: TypeArrayIndexes::default(),
        }
    }
}

impl<T: Type> CompiledExpr<T> {
    fn map_ty<U: Type>(self, f: impl FnMut(T) -> U) -> CompiledExpr<U> {
        let Self {
            static_part,
            indexes,
        } = self;
        CompiledExpr {
            static_part: static_part.map_ty(f),
            indexes,
        }
    }
    fn add_target_without_indexes_to_set(self, inputs: &mut SlotSet) {
        let Self {
            static_part,
            indexes,
        } = self;
        indexes.as_ref().for_each_offset(|offset| {
            inputs.extend([static_part.range.offset(offset)]);
        });
    }
    fn add_target_and_indexes_to_set(self, inputs: &mut SlotSet) {
        let Self {
            static_part: _,
            indexes,
        } = self;
        self.add_target_without_indexes_to_set(inputs);
        inputs.extend(indexes.as_ref().iter());
    }
}

impl CompiledExpr<Bundle> {
    fn field_by_index(self, field_index: usize) -> CompiledExpr<CanonicalType> {
        CompiledExpr {
            static_part: self.static_part.field_by_index(field_index),
            indexes: self.indexes,
        }
    }
    fn field_by_name(self, name: Interned<str>) -> CompiledExpr<CanonicalType> {
        CompiledExpr {
            static_part: self.static_part.field_by_name(name),
            indexes: self.indexes,
        }
    }
}

impl CompiledExpr<Array> {
    fn element(self, index: usize) -> CompiledExpr<CanonicalType> {
        CompiledExpr {
            static_part: self.static_part.element(index),
            indexes: self.indexes,
        }
    }
    fn element_dyn(
        self,
        index_slot: StatePartIndex<StatePartKindSmallSlots>,
    ) -> CompiledExpr<CanonicalType> {
        let CompiledTypeLayoutBody::Array { element } = self.static_part.layout.body else {
            unreachable!();
        };
        let stride = element.layout.len();
        let indexes = self.indexes.join(TypeArrayIndex::from_parts(
            index_slot,
            self.static_part.layout.ty.len(),
            stride,
        ));
        CompiledExpr {
            static_part: self.static_part.map(|layout, range| {
                let CompiledTypeLayoutBody::Array { element } = layout.body else {
                    unreachable!();
                };
                (*element, range.index_array(stride, 0))
            }),
            indexes,
        }
    }
}

#[derive(Copy, Clone, Debug, PartialEq, Eq, Hash)]
enum AssignmentOrSlotIndex {
    AssignmentIndex(usize),
    SmallSlot(StatePartIndex<StatePartKindSmallSlots>),
    BigSlot(StatePartIndex<StatePartKindBigSlots>),
}

#[derive(Copy, Clone, Debug, PartialEq, Eq, Hash)]
enum AssignmentIO {
    BigInput {
        assignment_index: usize,
        slot: StatePartIndex<StatePartKindBigSlots>,
    },
    SmallInput {
        assignment_index: usize,
        slot: StatePartIndex<StatePartKindSmallSlots>,
    },
    BigOutput {
        assignment_index: usize,
        slot: StatePartIndex<StatePartKindBigSlots>,
    },
    SmallOutput {
        assignment_index: usize,
        slot: StatePartIndex<StatePartKindSmallSlots>,
    },
}

#[derive(Copy, Clone, Debug, PartialEq, Eq, Hash)]
enum AssignmentsEdge {
    IO(AssignmentIO),
    AssignmentImmediatePredecessor {
        predecessor_assignment_index: usize,
        assignment_index: usize,
    },
}

#[derive(Debug)]
enum Assignments {
    Accumulating {
        assignments: Vec<Assignment>,
    },
    Finalized {
        assignments: Box<[Assignment]>,
        slots_layout: TypeLayout<InsnsBuildingDone>,
        slot_readers: SlotToAssignmentIndexFullMap,
        slot_writers: SlotToAssignmentIndexFullMap,
        assignment_immediate_predecessors: Box<[Box<[usize]>]>,
        assignment_immediate_successors: Box<[Box<[usize]>]>,
    },
}

impl Default for Assignments {
    fn default() -> Self {
        Self::Accumulating {
            assignments: Vec::new(),
        }
    }
}

impl Assignments {
    fn finalize(&mut self, slots_layout: TypeLayout<InsnsBuildingDone>) {
        let Self::Accumulating { assignments } = self else {
            unreachable!("already finalized");
        };
        let assignments = mem::take(assignments).into_boxed_slice();
        let mut slot_readers = SlotToAssignmentIndexFullMap::new(slots_layout.len());
        let mut slot_writers = SlotToAssignmentIndexFullMap::new(slots_layout.len());
        let mut assignment_immediate_predecessors = vec![BTreeSet::new(); assignments.len()];
        let mut assignment_immediate_successors = vec![BTreeSet::new(); assignments.len()];
        for (assignment_index, assignment) in assignments.iter().enumerate() {
            slot_readers
                .keys_for_assignment(assignment_index)
                .extend([&assignment.inputs]);
            slot_readers
                .keys_for_assignment(assignment_index)
                .extend(&assignment.conditions);
            let SlotSet(TypeParts {
                small_slots,
                big_slots,
            }) = &assignment.outputs;
            for &slot in small_slots {
                if let Some(&pred) = slot_writers[slot].last() {
                    assignment_immediate_predecessors[assignment_index].insert(pred);
                    assignment_immediate_successors[pred].insert(assignment_index);
                }
                slot_writers[slot].push(assignment_index);
            }
            for &slot in big_slots {
                if let Some(&pred) = slot_writers[slot].last() {
                    assignment_immediate_predecessors[assignment_index].insert(pred);
                    assignment_immediate_successors[pred].insert(assignment_index);
                }
                slot_writers[slot].push(assignment_index);
            }
        }
        *self = Self::Finalized {
            assignments,
            slots_layout,
            slot_readers,
            slot_writers,
            assignment_immediate_predecessors: assignment_immediate_predecessors
                .into_iter()
                .map(Box::from_iter)
                .collect(),
            assignment_immediate_successors: assignment_immediate_successors
                .into_iter()
                .map(Box::from_iter)
                .collect(),
        };
    }
    fn push(&mut self, v: Assignment) {
        let Self::Accumulating { assignments } = self else {
            unreachable!("already finalized");
        };
        assignments.push(v);
    }
    fn assignments(&self) -> &[Assignment] {
        let Self::Finalized { assignments, .. } = self else {
            unreachable!("Assignments::finalize should have been called");
        };
        assignments
    }
    fn slots_layout(&self) -> TypeLayout<InsnsBuildingDone> {
        let Self::Finalized { slots_layout, .. } = self else {
            unreachable!("Assignments::finalize should have been called");
        };
        *slots_layout
    }
    fn slot_readers(&self) -> &SlotToAssignmentIndexFullMap {
        let Self::Finalized { slot_readers, .. } = self else {
            unreachable!("Assignments::finalize should have been called");
        };
        slot_readers
    }
    fn slot_writers(&self) -> &SlotToAssignmentIndexFullMap {
        let Self::Finalized { slot_writers, .. } = self else {
            unreachable!("Assignments::finalize should have been called");
        };
        slot_writers
    }
    fn assignment_immediate_predecessors(&self) -> &[Box<[usize]>] {
        let Self::Finalized {
            assignment_immediate_predecessors,
            ..
        } = self
        else {
            unreachable!("Assignments::finalize should have been called");
        };
        assignment_immediate_predecessors
    }
    fn assignment_immediate_successors(&self) -> &[Box<[usize]>] {
        let Self::Finalized {
            assignment_immediate_successors,
            ..
        } = self
        else {
            unreachable!("Assignments::finalize should have been called");
        };
        assignment_immediate_successors
    }
    fn elements(&self) -> AssignmentsElements<'_> {
        let SlotToAssignmentIndexFullMap(TypeParts {
            small_slots,
            big_slots,
        }) = self.slot_readers();
        AssignmentsElements {
            node_indexes: HashMap::with_capacity_and_hasher(
                self.assignments().len() + small_slots.len() + big_slots.len(),
                Default::default(),
            ),
            nodes: self.node_references(),
            edges: self.edge_references(),
        }
    }
}

impl GraphBase for Assignments {
    type EdgeId = AssignmentsEdge;
    type NodeId = AssignmentOrSlotIndex;
}

#[derive(Debug, Clone, Copy)]
enum AssignmentsNodeRef<'a> {
    Assignment {
        index: usize,
        assignment: &'a Assignment,
    },
    SmallSlot(StatePartIndex<StatePartKindSmallSlots>, SlotDebugData),
    BigSlot(StatePartIndex<StatePartKindBigSlots>, SlotDebugData),
}

impl<'a> NodeRef for AssignmentsNodeRef<'a> {
    type NodeId = AssignmentOrSlotIndex;
    type Weight = AssignmentsNodeRef<'a>;

    fn id(&self) -> Self::NodeId {
        match *self {
            AssignmentsNodeRef::Assignment {
                index,
                assignment: _,
            } => AssignmentOrSlotIndex::AssignmentIndex(index),
            AssignmentsNodeRef::SmallSlot(slot, _) => AssignmentOrSlotIndex::SmallSlot(slot),
            AssignmentsNodeRef::BigSlot(slot, _) => AssignmentOrSlotIndex::BigSlot(slot),
        }
    }

    fn weight(&self) -> &Self::Weight {
        self
    }
}

impl<'a> petgraph::visit::Data for &'a Assignments {
    type NodeWeight = AssignmentsNodeRef<'a>;
    type EdgeWeight = AssignmentsEdge;
}

struct AssignmentsElements<'a> {
    node_indexes: HashMap<AssignmentOrSlotIndex, usize>,
    nodes: AssignmentsNodes<'a>,
    edges: AssignmentsEdges<'a>,
}

impl<'a> Iterator for AssignmentsElements<'a> {
    type Item = petgraph::data::Element<
        <&'a Assignments as petgraph::visit::Data>::NodeWeight,
        <&'a Assignments as petgraph::visit::Data>::EdgeWeight,
    >;

    fn next(&mut self) -> Option<Self::Item> {
        let Self {
            node_indexes,
            nodes,
            edges,
        } = self;
        if let Some(node) = nodes.next() {
            node_indexes.insert(node.id(), node_indexes.len());
            return Some(petgraph::data::Element::Node { weight: node });
        }
        let edge = edges.next()?;
        Some(petgraph::data::Element::Edge {
            source: node_indexes[&edge.source()],
            target: node_indexes[&edge.target()],
            weight: *edge.weight(),
        })
    }
}

#[derive(Clone)]
struct AssignmentsNodeIdentifiers {
    assignment_indexes: std::ops::Range<usize>,
    small_slots: std::ops::Range<u32>,
    big_slots: std::ops::Range<u32>,
}

impl AssignmentsNodeIdentifiers {
    fn internal_iter<'a>(&'a mut self) -> impl Iterator<Item = AssignmentOrSlotIndex> + 'a {
        let Self {
            assignment_indexes,
            small_slots,
            big_slots,
        } = self;
        assignment_indexes
            .map(AssignmentOrSlotIndex::AssignmentIndex)
            .chain(small_slots.map(|value| {
                AssignmentOrSlotIndex::SmallSlot(StatePartIndex {
                    value,
                    _phantom: PhantomData,
                })
            }))
            .chain(big_slots.map(|value| {
                AssignmentOrSlotIndex::BigSlot(StatePartIndex {
                    value,
                    _phantom: PhantomData,
                })
            }))
    }
}

impl Iterator for AssignmentsNodeIdentifiers {
    type Item = AssignmentOrSlotIndex;
    fn next(&mut self) -> Option<Self::Item> {
        self.internal_iter().next()
    }

    fn nth(&mut self, n: usize) -> Option<Self::Item> {
        self.internal_iter().nth(n)
    }
}

impl<'a> IntoNodeIdentifiers for &'a Assignments {
    type NodeIdentifiers = AssignmentsNodeIdentifiers;

    fn node_identifiers(self) -> Self::NodeIdentifiers {
        let TypeLen {
            small_slots,
            big_slots,
        } = self.slot_readers().len();
        AssignmentsNodeIdentifiers {
            assignment_indexes: 0..self.assignments().len(),
            small_slots: 0..small_slots.value,
            big_slots: 0..big_slots.value,
        }
    }
}

struct AssignmentsNodes<'a> {
    assignments: &'a Assignments,
    nodes: AssignmentsNodeIdentifiers,
}

impl<'a> Iterator for AssignmentsNodes<'a> {
    type Item = AssignmentsNodeRef<'a>;

    fn next(&mut self) -> Option<Self::Item> {
        self.nodes.next().map(|node| match node {
            AssignmentOrSlotIndex::AssignmentIndex(index) => AssignmentsNodeRef::Assignment {
                index,
                assignment: &self.assignments.assignments()[index],
            },
            AssignmentOrSlotIndex::SmallSlot(slot) => AssignmentsNodeRef::SmallSlot(
                slot,
                *self.assignments.slots_layout().small_slots.debug_data(slot),
            ),
            AssignmentOrSlotIndex::BigSlot(slot) => AssignmentsNodeRef::BigSlot(
                slot,
                *self.assignments.slots_layout().big_slots.debug_data(slot),
            ),
        })
    }
}

impl<'a> IntoNodeReferences for &'a Assignments {
    type NodeRef = AssignmentsNodeRef<'a>;
    type NodeReferences = AssignmentsNodes<'a>;

    fn node_references(self) -> Self::NodeReferences {
        AssignmentsNodes {
            assignments: self,
            nodes: self.node_identifiers(),
        }
    }
}

struct AssignmentsNeighborsDirected<'a> {
    assignment_indexes: std::slice::Iter<'a, usize>,
    small_slots: std::collections::btree_set::Iter<'a, StatePartIndex<StatePartKindSmallSlots>>,
    big_slots: std::collections::btree_set::Iter<'a, StatePartIndex<StatePartKindBigSlots>>,
}

impl Iterator for AssignmentsNeighborsDirected<'_> {
    type Item = AssignmentOrSlotIndex;
    fn next(&mut self) -> Option<Self::Item> {
        let Self {
            assignment_indexes,
            small_slots,
            big_slots,
        } = self;
        if let retval @ Some(_) = assignment_indexes
            .next()
            .copied()
            .map(AssignmentOrSlotIndex::AssignmentIndex)
        {
            retval
        } else if let retval @ Some(_) = small_slots
            .next()
            .copied()
            .map(AssignmentOrSlotIndex::SmallSlot)
        {
            retval
        } else if let retval @ Some(_) = big_slots
            .next()
            .copied()
            .map(AssignmentOrSlotIndex::BigSlot)
        {
            retval
        } else {
            None
        }
    }
}

impl<'a> IntoNeighbors for &'a Assignments {
    type Neighbors = AssignmentsNeighborsDirected<'a>;

    fn neighbors(self, n: Self::NodeId) -> Self::Neighbors {
        self.neighbors_directed(n, petgraph::Direction::Outgoing)
    }
}

impl<'a> IntoNeighborsDirected for &'a Assignments {
    type NeighborsDirected = AssignmentsNeighborsDirected<'a>;

    fn neighbors_directed(
        self,
        n: Self::NodeId,
        d: petgraph::Direction,
    ) -> Self::NeighborsDirected {
        use petgraph::Direction::*;
        let slot_map = match d {
            Outgoing => self.slot_readers(),
            Incoming => self.slot_writers(),
        };
        match n {
            AssignmentOrSlotIndex::AssignmentIndex(assignment_index) => {
                let assignment = &self.assignments()[assignment_index];
                let (
                    assignment_indexes,
                    SlotSet(TypeParts {
                        small_slots,
                        big_slots,
                    }),
                ) = match d {
                    Outgoing => (
                        &self.assignment_immediate_successors()[assignment_index],
                        &assignment.outputs,
                    ),
                    Incoming => (
                        &self.assignment_immediate_predecessors()[assignment_index],
                        &assignment.inputs,
                    ),
                };
                AssignmentsNeighborsDirected {
                    assignment_indexes: assignment_indexes.iter(),
                    small_slots: small_slots.iter(),
                    big_slots: big_slots.iter(),
                }
            }
            AssignmentOrSlotIndex::SmallSlot(slot) => AssignmentsNeighborsDirected {
                assignment_indexes: slot_map[slot].iter(),
                small_slots: Default::default(),
                big_slots: Default::default(),
            },
            AssignmentOrSlotIndex::BigSlot(slot) => AssignmentsNeighborsDirected {
                assignment_indexes: slot_map[slot].iter(),
                small_slots: Default::default(),
                big_slots: Default::default(),
            },
        }
    }
}

impl EdgeRef for AssignmentsEdge {
    type NodeId = AssignmentOrSlotIndex;
    type EdgeId = AssignmentsEdge;
    type Weight = AssignmentsEdge;

    fn source(&self) -> Self::NodeId {
        match *self {
            AssignmentsEdge::IO(AssignmentIO::BigInput {
                assignment_index: _,
                slot,
            }) => AssignmentOrSlotIndex::BigSlot(slot),
            AssignmentsEdge::IO(AssignmentIO::SmallInput {
                assignment_index: _,
                slot,
            }) => AssignmentOrSlotIndex::SmallSlot(slot),
            AssignmentsEdge::IO(AssignmentIO::BigOutput {
                assignment_index,
                slot: _,
            }) => AssignmentOrSlotIndex::AssignmentIndex(assignment_index),
            AssignmentsEdge::IO(AssignmentIO::SmallOutput {
                assignment_index,
                slot: _,
            }) => AssignmentOrSlotIndex::AssignmentIndex(assignment_index),
            AssignmentsEdge::AssignmentImmediatePredecessor {
                predecessor_assignment_index,
                assignment_index: _,
            } => AssignmentOrSlotIndex::AssignmentIndex(predecessor_assignment_index),
        }
    }

    fn target(&self) -> Self::NodeId {
        match *self {
            AssignmentsEdge::IO(AssignmentIO::BigInput {
                assignment_index,
                slot: _,
            }) => AssignmentOrSlotIndex::AssignmentIndex(assignment_index),
            AssignmentsEdge::IO(AssignmentIO::SmallInput {
                assignment_index,
                slot: _,
            }) => AssignmentOrSlotIndex::AssignmentIndex(assignment_index),
            AssignmentsEdge::IO(AssignmentIO::BigOutput {
                assignment_index: _,
                slot,
            }) => AssignmentOrSlotIndex::BigSlot(slot),
            AssignmentsEdge::IO(AssignmentIO::SmallOutput {
                assignment_index: _,
                slot,
            }) => AssignmentOrSlotIndex::SmallSlot(slot),
            AssignmentsEdge::AssignmentImmediatePredecessor {
                predecessor_assignment_index: _,
                assignment_index,
            } => AssignmentOrSlotIndex::AssignmentIndex(assignment_index),
        }
    }

    fn weight(&self) -> &Self::Weight {
        self
    }

    fn id(&self) -> Self::EdgeId {
        *self
    }
}

struct AssignmentsEdges<'a> {
    assignments: &'a Assignments,
    nodes: AssignmentsNodeIdentifiers,
    outgoing_neighbors: Option<(AssignmentOrSlotIndex, AssignmentsNeighborsDirected<'a>)>,
}

impl Iterator for AssignmentsEdges<'_> {
    type Item = AssignmentsEdge;

    fn next(&mut self) -> Option<Self::Item> {
        loop {
            if let Some((node, outgoing_neighbors)) = &mut self.outgoing_neighbors {
                if let Some(outgoing_neighbor) = outgoing_neighbors.next() {
                    return Some(match (*node, outgoing_neighbor) {
                        (
                            AssignmentOrSlotIndex::SmallSlot(_) | AssignmentOrSlotIndex::BigSlot(_),
                            AssignmentOrSlotIndex::SmallSlot(_) | AssignmentOrSlotIndex::BigSlot(_),
                        ) => unreachable!(),
                        (
                            AssignmentOrSlotIndex::AssignmentIndex(predecessor_assignment_index),
                            AssignmentOrSlotIndex::AssignmentIndex(assignment_index),
                        ) => AssignmentsEdge::AssignmentImmediatePredecessor {
                            predecessor_assignment_index,
                            assignment_index,
                        },
                        (
                            AssignmentOrSlotIndex::AssignmentIndex(assignment_index),
                            AssignmentOrSlotIndex::SmallSlot(slot),
                        ) => AssignmentsEdge::IO(AssignmentIO::SmallOutput {
                            assignment_index,
                            slot,
                        }),
                        (
                            AssignmentOrSlotIndex::AssignmentIndex(assignment_index),
                            AssignmentOrSlotIndex::BigSlot(slot),
                        ) => AssignmentsEdge::IO(AssignmentIO::BigOutput {
                            assignment_index,
                            slot,
                        }),
                        (
                            AssignmentOrSlotIndex::SmallSlot(slot),
                            AssignmentOrSlotIndex::AssignmentIndex(assignment_index),
                        ) => AssignmentsEdge::IO(AssignmentIO::SmallInput {
                            assignment_index,
                            slot,
                        }),
                        (
                            AssignmentOrSlotIndex::BigSlot(slot),
                            AssignmentOrSlotIndex::AssignmentIndex(assignment_index),
                        ) => AssignmentsEdge::IO(AssignmentIO::BigInput {
                            assignment_index,
                            slot,
                        }),
                    });
                }
            }
            let node = self.nodes.next()?;
            self.outgoing_neighbors = Some((
                node,
                self.assignments
                    .neighbors_directed(node, petgraph::Direction::Outgoing),
            ));
        }
    }
}

impl<'a> IntoEdgeReferences for &'a Assignments {
    type EdgeRef = AssignmentsEdge;
    type EdgeReferences = AssignmentsEdges<'a>;

    fn edge_references(self) -> Self::EdgeReferences {
        AssignmentsEdges {
            assignments: self,
            nodes: self.node_identifiers(),
            outgoing_neighbors: None,
        }
    }
}

struct AssignmentsVisitMap {
    assignments: Vec<bool>,
    slots: DenseSlotSet,
}

impl VisitMap<AssignmentOrSlotIndex> for AssignmentsVisitMap {
    fn visit(&mut self, n: AssignmentOrSlotIndex) -> bool {
        match n {
            AssignmentOrSlotIndex::AssignmentIndex(assignment_index) => {
                !mem::replace(&mut self.assignments[assignment_index], true)
            }
            AssignmentOrSlotIndex::SmallSlot(slot) => self.slots.insert(slot),
            AssignmentOrSlotIndex::BigSlot(slot) => self.slots.insert(slot),
        }
    }

    fn is_visited(&self, n: &AssignmentOrSlotIndex) -> bool {
        match *n {
            AssignmentOrSlotIndex::AssignmentIndex(assignment_index) => {
                self.assignments[assignment_index]
            }
            AssignmentOrSlotIndex::SmallSlot(slot) => self.slots.contains(slot),
            AssignmentOrSlotIndex::BigSlot(slot) => self.slots.contains(slot),
        }
    }
}

impl Visitable for Assignments {
    type Map = AssignmentsVisitMap;

    fn visit_map(self: &Self) -> Self::Map {
        AssignmentsVisitMap {
            assignments: vec![false; self.assignments().len()],
            slots: DenseSlotSet::new(self.slot_readers().len()),
        }
    }

    fn reset_map(self: &Self, map: &mut Self::Map) {
        let AssignmentsVisitMap { assignments, slots } = map;
        assignments.clear();
        assignments.resize(self.assignments().len(), false);
        if slots.len() != self.slot_readers().len() {
            *slots = DenseSlotSet::new(self.slot_readers().len());
        } else {
            slots.clear();
        }
    }
}

#[derive(Clone, Debug, PartialEq, Eq, Hash)]
struct DenseSlotSet(TypeParts<Self>);

impl DenseSlotSet {
    fn new(len: TypeLen) -> Self {
        let TypeLen {
            small_slots,
            big_slots,
        } = len;
        Self(TypeParts {
            small_slots: vec![false; small_slots.value.try_into().expect("length too big")]
                .into_boxed_slice(),
            big_slots: vec![false; big_slots.value.try_into().expect("length too big")]
                .into_boxed_slice(),
        })
    }
    fn len(&self) -> TypeLen {
        TypeLen {
            small_slots: StatePartLen {
                value: self.0.small_slots.len() as _,
                _phantom: PhantomData,
            },
            big_slots: StatePartLen {
                value: self.0.big_slots.len() as _,
                _phantom: PhantomData,
            },
        }
    }
    fn clear(&mut self) {
        let Self(TypeParts {
            small_slots,
            big_slots,
        }) = self;
        small_slots.fill(false);
        big_slots.fill(false);
    }
}

impl StatePartsValue for DenseSlotSet {
    type Value<K: StatePartKind> = Box<[bool]>;
}

trait DenseSlotSetMethods<K: StatePartKind>: Extend<StatePartIndex<K>> {
    fn contains(&self, k: StatePartIndex<K>) -> bool;
    fn remove(&mut self, k: StatePartIndex<K>) -> bool {
        self.take(k).is_some()
    }
    fn take(&mut self, k: StatePartIndex<K>) -> Option<StatePartIndex<K>>;
    fn replace(&mut self, k: StatePartIndex<K>) -> Option<StatePartIndex<K>>;
    fn insert(&mut self, k: StatePartIndex<K>) -> bool {
        self.replace(k).is_none()
    }
}

impl<K: StatePartKind> Extend<StatePartIndex<K>> for DenseSlotSet
where
    Self: DenseSlotSetMethods<K>,
{
    fn extend<T: IntoIterator<Item = StatePartIndex<K>>>(&mut self, iter: T) {
        iter.into_iter().for_each(|v| {
            self.insert(v);
        });
    }
}

impl DenseSlotSetMethods<StatePartKindSmallSlots> for DenseSlotSet {
    fn contains(&self, k: StatePartIndex<StatePartKindSmallSlots>) -> bool {
        self.0.small_slots[k.as_usize()]
    }

    fn take(
        &mut self,
        k: StatePartIndex<StatePartKindSmallSlots>,
    ) -> Option<StatePartIndex<StatePartKindSmallSlots>> {
        mem::replace(self.0.small_slots.get_mut(k.as_usize())?, false).then_some(k)
    }

    fn replace(
        &mut self,
        k: StatePartIndex<StatePartKindSmallSlots>,
    ) -> Option<StatePartIndex<StatePartKindSmallSlots>> {
        mem::replace(&mut self.0.small_slots[k.as_usize()], true).then_some(k)
    }
}

impl DenseSlotSetMethods<StatePartKindBigSlots> for DenseSlotSet {
    fn contains(&self, k: StatePartIndex<StatePartKindBigSlots>) -> bool {
        self.0.big_slots[k.as_usize()]
    }

    fn take(
        &mut self,
        k: StatePartIndex<StatePartKindBigSlots>,
    ) -> Option<StatePartIndex<StatePartKindBigSlots>> {
        mem::replace(self.0.big_slots.get_mut(k.as_usize())?, false).then_some(k)
    }

    fn replace(
        &mut self,
        k: StatePartIndex<StatePartKindBigSlots>,
    ) -> Option<StatePartIndex<StatePartKindBigSlots>> {
        mem::replace(&mut self.0.big_slots[k.as_usize()], true).then_some(k)
    }
}

#[derive(Clone, Debug, Default, PartialEq, Eq, Hash)]
struct SlotVec(TypeParts<Self>);

impl SlotVec {
    fn is_empty(&self) -> bool {
        let Self(TypeParts {
            small_slots,
            big_slots,
        }) = self;
        small_slots.is_empty() && big_slots.is_empty()
    }
}

impl StatePartsValue for SlotVec {
    type Value<K: StatePartKind> = Vec<StatePartIndex<K>>;
}

#[derive(Clone, Debug, Default, PartialEq, Eq, Hash)]
struct SlotSet(TypeParts<Self>);

impl SlotSet {
    fn is_empty(&self) -> bool {
        let Self(TypeParts {
            small_slots,
            big_slots,
        }) = self;
        small_slots.is_empty() && big_slots.is_empty()
    }
    fn for_each(
        &self,
        small_slots_fn: impl FnMut(StatePartIndex<StatePartKindSmallSlots>),
        big_slots_fn: impl FnMut(StatePartIndex<StatePartKindBigSlots>),
    ) {
        let Self(TypeParts {
            small_slots,
            big_slots,
        }) = self;
        small_slots.iter().copied().for_each(small_slots_fn);
        big_slots.iter().copied().for_each(big_slots_fn);
    }
    fn all(
        &self,
        small_slots_fn: impl FnMut(StatePartIndex<StatePartKindSmallSlots>) -> bool,
        big_slots_fn: impl FnMut(StatePartIndex<StatePartKindBigSlots>) -> bool,
    ) -> bool {
        let Self(TypeParts {
            small_slots,
            big_slots,
        }) = self;
        small_slots.iter().copied().all(small_slots_fn)
            && big_slots.iter().copied().all(big_slots_fn)
    }
}

impl StatePartsValue for SlotSet {
    type Value<K: StatePartKind> = BTreeSet<StatePartIndex<K>>;
}

impl Extend<StatePartIndex<StatePartKindSmallSlots>> for SlotSet {
    fn extend<T: IntoIterator<Item = StatePartIndex<StatePartKindSmallSlots>>>(&mut self, iter: T) {
        self.0.small_slots.extend(iter);
    }
}

impl Extend<StatePartIndex<StatePartKindBigSlots>> for SlotSet {
    fn extend<T: IntoIterator<Item = StatePartIndex<StatePartKindBigSlots>>>(&mut self, iter: T) {
        self.0.big_slots.extend(iter);
    }
}

impl<K: StatePartKind> Extend<StatePartIndexRange<K>> for SlotSet
where
    Self: Extend<StatePartIndex<K>>,
{
    fn extend<T: IntoIterator<Item = StatePartIndexRange<K>>>(&mut self, iter: T) {
        self.extend(iter.into_iter().flat_map(|v| v.iter()));
    }
}

impl Extend<TypeIndexRange> for SlotSet {
    fn extend<T: IntoIterator<Item = TypeIndexRange>>(&mut self, iter: T) {
        iter.into_iter().for_each(
            |TypeIndexRange {
                 small_slots,
                 big_slots,
             }| {
                self.extend(small_slots.iter());
                self.extend(big_slots.iter());
            },
        )
    }
}

impl Extend<TypeArrayIndex> for SlotSet {
    fn extend<T: IntoIterator<Item = TypeArrayIndex>>(&mut self, iter: T) {
        iter.into_iter().for_each(
            |TypeArrayIndex {
                 small_slots,
                 big_slots,
             }| {
                self.extend([small_slots]);
                self.extend([big_slots]);
            },
        )
    }
}

impl<K: StatePartKind> Extend<StatePartArrayIndex<K>> for SlotSet {
    fn extend<T: IntoIterator<Item = StatePartArrayIndex<K>>>(&mut self, iter: T) {
        self.extend(iter.into_iter().map(|v| v.index));
    }
}

impl Extend<CondBody> for SlotSet {
    fn extend<T: IntoIterator<Item = CondBody>>(&mut self, iter: T) {
        iter.into_iter().for_each(|cond_body| match cond_body {
            CondBody::IfTrue { cond } | CondBody::IfFalse { cond } => {
                self.extend([cond.range]);
            }
            CondBody::MatchArm {
                discriminant,
                variant_index: _,
            } => self.extend([discriminant]),
        })
    }
}

impl Extend<Cond> for SlotSet {
    fn extend<T: IntoIterator<Item = Cond>>(&mut self, iter: T) {
        self.extend(iter.into_iter().map(|v| v.body))
    }
}

#[derive(Debug)]
struct Assignment {
    inputs: SlotSet,
    outputs: SlotSet,
    conditions: Interned<[Cond]>,
    insns: Vec<InsnOrLabel>,
    source_location: SourceLocation,
}

#[derive(Debug)]
struct SlotToAssignmentIndexFullMap(TypeParts<Self>);

impl StatePartsValue for SlotToAssignmentIndexFullMap {
    type Value<K: StatePartKind> = Box<[Vec<usize>]>;
}

impl SlotToAssignmentIndexFullMap {
    fn new(len: TypeLen) -> Self {
        let TypeLen {
            small_slots,
            big_slots,
        } = len;
        Self(TypeParts {
            small_slots: vec![Vec::new(); small_slots.value.try_into().expect("length too big")]
                .into_boxed_slice(),
            big_slots: vec![Vec::new(); big_slots.value.try_into().expect("length too big")]
                .into_boxed_slice(),
        })
    }
    fn len(&self) -> TypeLen {
        TypeLen {
            small_slots: StatePartLen {
                value: self.0.small_slots.len() as _,
                _phantom: PhantomData,
            },
            big_slots: StatePartLen {
                value: self.0.big_slots.len() as _,
                _phantom: PhantomData,
            },
        }
    }
    fn keys_for_assignment(
        &mut self,
        assignment_index: usize,
    ) -> SlotToAssignmentIndexFullMapKeysForAssignment<'_> {
        SlotToAssignmentIndexFullMapKeysForAssignment {
            map: self,
            assignment_index,
        }
    }
    fn for_each(
        &self,
        mut small_slots_fn: impl FnMut(StatePartIndex<StatePartKindSmallSlots>, &[usize]),
        mut big_slots_fn: impl FnMut(StatePartIndex<StatePartKindBigSlots>, &[usize]),
    ) {
        let Self(TypeParts {
            small_slots,
            big_slots,
        }) = self;
        small_slots.iter().enumerate().for_each(|(k, v)| {
            small_slots_fn(
                StatePartIndex {
                    value: k as _,
                    _phantom: PhantomData,
                },
                v,
            )
        });
        big_slots.iter().enumerate().for_each(|(k, v)| {
            big_slots_fn(
                StatePartIndex {
                    value: k as _,
                    _phantom: PhantomData,
                },
                v,
            )
        });
    }
}

impl std::ops::Index<StatePartIndex<StatePartKindSmallSlots>> for SlotToAssignmentIndexFullMap {
    type Output = Vec<usize>;

    fn index(&self, index: StatePartIndex<StatePartKindSmallSlots>) -> &Self::Output {
        &self.0.small_slots[index.as_usize()]
    }
}

impl std::ops::IndexMut<StatePartIndex<StatePartKindSmallSlots>> for SlotToAssignmentIndexFullMap {
    fn index_mut(&mut self, index: StatePartIndex<StatePartKindSmallSlots>) -> &mut Self::Output {
        &mut self.0.small_slots[index.as_usize()]
    }
}

impl std::ops::Index<StatePartIndex<StatePartKindBigSlots>> for SlotToAssignmentIndexFullMap {
    type Output = Vec<usize>;

    fn index(&self, index: StatePartIndex<StatePartKindBigSlots>) -> &Self::Output {
        &self.0.big_slots[index.as_usize()]
    }
}

impl std::ops::IndexMut<StatePartIndex<StatePartKindBigSlots>> for SlotToAssignmentIndexFullMap {
    fn index_mut(&mut self, index: StatePartIndex<StatePartKindBigSlots>) -> &mut Self::Output {
        &mut self.0.big_slots[index.as_usize()]
    }
}

struct SlotToAssignmentIndexFullMapKeysForAssignment<'a> {
    map: &'a mut SlotToAssignmentIndexFullMap,
    assignment_index: usize,
}

impl<'a, K: StatePartKind> Extend<&'a StatePartIndex<K>>
    for SlotToAssignmentIndexFullMapKeysForAssignment<'_>
where
    Self: Extend<StatePartIndex<K>>,
{
    fn extend<T: IntoIterator<Item = &'a StatePartIndex<K>>>(&mut self, iter: T) {
        self.extend(iter.into_iter().copied());
    }
}

impl<K: StatePartKind> Extend<StatePartIndex<K>>
    for SlotToAssignmentIndexFullMapKeysForAssignment<'_>
where
    SlotToAssignmentIndexFullMap: IndexMut<StatePartIndex<K>, Output = Vec<usize>>,
{
    fn extend<T: IntoIterator<Item = StatePartIndex<K>>>(&mut self, iter: T) {
        iter.into_iter()
            .for_each(|slot| self.map[slot].push(self.assignment_index));
    }
}

impl<'a> Extend<&'a SlotSet> for SlotToAssignmentIndexFullMapKeysForAssignment<'_> {
    fn extend<T: IntoIterator<Item = &'a SlotSet>>(&mut self, iter: T) {
        iter.into_iter().for_each(
            |SlotSet(TypeParts {
                 small_slots,
                 big_slots,
             })| {
                self.extend(small_slots);
                self.extend(big_slots);
            },
        );
    }
}

impl<'a> Extend<&'a Cond> for SlotToAssignmentIndexFullMapKeysForAssignment<'_> {
    fn extend<T: IntoIterator<Item = &'a Cond>>(&mut self, iter: T) {
        iter.into_iter().for_each(|cond| match cond.body {
            CondBody::IfTrue { cond } | CondBody::IfFalse { cond } => {
                let CompiledValue {
                    range:
                        TypeIndexRange {
                            small_slots,
                            big_slots,
                        },
                    layout: _,
                    write: _,
                } = cond;
                self.extend(small_slots.iter());
                self.extend(big_slots.iter());
            }
            CondBody::MatchArm {
                discriminant,
                variant_index: _,
            } => self.extend([discriminant]),
        });
    }
}

impl Assignment {
    fn new(
        conditions: Interned<[Cond]>,
        insns: Vec<InsnOrLabel>,
        source_location: SourceLocation,
    ) -> Self {
        let mut inputs = SlotSet::default();
        let mut outputs = SlotSet::default();
        for insn in &insns {
            let insn = match insn {
                InsnOrLabel::Insn(insn) => insn,
                InsnOrLabel::Label(_) => continue,
            };
            for InsnField { ty, kind } in insn.fields() {
                match (kind, ty) {
                    (InsnFieldKind::Input, InsnFieldType::SmallSlot(&slot)) => {
                        inputs.extend([slot]);
                    }
                    (InsnFieldKind::Input, InsnFieldType::BigSlot(&slot)) => {
                        inputs.extend([slot]);
                    }
                    (
                        InsnFieldKind::Input,
                        InsnFieldType::SmallSlotArrayIndexed(&array_indexed),
                    ) => {
                        array_indexed.for_each_target(|slot| inputs.extend([slot]));
                        inputs.extend(array_indexed.indexes);
                    }
                    (InsnFieldKind::Input, InsnFieldType::BigSlotArrayIndexed(&array_indexed)) => {
                        array_indexed.for_each_target(|slot| inputs.extend([slot]));
                        inputs.extend(array_indexed.indexes);
                    }
                    (InsnFieldKind::Output, InsnFieldType::SmallSlot(&slot)) => {
                        outputs.extend([slot]);
                    }
                    (InsnFieldKind::Output, InsnFieldType::BigSlot(&slot)) => {
                        outputs.extend([slot]);
                    }
                    (
                        InsnFieldKind::Output,
                        InsnFieldType::SmallSlotArrayIndexed(&array_indexed),
                    ) => {
                        array_indexed.for_each_target(|slot| {
                            outputs.extend([slot]);
                        });
                        inputs.extend(array_indexed.indexes);
                    }
                    (InsnFieldKind::Output, InsnFieldType::BigSlotArrayIndexed(&array_indexed)) => {
                        array_indexed.for_each_target(|slot| {
                            outputs.extend([slot]);
                        });
                        inputs.extend(array_indexed.indexes);
                    }
                    (
                        _,
                        InsnFieldType::Memory(_)
                        | InsnFieldType::SmallUInt(_)
                        | InsnFieldType::SmallSInt(_)
                        | InsnFieldType::InternedBigInt(_)
                        | InsnFieldType::U8(_)
                        | InsnFieldType::USize(_)
                        | InsnFieldType::Empty(_),
                    )
                    | (
                        InsnFieldKind::Immediate
                        | InsnFieldKind::Memory
                        | InsnFieldKind::BranchTarget,
                        _,
                    ) => {}
                }
            }
        }
        Self {
            inputs,
            outputs,
            conditions,
            insns,
            source_location,
        }
    }
}

#[derive(Debug)]
struct RegisterReset {
    is_async: bool,
    init: CompiledValue<CanonicalType>,
    rst: StatePartIndex<StatePartKindSmallSlots>,
}

#[derive(Debug, Clone, Copy)]
struct ClockTrigger {
    last_clk_was_low: StatePartIndex<StatePartKindSmallSlots>,
    clk: StatePartIndex<StatePartKindSmallSlots>,
    clk_triggered: StatePartIndex<StatePartKindSmallSlots>,
    source_location: SourceLocation,
}

#[derive(Debug)]
struct Register {
    value: CompiledValue<CanonicalType>,
    clk_triggered: StatePartIndex<StatePartKindSmallSlots>,
    reset: Option<RegisterReset>,
    source_location: SourceLocation,
}

#[derive(Debug)]

struct MemoryPort {
    clk_triggered: StatePartIndex<StatePartKindSmallSlots>,
    addr_delayed: Vec<StatePartIndex<StatePartKindSmallSlots>>,
    en_delayed: Vec<StatePartIndex<StatePartKindSmallSlots>>,
    data_layout: CompiledTypeLayout<CanonicalType>,
    read_data_delayed: Vec<TypeIndexRange>,
    write_data_delayed: Vec<TypeIndexRange>,
    write_mask_delayed: Vec<TypeIndexRange>,
    write_mode_delayed: Vec<StatePartIndex<StatePartKindSmallSlots>>,
    write_insns: Vec<InsnOrLabel>,
}

struct MemoryPortReadInsns<'a> {
    addr: StatePartIndex<StatePartKindSmallSlots>,
    en: StatePartIndex<StatePartKindSmallSlots>,
    write_mode: Option<StatePartIndex<StatePartKindSmallSlots>>,
    data: TypeIndexRange,
    insns: &'a mut Vec<InsnOrLabel>,
}

struct MemoryPortWriteInsns<'a> {
    addr: StatePartIndex<StatePartKindSmallSlots>,
    en: StatePartIndex<StatePartKindSmallSlots>,
    write_mode: Option<StatePartIndex<StatePartKindSmallSlots>>,
    data: TypeIndexRange,
    mask: TypeIndexRange,
    insns: &'a mut Vec<InsnOrLabel>,
}

#[derive(Debug)]
struct Memory {
    mem: Mem,
    memory: StatePartIndex<StatePartKindMemories>,
    trace: TraceMem,
    ports: Vec<MemoryPort>,
}

#[derive(Copy, Clone)]
enum MakeTraceDeclTarget {
    Expr(Expr<CanonicalType>),
    Memory {
        id: TraceMemoryId,
        depth: usize,
        stride: usize,
        start: usize,
        ty: CanonicalType,
    },
}

impl MakeTraceDeclTarget {
    fn flow(self) -> Flow {
        match self {
            MakeTraceDeclTarget::Expr(expr) => Expr::flow(expr),
            MakeTraceDeclTarget::Memory { .. } => Flow::Duplex,
        }
    }
    fn ty(self) -> CanonicalType {
        match self {
            MakeTraceDeclTarget::Expr(expr) => Expr::ty(expr),
            MakeTraceDeclTarget::Memory { ty, .. } => ty,
        }
    }
}

struct DebugOpaque<T>(T);

impl<T> fmt::Debug for DebugOpaque<T> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.write_str("<...>")
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
struct CompiledExternModule {
    module_io_targets: Interned<[Target]>,
    module_io: Interned<[CompiledValue<CanonicalType>]>,
    simulation: ExternModuleSimulation,
}

#[derive(Debug)]
pub struct Compiler {
    insns: Insns<InsnsBuilding>,
    original_base_module: Interned<Module<Bundle>>,
    base_module: Interned<Module<Bundle>>,
    modules: HashMap<InstantiatedModule, CompiledModule>,
    extern_modules: Vec<CompiledExternModule>,
    compiled_values: HashMap<TargetInInstantiatedModule, CompiledValue<CanonicalType>>,
    compiled_exprs: HashMap<Expr<CanonicalType>, CompiledExpr<CanonicalType>>,
    compiled_exprs_to_values: HashMap<CompiledExpr<CanonicalType>, CompiledValue<CanonicalType>>,
    decl_conditions: HashMap<TargetInInstantiatedModule, Interned<[Cond]>>,
    compiled_values_to_dyn_array_indexes:
        HashMap<CompiledValue<UInt>, StatePartIndex<StatePartKindSmallSlots>>,
    compiled_value_bool_dest_is_small_map:
        HashMap<CompiledValue<CanonicalType>, StatePartIndex<StatePartKindSmallSlots>>,
    assignments: Assignments,
    clock_triggers: Vec<ClockTrigger>,
    compiled_value_to_clock_trigger_map: HashMap<CompiledValue<Clock>, ClockTrigger>,
    enum_discriminants: HashMap<CompiledValue<Enum>, StatePartIndex<StatePartKindSmallSlots>>,
    registers: Vec<Register>,
    traces: SimTraces<Vec<SimTrace<SimTraceKind, ()>>>,
    memories: Vec<Memory>,
    dump_assignments_dot: Option<DebugOpaque<Box<dyn Fn(&dyn fmt::Debug)>>>,
}

impl Compiler {
    pub fn new(base_module: Interned<Module<Bundle>>) -> Self {
        let original_base_module = base_module;
        let base_module = deduce_resets(base_module, true)
            .unwrap_or_else(|e| panic!("failed to deduce reset types: {e}"));
        Self {
            insns: Insns::new(),
            original_base_module,
            base_module,
            modules: HashMap::default(),
            extern_modules: Vec::new(),
            compiled_values: HashMap::default(),
            compiled_exprs: HashMap::default(),
            compiled_exprs_to_values: HashMap::default(),
            decl_conditions: HashMap::default(),
            compiled_values_to_dyn_array_indexes: HashMap::default(),
            compiled_value_bool_dest_is_small_map: HashMap::default(),
            assignments: Assignments::default(),
            clock_triggers: Vec::new(),
            compiled_value_to_clock_trigger_map: HashMap::default(),
            enum_discriminants: HashMap::default(),
            registers: Vec::new(),
            traces: SimTraces(Vec::new()),
            memories: Vec::new(),
            dump_assignments_dot: None,
        }
    }
    #[doc(hidden)]
    /// This is explicitly unstable and may be changed/removed at any time
    pub fn dump_assignments_dot(&mut self, callback: Box<dyn Fn(&dyn fmt::Debug)>) {
        self.dump_assignments_dot = Some(DebugOpaque(callback));
    }
    fn new_sim_trace(&mut self, kind: SimTraceKind) -> TraceScalarId {
        let id = TraceScalarId(self.traces.0.len());
        self.traces.0.push(SimTrace {
            kind,
            state: (),
            last_state: (),
        });
        id
    }
    fn make_trace_scalar_helper(
        &mut self,
        instantiated_module: InstantiatedModule,
        target: MakeTraceDeclTarget,
        source_location: SourceLocation,
        small_kind: impl FnOnce(StatePartIndex<StatePartKindSmallSlots>) -> SimTraceKind,
        big_kind: impl FnOnce(StatePartIndex<StatePartKindBigSlots>) -> SimTraceKind,
    ) -> TraceLocation {
        match target {
            MakeTraceDeclTarget::Expr(target) => {
                let compiled_value = self.compile_expr(instantiated_module, target);
                let compiled_value = self.compiled_expr_to_value(compiled_value, source_location);
                TraceLocation::Scalar(self.new_sim_trace(match compiled_value.range.len() {
                    TypeLen::A_SMALL_SLOT => small_kind(compiled_value.range.small_slots.start),
                    TypeLen::A_BIG_SLOT => big_kind(compiled_value.range.big_slots.start),
                    _ => unreachable!(),
                }))
            }
            MakeTraceDeclTarget::Memory {
                id,
                depth,
                stride,
                start,
                ty,
            } => TraceLocation::Memory(TraceMemoryLocation {
                id,
                depth,
                stride,
                start,
                len: ty.bit_width(),
            }),
        }
    }
    fn make_trace_scalar(
        &mut self,
        instantiated_module: InstantiatedModule,
        target: MakeTraceDeclTarget,
        name: Interned<str>,
        source_location: SourceLocation,
    ) -> TraceDecl {
        let flow = target.flow();
        match target.ty() {
            CanonicalType::UInt(ty) => TraceUInt {
                location: self.make_trace_scalar_helper(
                    instantiated_module,
                    target,
                    source_location,
                    |index| SimTraceKind::SmallUInt { index, ty },
                    |index| SimTraceKind::BigUInt { index, ty },
                ),
                name,
                ty,
                flow,
            }
            .into(),
            CanonicalType::SInt(ty) => TraceSInt {
                location: self.make_trace_scalar_helper(
                    instantiated_module,
                    target,
                    source_location,
                    |index| SimTraceKind::SmallSInt { index, ty },
                    |index| SimTraceKind::BigSInt { index, ty },
                ),
                name,
                ty,
                flow,
            }
            .into(),
            CanonicalType::Bool(_) => TraceBool {
                location: self.make_trace_scalar_helper(
                    instantiated_module,
                    target,
                    source_location,
                    |index| SimTraceKind::SmallBool { index },
                    |index| SimTraceKind::BigBool { index },
                ),
                name,
                flow,
            }
            .into(),
            CanonicalType::Array(_) => unreachable!(),
            CanonicalType::Enum(ty) => {
                assert_eq!(ty.discriminant_bit_width(), ty.type_properties().bit_width);
                let location = match target {
                    MakeTraceDeclTarget::Expr(target) => {
                        let compiled_value = self.compile_expr(instantiated_module, target);
                        let compiled_value =
                            self.compiled_expr_to_value(compiled_value, source_location);
                        let discriminant = self.compile_enum_discriminant(
                            compiled_value.map_ty(Enum::from_canonical),
                            source_location,
                        );
                        TraceLocation::Scalar(self.new_sim_trace(SimTraceKind::EnumDiscriminant {
                            index: discriminant,
                            ty,
                        }))
                    }
                    MakeTraceDeclTarget::Memory {
                        id,
                        depth,
                        stride,
                        start,
                        ty: _,
                    } => TraceLocation::Memory(TraceMemoryLocation {
                        id,
                        depth,
                        stride,
                        start,
                        len: ty.type_properties().bit_width,
                    }),
                };
                TraceFieldlessEnum {
                    location,
                    name,
                    ty,
                    flow,
                }
                .into()
            }
            CanonicalType::Bundle(_) | CanonicalType::PhantomConst(_) => unreachable!(),
            CanonicalType::AsyncReset(_) => TraceAsyncReset {
                location: self.make_trace_scalar_helper(
                    instantiated_module,
                    target,
                    source_location,
                    |index| SimTraceKind::SmallAsyncReset { index },
                    |index| SimTraceKind::BigAsyncReset { index },
                ),
                name,
                flow,
            }
            .into(),
            CanonicalType::SyncReset(_) => TraceSyncReset {
                location: self.make_trace_scalar_helper(
                    instantiated_module,
                    target,
                    source_location,
                    |index| SimTraceKind::SmallSyncReset { index },
                    |index| SimTraceKind::BigSyncReset { index },
                ),
                name,
                flow,
            }
            .into(),
            CanonicalType::Reset(_) => unreachable!(),
            CanonicalType::Clock(_) => TraceClock {
                location: self.make_trace_scalar_helper(
                    instantiated_module,
                    target,
                    source_location,
                    |index| SimTraceKind::SmallClock { index },
                    |index| SimTraceKind::BigClock { index },
                ),
                name,
                flow,
            }
            .into(),
        }
    }
    fn make_trace_decl_child(
        &mut self,
        instantiated_module: InstantiatedModule,
        target: MakeTraceDeclTarget,
        name: Interned<str>,
        source_location: SourceLocation,
    ) -> TraceDecl {
        match target.ty() {
            CanonicalType::Array(ty) => {
                let elements = Interned::from_iter((0..ty.len()).map(|index| {
                    self.make_trace_decl_child(
                        instantiated_module,
                        match target {
                            MakeTraceDeclTarget::Expr(target) => MakeTraceDeclTarget::Expr(
                                Expr::<Array>::from_canonical(target)[index],
                            ),
                            MakeTraceDeclTarget::Memory {
                                id,
                                depth,
                                stride,
                                start,
                                ty: _,
                            } => MakeTraceDeclTarget::Memory {
                                id,
                                depth,
                                stride,
                                start: start + ty.element().bit_width() * index,
                                ty: ty.element(),
                            },
                        },
                        Intern::intern_owned(format!("[{index}]")),
                        source_location,
                    )
                }));
                TraceArray {
                    name,
                    elements,
                    ty,
                    flow: target.flow(),
                }
                .into()
            }
            CanonicalType::Enum(ty) => {
                if ty.variants().iter().all(|v| v.ty.is_none()) {
                    self.make_trace_scalar(instantiated_module, target, name, source_location)
                } else {
                    let flow = target.flow();
                    let location = match target {
                        MakeTraceDeclTarget::Expr(target) => {
                            let compiled_value = self.compile_expr(instantiated_module, target);
                            let compiled_value =
                                self.compiled_expr_to_value(compiled_value, source_location);
                            let discriminant = self.compile_enum_discriminant(
                                compiled_value.map_ty(Enum::from_canonical),
                                source_location,
                            );
                            TraceLocation::Scalar(self.new_sim_trace(
                                SimTraceKind::EnumDiscriminant {
                                    index: discriminant,
                                    ty,
                                },
                            ))
                        }
                        MakeTraceDeclTarget::Memory {
                            id,
                            depth,
                            stride,
                            start,
                            ty: _,
                        } => TraceLocation::Memory(TraceMemoryLocation {
                            id,
                            depth,
                            stride,
                            start,
                            len: ty.discriminant_bit_width(),
                        }),
                    };
                    let discriminant = TraceEnumDiscriminant {
                        location,
                        name: "$tag".intern(),
                        ty,
                        flow,
                    };
                    let non_empty_fields =
                        Interned::from_iter(ty.variants().into_iter().enumerate().flat_map(
                            |(variant_index, variant)| {
                                variant.ty.map(|variant_ty| {
                                    self.make_trace_decl_child(
                                        instantiated_module,
                                        match target {
                                            MakeTraceDeclTarget::Expr(target) => {
                                                MakeTraceDeclTarget::Expr(
                                                    ops::VariantAccess::new_by_index(
                                                        Expr::<Enum>::from_canonical(target),
                                                        variant_index,
                                                    )
                                                    .to_expr(),
                                                )
                                            }
                                            MakeTraceDeclTarget::Memory {
                                                id,
                                                depth,
                                                stride,
                                                start,
                                                ty: _,
                                            } => MakeTraceDeclTarget::Memory {
                                                id,
                                                depth,
                                                stride,
                                                start: start + ty.discriminant_bit_width(),
                                                ty: variant_ty,
                                            },
                                        },
                                        variant.name,
                                        source_location,
                                    )
                                })
                            },
                        ));
                    TraceEnumWithFields {
                        name,
                        discriminant,
                        non_empty_fields,
                        ty,
                        flow,
                    }
                    .into()
                }
            }
            CanonicalType::Bundle(ty) => {
                let fields = Interned::from_iter(ty.fields().iter().zip(ty.field_offsets()).map(
                    |(field, field_offset)| {
                        self.make_trace_decl_child(
                            instantiated_module,
                            match target {
                                MakeTraceDeclTarget::Expr(target) => {
                                    MakeTraceDeclTarget::Expr(Expr::field(
                                        Expr::<Bundle>::from_canonical(target),
                                        &field.name,
                                    ))
                                }
                                MakeTraceDeclTarget::Memory {
                                    id,
                                    depth,
                                    stride,
                                    start,
                                    ty: _,
                                } => MakeTraceDeclTarget::Memory {
                                    id,
                                    depth,
                                    stride,
                                    start: start + field_offset,
                                    ty: field.ty,
                                },
                            },
                            field.name,
                            source_location,
                        )
                    },
                ));
                TraceBundle {
                    name,
                    fields,
                    ty,
                    flow: target.flow(),
                }
                .into()
            }
            CanonicalType::UInt(_)
            | CanonicalType::SInt(_)
            | CanonicalType::Bool(_)
            | CanonicalType::AsyncReset(_)
            | CanonicalType::SyncReset(_)
            | CanonicalType::Reset(_)
            | CanonicalType::Clock(_) => {
                self.make_trace_scalar(instantiated_module, target, name, source_location)
            }
            CanonicalType::PhantomConst(_) => TraceBundle {
                name,
                fields: Interned::default(),
                ty: Bundle::new(Interned::default()),
                flow: target.flow(),
            }
            .into(),
        }
    }
    fn make_trace_decl(
        &mut self,
        instantiated_module: InstantiatedModule,
        target_base: TargetBase,
    ) -> TraceDecl {
        let target = MakeTraceDeclTarget::Expr(target_base.to_expr());
        match target_base {
            TargetBase::ModuleIO(module_io) => TraceModuleIO {
                name: module_io.name(),
                child: self
                    .make_trace_decl_child(
                        instantiated_module,
                        target,
                        module_io.name(),
                        module_io.source_location(),
                    )
                    .intern(),
                ty: module_io.ty(),
                flow: module_io.flow(),
            }
            .into(),
            TargetBase::MemPort(mem_port) => {
                let name = Intern::intern_owned(mem_port.port_name().to_string());
                let TraceDecl::Scope(TraceScope::Bundle(bundle)) = self.make_trace_decl_child(
                    instantiated_module,
                    target,
                    name,
                    mem_port.source_location(),
                ) else {
                    unreachable!()
                };
                TraceMemPort {
                    name,
                    bundle,
                    ty: mem_port.ty(),
                }
                .into()
            }
            TargetBase::Reg(reg) => TraceReg {
                name: reg.name(),
                child: self
                    .make_trace_decl_child(
                        instantiated_module,
                        target,
                        reg.name(),
                        reg.source_location(),
                    )
                    .intern(),
                ty: reg.ty(),
            }
            .into(),
            TargetBase::RegSync(reg) => TraceReg {
                name: reg.name(),
                child: self
                    .make_trace_decl_child(
                        instantiated_module,
                        target,
                        reg.name(),
                        reg.source_location(),
                    )
                    .intern(),
                ty: reg.ty(),
            }
            .into(),
            TargetBase::RegAsync(reg) => TraceReg {
                name: reg.name(),
                child: self
                    .make_trace_decl_child(
                        instantiated_module,
                        target,
                        reg.name(),
                        reg.source_location(),
                    )
                    .intern(),
                ty: reg.ty(),
            }
            .into(),
            TargetBase::Wire(wire) => TraceWire {
                name: wire.name(),
                child: self
                    .make_trace_decl_child(
                        instantiated_module,
                        target,
                        wire.name(),
                        wire.source_location(),
                    )
                    .intern(),
                ty: wire.ty(),
            }
            .into(),
            TargetBase::Instance(instance) => {
                let TraceDecl::Scope(TraceScope::Bundle(instance_io)) = self.make_trace_decl_child(
                    instantiated_module,
                    target,
                    instance.name(),
                    instance.source_location(),
                ) else {
                    unreachable!()
                };
                let compiled_module = &self.modules[&InstantiatedModule::Child {
                    parent: instantiated_module.intern(),
                    instance: instance.intern(),
                }];
                TraceInstance {
                    name: instance.name(),
                    instance_io,
                    module: compiled_module.trace_decls,
                    ty: instance.ty(),
                }
                .into()
            }
        }
    }
    fn compile_value(
        &mut self,
        target: TargetInInstantiatedModule,
    ) -> CompiledValue<CanonicalType> {
        if let Some(&retval) = self.compiled_values.get(&target) {
            return retval;
        }
        let retval = match target.target {
            Target::Base(base) => {
                let unprefixed_layout = CompiledTypeLayout::get(base.canonical_ty());
                let layout = unprefixed_layout.with_prefixed_debug_names(&format!(
                    "{:?}.{:?}",
                    target.instantiated_module,
                    base.target_name()
                ));
                let range = self.insns.allocate_variable(&layout.layout);
                let write = match *base {
                    TargetBase::ModuleIO(_)
                    | TargetBase::MemPort(_)
                    | TargetBase::Wire(_)
                    | TargetBase::Instance(_) => None,
                    TargetBase::Reg(_) | TargetBase::RegSync(_) | TargetBase::RegAsync(_) => {
                        let write_layout = unprefixed_layout.with_prefixed_debug_names(&format!(
                            "{:?}.{:?}$next",
                            target.instantiated_module,
                            base.target_name()
                        ));
                        Some((
                            write_layout,
                            self.insns.allocate_variable(&write_layout.layout),
                        ))
                    }
                };
                CompiledValue {
                    range,
                    layout,
                    write,
                }
            }
            Target::Child(target_child) => {
                let parent = self.compile_value(TargetInInstantiatedModule {
                    instantiated_module: target.instantiated_module,
                    target: *target_child.parent(),
                });
                match *target_child.path_element() {
                    TargetPathElement::BundleField(TargetPathBundleField { name }) => {
                        parent.map_ty(Bundle::from_canonical).field_by_name(name)
                    }
                    TargetPathElement::ArrayElement(TargetPathArrayElement { index }) => {
                        parent.map_ty(Array::from_canonical).element(index)
                    }
                    TargetPathElement::DynArrayElement(_) => unreachable!(),
                }
            }
        };
        self.compiled_values.insert(target, retval);
        retval
    }
    fn compiled_expr_to_value(
        &mut self,
        expr: CompiledExpr<CanonicalType>,
        source_location: SourceLocation,
    ) -> CompiledValue<CanonicalType> {
        if let Some(&retval) = self.compiled_exprs_to_values.get(&expr) {
            return retval;
        }
        assert!(
            expr.static_part.layout.ty.is_passive(),
            "invalid expression passed to compiled_expr_to_value -- type must be passive",
        );
        let CompiledExpr {
            static_part,
            indexes,
        } = expr;
        let retval = if indexes.as_ref().is_empty() {
            CompiledValue {
                layout: static_part.layout,
                range: static_part.range,
                write: None,
            }
        } else {
            let layout = static_part.layout.with_anonymized_debug_info();
            let retval = CompiledValue {
                layout,
                range: self.insns.allocate_variable(&layout.layout),
                write: None,
            };
            let TypeIndexRange {
                small_slots,
                big_slots,
            } = retval.range;
            self.add_assignment(
                Interned::default(),
                small_slots
                    .iter()
                    .zip(static_part.range.small_slots.iter())
                    .map(|(dest, base)| Insn::ReadSmallIndexed {
                        dest,
                        src: StatePartArrayIndexed {
                            base,
                            indexes: indexes.small_slots,
                        },
                    })
                    .chain(
                        big_slots
                            .iter()
                            .zip(static_part.range.big_slots.iter())
                            .map(|(dest, base)| Insn::ReadIndexed {
                                dest,
                                src: StatePartArrayIndexed {
                                    base,
                                    indexes: indexes.big_slots,
                                },
                            }),
                    ),
                source_location,
            );
            retval
        };
        self.compiled_exprs_to_values.insert(expr, retval);
        retval
    }
    fn add_assignment<T: Into<InsnOrLabel>>(
        &mut self,
        conditions: Interned<[Cond]>,
        insns: impl IntoIterator<Item = T>,
        source_location: SourceLocation,
    ) {
        let insns = Vec::from_iter(insns.into_iter().map(Into::into));
        self.assignments
            .push(Assignment::new(conditions, insns, source_location));
    }
    fn simple_big_expr_input(
        &mut self,
        instantiated_module: InstantiatedModule,
        input: Expr<CanonicalType>,
    ) -> StatePartIndex<StatePartKindBigSlots> {
        let input = self.compile_expr(instantiated_module, input);
        let input =
            self.compiled_expr_to_value(input, instantiated_module.leaf_module().source_location());
        assert_eq!(input.range.len(), TypeLen::A_BIG_SLOT);
        input.range.big_slots.start
    }
    fn compile_expr_helper(
        &mut self,
        instantiated_module: InstantiatedModule,
        dest_ty: CanonicalType,
        make_insns: impl FnOnce(&mut Self, TypeIndexRange) -> Vec<Insn>,
    ) -> CompiledValue<CanonicalType> {
        let layout = CompiledTypeLayout::get(dest_ty);
        let range = self.insns.allocate_variable(&layout.layout);
        let retval = CompiledValue {
            layout,
            range,
            write: None,
        };
        let insns = make_insns(self, range);
        self.add_assignment(
            Interned::default(),
            insns,
            instantiated_module.leaf_module().source_location(),
        );
        retval
    }
    fn simple_nary_big_expr_helper(
        &mut self,
        instantiated_module: InstantiatedModule,
        dest_ty: CanonicalType,
        make_insns: impl FnOnce(StatePartIndex<StatePartKindBigSlots>) -> Vec<Insn>,
    ) -> CompiledValue<CanonicalType> {
        self.compile_expr_helper(instantiated_module, dest_ty, |_, dest| {
            assert_eq!(dest.len(), TypeLen::A_BIG_SLOT);
            make_insns(dest.big_slots.start)
        })
    }
    fn simple_nary_big_expr<const N: usize>(
        &mut self,
        instantiated_module: InstantiatedModule,
        dest_ty: CanonicalType,
        inputs: [Expr<CanonicalType>; N],
        make_insns: impl FnOnce(
            StatePartIndex<StatePartKindBigSlots>,
            [StatePartIndex<StatePartKindBigSlots>; N],
        ) -> Vec<Insn>,
    ) -> CompiledValue<CanonicalType> {
        let inputs = inputs.map(|input| self.simple_big_expr_input(instantiated_module, input));
        self.simple_nary_big_expr_helper(instantiated_module, dest_ty, |dest| {
            make_insns(dest, inputs)
        })
    }
    fn compiled_value_to_dyn_array_index(
        &mut self,
        compiled_value: CompiledValue<UInt>,
        source_location: SourceLocation,
    ) -> StatePartIndex<StatePartKindSmallSlots> {
        if let Some(&retval) = self
            .compiled_values_to_dyn_array_indexes
            .get(&compiled_value)
        {
            return retval;
        }
        let mut ty = compiled_value.layout.ty;
        ty.width = ty.width.min(SmallUInt::BITS as usize);
        let retval = match compiled_value.range.len() {
            TypeLen::A_SMALL_SLOT => compiled_value.range.small_slots.start,
            TypeLen::A_BIG_SLOT => {
                let debug_data = SlotDebugData {
                    name: Interned::default(),
                    ty: ty.canonical(),
                };
                let dest = self
                    .insns
                    .allocate_variable(&TypeLayout {
                        small_slots: StatePartLayout::scalar(debug_data, ()),
                        big_slots: StatePartLayout::empty(),
                    })
                    .small_slots
                    .start;
                self.add_assignment(
                    Interned::default(),
                    vec![Insn::CastBigToArrayIndex {
                        dest,
                        src: compiled_value.range.big_slots.start,
                    }],
                    source_location,
                );
                dest
            }
            _ => unreachable!(),
        };
        self.compiled_values_to_dyn_array_indexes
            .insert(compiled_value, retval);
        retval
    }
    fn compiled_value_bool_dest_is_small(
        &mut self,
        compiled_value: CompiledValue<CanonicalType>,
        source_location: SourceLocation,
    ) -> StatePartIndex<StatePartKindSmallSlots> {
        if let Some(&retval) = self
            .compiled_value_bool_dest_is_small_map
            .get(&compiled_value)
        {
            return retval;
        }
        let retval = match compiled_value.range.len() {
            TypeLen::A_SMALL_SLOT => compiled_value.range.small_slots.start,
            TypeLen::A_BIG_SLOT => {
                let debug_data = SlotDebugData {
                    name: Interned::default(),
                    ty: Bool.canonical(),
                };
                let dest = self
                    .insns
                    .allocate_variable(&TypeLayout {
                        small_slots: StatePartLayout::scalar(debug_data, ()),
                        big_slots: StatePartLayout::empty(),
                    })
                    .small_slots
                    .start;
                self.add_assignment(
                    Interned::default(),
                    vec![Insn::IsNonZeroDestIsSmall {
                        dest,
                        src: compiled_value.range.big_slots.start,
                    }],
                    source_location,
                );
                dest
            }
            _ => unreachable!(),
        };
        self.compiled_value_bool_dest_is_small_map
            .insert(compiled_value, retval);
        retval
    }
    fn compile_cast_scalar_to_bits<T: Type>(
        &mut self,
        instantiated_module: InstantiatedModule,
        arg: Expr<CanonicalType>,
        cast_fn: impl FnOnce(Expr<T>) -> Expr<UInt>,
    ) -> CompiledValue<UInt> {
        let arg = Expr::<T>::from_canonical(arg);
        let retval = cast_fn(arg);
        let retval = self.compile_expr(instantiated_module, Expr::canonical(retval));
        let retval = self
            .compiled_expr_to_value(retval, instantiated_module.leaf_module().source_location());
        retval.map_ty(UInt::from_canonical)
    }
    fn compile_cast_aggregate_to_bits(
        &mut self,
        instantiated_module: InstantiatedModule,
        parts: impl IntoIterator<Item = Expr<CanonicalType>>,
    ) -> CompiledValue<UInt> {
        let retval = parts
            .into_iter()
            .map(|part| part.cast_to_bits())
            .reduce(|accumulator, part| accumulator | (part << Expr::ty(accumulator).width))
            .unwrap_or_else(|| UInt[0].zero().to_expr());
        let retval = self.compile_expr(instantiated_module, Expr::canonical(retval));
        let retval = self
            .compiled_expr_to_value(retval, instantiated_module.leaf_module().source_location());
        retval.map_ty(UInt::from_canonical)
    }
    fn compile_cast_to_bits(
        &mut self,
        instantiated_module: InstantiatedModule,
        expr: ops::CastToBits,
    ) -> CompiledValue<UInt> {
        match Expr::ty(expr.arg()) {
            CanonicalType::UInt(_) => {
                self.compile_cast_scalar_to_bits(instantiated_module, expr.arg(), |arg| arg)
            }
            CanonicalType::SInt(ty) => self.compile_cast_scalar_to_bits(
                instantiated_module,
                expr.arg(),
                |arg: Expr<SInt>| arg.cast_to(ty.as_same_width_uint()),
            ),
            CanonicalType::Bool(_)
            | CanonicalType::AsyncReset(_)
            | CanonicalType::SyncReset(_)
            | CanonicalType::Reset(_)
            | CanonicalType::Clock(_) => self.compile_cast_scalar_to_bits(
                instantiated_module,
                expr.arg(),
                |arg: Expr<Bool>| arg.cast_to(UInt[1]),
            ),
            CanonicalType::Array(ty) => self.compile_cast_aggregate_to_bits(
                instantiated_module,
                (0..ty.len()).map(|index| Expr::<Array>::from_canonical(expr.arg())[index]),
            ),
            CanonicalType::Enum(ty) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    UInt[ty.type_properties().bit_width].canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| vec![Insn::Copy { dest, src }],
                )
                .map_ty(UInt::from_canonical),
            CanonicalType::Bundle(ty) => self.compile_cast_aggregate_to_bits(
                instantiated_module,
                ty.fields().iter().map(|field| {
                    Expr::field(Expr::<Bundle>::from_canonical(expr.arg()), &field.name)
                }),
            ),
            CanonicalType::PhantomConst(_) => {
                self.compile_cast_aggregate_to_bits(instantiated_module, [])
            }
        }
    }
    fn compile_cast_bits_to(
        &mut self,
        instantiated_module: InstantiatedModule,
        expr: ops::CastBitsTo,
    ) -> CompiledValue<CanonicalType> {
        let retval = match expr.ty() {
            CanonicalType::UInt(_) => Expr::canonical(expr.arg()),
            CanonicalType::SInt(ty) => Expr::canonical(expr.arg().cast_to(ty)),
            CanonicalType::Bool(ty) => Expr::canonical(expr.arg().cast_to(ty)),
            CanonicalType::Array(ty) => {
                let stride = ty.element().bit_width();
                Expr::<Array>::canonical(
                    ops::ArrayLiteral::new(
                        ty.element(),
                        Interned::from_iter((0..ty.len()).map(|index| {
                            let start = stride * index;
                            let end = start + stride;
                            expr.arg()[start..end].cast_bits_to(ty.element())
                        })),
                    )
                    .to_expr(),
                )
            }
            ty @ CanonicalType::Enum(_) => {
                return self.simple_nary_big_expr(
                    instantiated_module,
                    ty,
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| vec![Insn::Copy { dest, src }],
                );
            }
            CanonicalType::Bundle(ty) => Expr::canonical(
                ops::BundleLiteral::new(
                    ty,
                    Interned::from_iter(ty.field_offsets().iter().zip(&ty.fields()).map(
                        |(&offset, &field)| {
                            let end = offset + field.ty.bit_width();
                            expr.arg()[offset..end].cast_bits_to(field.ty)
                        },
                    )),
                )
                .to_expr(),
            ),
            CanonicalType::AsyncReset(ty) => Expr::canonical(expr.arg().cast_to(ty)),
            CanonicalType::SyncReset(ty) => Expr::canonical(expr.arg().cast_to(ty)),
            CanonicalType::Reset(_) => unreachable!(),
            CanonicalType::Clock(ty) => Expr::canonical(expr.arg().cast_to(ty)),
            CanonicalType::PhantomConst(ty) => {
                let _ = self.compile_expr(instantiated_module, Expr::canonical(expr.arg()));
                Expr::canonical(ty.to_expr())
            }
        };
        let retval = self.compile_expr(instantiated_module, Expr::canonical(retval));
        self.compiled_expr_to_value(retval, instantiated_module.leaf_module().source_location())
    }
    fn compile_aggregate_literal(
        &mut self,
        instantiated_module: InstantiatedModule,
        dest_ty: CanonicalType,
        inputs: Interned<[Expr<CanonicalType>]>,
    ) -> CompiledValue<CanonicalType> {
        self.compile_expr_helper(instantiated_module, dest_ty, |this, dest| {
            let mut insns = Vec::new();
            let mut offset = TypeIndex::ZERO;
            for input in inputs {
                let input = this.compile_expr(instantiated_module, input);
                let input = this
                    .compiled_expr_to_value(
                        input,
                        instantiated_module.leaf_module().source_location(),
                    )
                    .range;
                insns.extend(
                    input.insns_for_copy_to(dest.slice(TypeIndexRange::new(offset, input.len()))),
                );
                offset = offset.offset(input.len().as_index());
            }
            insns
        })
    }
    fn compile_expr(
        &mut self,
        instantiated_module: InstantiatedModule,
        expr: Expr<CanonicalType>,
    ) -> CompiledExpr<CanonicalType> {
        if let Some(&retval) = self.compiled_exprs.get(&expr) {
            return retval;
        }
        let mut cast_bit = |arg: Expr<CanonicalType>| {
            let src_signed = match Expr::ty(arg) {
                CanonicalType::UInt(_) => false,
                CanonicalType::SInt(_) => true,
                CanonicalType::Bool(_) => false,
                CanonicalType::Array(_) => unreachable!(),
                CanonicalType::Enum(_) => unreachable!(),
                CanonicalType::Bundle(_) => unreachable!(),
                CanonicalType::AsyncReset(_) => false,
                CanonicalType::SyncReset(_) => false,
                CanonicalType::Reset(_) => false,
                CanonicalType::Clock(_) => false,
                CanonicalType::PhantomConst(_) => unreachable!(),
            };
            let dest_signed = match Expr::ty(expr) {
                CanonicalType::UInt(_) => false,
                CanonicalType::SInt(_) => true,
                CanonicalType::Bool(_) => false,
                CanonicalType::Array(_) => unreachable!(),
                CanonicalType::Enum(_) => unreachable!(),
                CanonicalType::Bundle(_) => unreachable!(),
                CanonicalType::AsyncReset(_) => false,
                CanonicalType::SyncReset(_) => false,
                CanonicalType::Reset(_) => false,
                CanonicalType::Clock(_) => false,
                CanonicalType::PhantomConst(_) => unreachable!(),
            };
            self.simple_nary_big_expr(instantiated_module, Expr::ty(expr), [arg], |dest, [src]| {
                match (src_signed, dest_signed) {
                    (false, false) | (true, true) => {
                        vec![Insn::Copy { dest, src }]
                    }
                    (false, true) => vec![Insn::CastToSInt {
                        dest,
                        src,
                        dest_width: 1,
                    }],
                    (true, false) => vec![Insn::CastToUInt {
                        dest,
                        src,
                        dest_width: 1,
                    }],
                }
            })
            .into()
        };
        let retval: CompiledExpr<_> = match *Expr::expr_enum(expr) {
            ExprEnum::UIntLiteral(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [],
                    |dest, []| {
                        vec![Insn::Const {
                            dest,
                            value: expr.to_bigint().intern_sized(),
                        }]
                    },
                )
                .into(),
            ExprEnum::SIntLiteral(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [],
                    |dest, []| {
                        vec![Insn::Const {
                            dest,
                            value: expr.to_bigint().intern_sized(),
                        }]
                    },
                )
                .into(),
            ExprEnum::BoolLiteral(expr) => self
                .simple_nary_big_expr(instantiated_module, Bool.canonical(), [], |dest, []| {
                    vec![Insn::Const {
                        dest,
                        value: BigInt::from(expr).intern_sized(),
                    }]
                })
                .into(),
            ExprEnum::PhantomConst(_) => self
                .compile_aggregate_literal(instantiated_module, Expr::ty(expr), Interned::default())
                .into(),
            ExprEnum::BundleLiteral(literal) => self
                .compile_aggregate_literal(
                    instantiated_module,
                    Expr::ty(expr),
                    literal.field_values(),
                )
                .into(),
            ExprEnum::ArrayLiteral(literal) => self
                .compile_aggregate_literal(
                    instantiated_module,
                    Expr::ty(expr),
                    literal.element_values(),
                )
                .into(),
            ExprEnum::EnumLiteral(expr) => {
                let enum_bits_ty = UInt[expr.ty().type_properties().bit_width];
                let enum_bits = if let Some(variant_value) = expr.variant_value() {
                    (
                        UInt[expr.ty().discriminant_bit_width()]
                            .from_int_wrapping(expr.variant_index()),
                        variant_value,
                    )
                        .cast_to_bits()
                        .cast_to(enum_bits_ty)
                } else {
                    enum_bits_ty
                        .from_int_wrapping(expr.variant_index())
                        .to_expr()
                };
                self.compile_expr(
                    instantiated_module,
                    enum_bits.cast_bits_to(expr.ty().canonical()),
                )
            }
            ExprEnum::Uninit(expr) => self.compile_expr(
                instantiated_module,
                UInt[expr.ty().bit_width()].zero().cast_bits_to(expr.ty()),
            ),
            ExprEnum::NotU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Expr::ty(expr.arg()).canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| {
                        vec![Insn::NotU {
                            dest,
                            src,
                            width: Expr::ty(expr.arg()).width(),
                        }]
                    },
                )
                .into(),
            ExprEnum::NotS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Expr::ty(expr.arg()).canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| vec![Insn::NotS { dest, src }],
                )
                .into(),
            ExprEnum::NotB(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Expr::ty(expr.arg()).canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| {
                        vec![Insn::NotU {
                            dest,
                            src,
                            width: 1,
                        }]
                    },
                )
                .into(),
            ExprEnum::Neg(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| vec![Insn::Neg { dest, src }],
                )
                .into(),
            ExprEnum::BitAndU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::And { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitAndS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::And { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitAndB(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::And { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitOrU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Or { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitOrS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Or { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitOrB(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Or { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitXorU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Xor { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitXorS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Xor { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitXorB(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Xor { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::AddU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Add { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::AddS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Add { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::SubU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| {
                        vec![Insn::SubU {
                            dest,
                            lhs,
                            rhs,
                            dest_width: expr.ty().width(),
                        }]
                    },
                )
                .into(),
            ExprEnum::SubS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::SubS { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::MulU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Mul { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::MulS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Mul { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::DivU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Div { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::DivS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Div { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::RemU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Rem { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::RemS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Rem { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::DynShlU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::DynShl { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::DynShlS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::DynShl { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::DynShrU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::DynShr { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::DynShrS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::DynShr { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::FixedShlU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs())],
                    |dest, [lhs]| {
                        vec![Insn::Shl {
                            dest,
                            lhs,
                            rhs: expr.rhs(),
                        }]
                    },
                )
                .into(),
            ExprEnum::FixedShlS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs())],
                    |dest, [lhs]| {
                        vec![Insn::Shl {
                            dest,
                            lhs,
                            rhs: expr.rhs(),
                        }]
                    },
                )
                .into(),
            ExprEnum::FixedShrU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs())],
                    |dest, [lhs]| {
                        vec![Insn::Shr {
                            dest,
                            lhs,
                            rhs: expr.rhs(),
                        }]
                    },
                )
                .into(),
            ExprEnum::FixedShrS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs())],
                    |dest, [lhs]| {
                        vec![Insn::Shr {
                            dest,
                            lhs,
                            rhs: expr.rhs(),
                        }]
                    },
                )
                .into(),
            ExprEnum::CmpLtB(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLt { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpLeB(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpGtB(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    // swap both comparison direction and lhs/rhs
                    [Expr::canonical(expr.rhs()), Expr::canonical(expr.lhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLt { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpGeB(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    // swap both comparison direction and lhs/rhs
                    [Expr::canonical(expr.rhs()), Expr::canonical(expr.lhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpEqB(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpEq { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpNeB(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpNe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpLtU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLt { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpLeU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpGtU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    // swap both comparison direction and lhs/rhs
                    [Expr::canonical(expr.rhs()), Expr::canonical(expr.lhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLt { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpGeU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    // swap both comparison direction and lhs/rhs
                    [Expr::canonical(expr.rhs()), Expr::canonical(expr.lhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpEqU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpEq { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpNeU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpNe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpLtS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLt { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpLeS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpGtS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    // swap both comparison direction and lhs/rhs
                    [Expr::canonical(expr.rhs()), Expr::canonical(expr.lhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLt { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpGeS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    // swap both comparison direction and lhs/rhs
                    [Expr::canonical(expr.rhs()), Expr::canonical(expr.lhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpEqS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpEq { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpNeS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpNe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CastUIntToUInt(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| {
                        vec![Insn::CastToUInt {
                            dest,
                            src,
                            dest_width: expr.ty().width(),
                        }]
                    },
                )
                .into(),
            ExprEnum::CastUIntToSInt(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| {
                        vec![Insn::CastToSInt {
                            dest,
                            src,
                            dest_width: expr.ty().width(),
                        }]
                    },
                )
                .into(),
            ExprEnum::CastSIntToUInt(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| {
                        vec![Insn::CastToUInt {
                            dest,
                            src,
                            dest_width: expr.ty().width(),
                        }]
                    },
                )
                .into(),
            ExprEnum::CastSIntToSInt(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| {
                        vec![Insn::CastToSInt {
                            dest,
                            src,
                            dest_width: expr.ty().width(),
                        }]
                    },
                )
                .into(),
            ExprEnum::CastBoolToUInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastBoolToSInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastUIntToBool(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSIntToBool(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastBoolToSyncReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastUIntToSyncReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSIntToSyncReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastBoolToAsyncReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastUIntToAsyncReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSIntToAsyncReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSyncResetToBool(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSyncResetToUInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSyncResetToSInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSyncResetToReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastAsyncResetToBool(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastAsyncResetToUInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastAsyncResetToSInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastAsyncResetToReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastResetToBool(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastResetToUInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastResetToSInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastBoolToClock(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastUIntToClock(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSIntToClock(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastClockToBool(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastClockToUInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastClockToSInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::FieldAccess(expr) => self
                .compile_expr(instantiated_module, Expr::canonical(expr.base()))
                .map_ty(Bundle::from_canonical)
                .field_by_index(expr.field_index()),
            ExprEnum::VariantAccess(variant_access) => {
                let start = Expr::ty(variant_access.base()).discriminant_bit_width();
                let len = Expr::ty(expr).bit_width();
                self.compile_expr(
                    instantiated_module,
                    variant_access.base().cast_to_bits()[start..start + len]
                        .cast_bits_to(Expr::ty(expr)),
                )
            }
            ExprEnum::ArrayIndex(expr) => self
                .compile_expr(instantiated_module, Expr::canonical(expr.base()))
                .map_ty(Array::from_canonical)
                .element(expr.element_index()),
            ExprEnum::DynArrayIndex(expr) => {
                let element_index =
                    self.compile_expr(instantiated_module, Expr::canonical(expr.element_index()));
                let element_index = self.compiled_expr_to_value(
                    element_index,
                    instantiated_module.leaf_module().source_location(),
                );
                let index_slot = self.compiled_value_to_dyn_array_index(
                    element_index.map_ty(UInt::from_canonical),
                    instantiated_module.leaf_module().source_location(),
                );
                self.compile_expr(instantiated_module, Expr::canonical(expr.base()))
                    .map_ty(Array::from_canonical)
                    .element_dyn(index_slot)
            }
            ExprEnum::ReduceBitAndU(expr) => if Expr::ty(expr.arg()).width() == 0 {
                self.compile_expr(instantiated_module, Expr::canonical(true.to_expr()))
            } else {
                self.compile_expr(
                    instantiated_module,
                    Expr::canonical(
                        expr.arg()
                            .cmp_eq(Expr::ty(expr.arg()).from_int_wrapping(-1)),
                    ),
                )
            }
            .into(),
            ExprEnum::ReduceBitAndS(expr) => if Expr::ty(expr.arg()).width() == 0 {
                self.compile_expr(instantiated_module, Expr::canonical(true.to_expr()))
            } else {
                self.compile_expr(
                    instantiated_module,
                    Expr::canonical(
                        expr.arg()
                            .cmp_eq(Expr::ty(expr.arg()).from_int_wrapping(-1)),
                    ),
                )
            }
            .into(),
            ExprEnum::ReduceBitOrU(expr) => if Expr::ty(expr.arg()).width() == 0 {
                self.compile_expr(instantiated_module, Expr::canonical(false.to_expr()))
            } else {
                self.compile_expr(
                    instantiated_module,
                    Expr::canonical(expr.arg().cmp_ne(Expr::ty(expr.arg()).from_int_wrapping(0))),
                )
            }
            .into(),
            ExprEnum::ReduceBitOrS(expr) => if Expr::ty(expr.arg()).width() == 0 {
                self.compile_expr(instantiated_module, Expr::canonical(false.to_expr()))
            } else {
                self.compile_expr(
                    instantiated_module,
                    Expr::canonical(expr.arg().cmp_ne(Expr::ty(expr.arg()).from_int_wrapping(0))),
                )
            }
            .into(),
            ExprEnum::ReduceBitXorU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    UInt::<1>::TYPE.canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| {
                        vec![Insn::ReduceBitXor {
                            dest,
                            src,
                            input_width: Expr::ty(expr.arg()).width(),
                        }]
                    },
                )
                .into(),
            ExprEnum::ReduceBitXorS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    UInt::<1>::TYPE.canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| {
                        vec![Insn::ReduceBitXor {
                            dest,
                            src,
                            input_width: Expr::ty(expr.arg()).width(),
                        }]
                    },
                )
                .into(),
            ExprEnum::SliceUInt(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    UInt::new_dyn(expr.range().len()).canonical(),
                    [Expr::canonical(expr.base())],
                    |dest, [src]| {
                        vec![Insn::SliceInt {
                            dest,
                            src,
                            start: expr.range().start,
                            len: expr.range().len(),
                        }]
                    },
                )
                .into(),
            ExprEnum::SliceSInt(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    UInt::new_dyn(expr.range().len()).canonical(),
                    [Expr::canonical(expr.base())],
                    |dest, [src]| {
                        vec![Insn::SliceInt {
                            dest,
                            src,
                            start: expr.range().start,
                            len: expr.range().len(),
                        }]
                    },
                )
                .into(),
            ExprEnum::CastToBits(expr) => self
                .compile_cast_to_bits(instantiated_module, expr)
                .map_ty(CanonicalType::UInt)
                .into(),
            ExprEnum::CastBitsTo(expr) => {
                self.compile_cast_bits_to(instantiated_module, expr).into()
            }
            ExprEnum::ModuleIO(expr) => self
                .compile_value(TargetInInstantiatedModule {
                    instantiated_module,
                    target: expr.into(),
                })
                .into(),
            ExprEnum::Instance(expr) => self
                .compile_value(TargetInInstantiatedModule {
                    instantiated_module,
                    target: expr.into(),
                })
                .into(),
            ExprEnum::Wire(expr) => self
                .compile_value(TargetInInstantiatedModule {
                    instantiated_module,
                    target: expr.into(),
                })
                .into(),
            ExprEnum::Reg(expr) => self
                .compile_value(TargetInInstantiatedModule {
                    instantiated_module,
                    target: expr.into(),
                })
                .into(),
            ExprEnum::RegSync(expr) => self
                .compile_value(TargetInInstantiatedModule {
                    instantiated_module,
                    target: expr.into(),
                })
                .into(),
            ExprEnum::RegAsync(expr) => self
                .compile_value(TargetInInstantiatedModule {
                    instantiated_module,
                    target: expr.into(),
                })
                .into(),
            ExprEnum::MemPort(expr) => self
                .compile_value(TargetInInstantiatedModule {
                    instantiated_module,
                    target: expr.into(),
                })
                .into(),
        };
        self.compiled_exprs.insert(expr, retval);
        retval
    }
    fn compile_simple_connect(
        &mut self,
        conditions: Interned<[Cond]>,
        lhs: CompiledExpr<CanonicalType>,
        rhs: CompiledValue<CanonicalType>,
        source_location: SourceLocation,
    ) {
        let CompiledExpr {
            static_part: lhs_static_part,
            indexes,
        } = lhs;
        let (lhs_layout, lhs_range) = lhs_static_part.write();
        assert!(
            lhs_layout.ty.is_passive(),
            "invalid expression passed to compile_simple_connect -- type must be passive",
        );
        let TypeIndexRange {
            small_slots,
            big_slots,
        } = lhs_range;
        self.add_assignment(
            conditions,
            small_slots
                .iter()
                .zip(rhs.range.small_slots.iter())
                .map(|(base, src)| {
                    if indexes.small_slots.is_empty() {
                        Insn::CopySmall { dest: base, src }
                    } else {
                        Insn::WriteSmallIndexed {
                            dest: StatePartArrayIndexed {
                                base,
                                indexes: indexes.small_slots,
                            },
                            src,
                        }
                    }
                })
                .chain(
                    big_slots
                        .iter()
                        .zip(rhs.range.big_slots.iter())
                        .map(|(base, src)| {
                            if indexes.big_slots.is_empty() {
                                Insn::Copy { dest: base, src }
                            } else {
                                Insn::WriteIndexed {
                                    dest: StatePartArrayIndexed {
                                        base,
                                        indexes: indexes.big_slots,
                                    },
                                    src,
                                }
                            }
                        }),
                ),
            source_location,
        );
    }
    fn compile_connect(
        &mut self,
        lhs_instantiated_module: InstantiatedModule,
        lhs_conditions: Interned<[Cond]>,
        lhs: Expr<CanonicalType>,
        rhs_instantiated_module: InstantiatedModule,
        rhs_conditions: Interned<[Cond]>,
        mut rhs: Expr<CanonicalType>,
        source_location: SourceLocation,
    ) {
        if Expr::ty(lhs) != Expr::ty(rhs) || !Expr::ty(lhs).is_passive() {
            match Expr::ty(lhs) {
                CanonicalType::UInt(lhs_ty) => {
                    rhs = Expr::canonical(Expr::<UInt>::from_canonical(rhs).cast_to(lhs_ty));
                }
                CanonicalType::SInt(lhs_ty) => {
                    rhs = Expr::canonical(Expr::<SInt>::from_canonical(rhs).cast_to(lhs_ty));
                }
                CanonicalType::Bool(_) => unreachable!(),
                CanonicalType::Array(lhs_ty) => {
                    let CanonicalType::Array(rhs_ty) = Expr::ty(rhs) else {
                        unreachable!();
                    };
                    assert_eq!(lhs_ty.len(), rhs_ty.len());
                    let lhs = Expr::<Array>::from_canonical(lhs);
                    let rhs = Expr::<Array>::from_canonical(rhs);
                    for index in 0..lhs_ty.len() {
                        self.compile_connect(
                            lhs_instantiated_module,
                            lhs_conditions,
                            lhs[index],
                            rhs_instantiated_module,
                            rhs_conditions,
                            rhs[index],
                            source_location,
                        );
                    }
                    return;
                }
                CanonicalType::Enum(lhs_ty) => {
                    let CanonicalType::Enum(rhs_ty) = Expr::ty(rhs) else {
                        unreachable!();
                    };
                    todo!("handle connect with different enum types");
                }
                CanonicalType::Bundle(lhs_ty) => {
                    let CanonicalType::Bundle(rhs_ty) = Expr::ty(rhs) else {
                        unreachable!();
                    };
                    assert_eq!(lhs_ty.fields().len(), rhs_ty.fields().len());
                    let lhs = Expr::<Bundle>::from_canonical(lhs);
                    let rhs = Expr::<Bundle>::from_canonical(rhs);
                    for (
                        field_index,
                        (
                            BundleField {
                                name,
                                flipped,
                                ty: _,
                            },
                            rhs_field,
                        ),
                    ) in lhs_ty.fields().into_iter().zip(rhs_ty.fields()).enumerate()
                    {
                        assert_eq!(name, rhs_field.name);
                        assert_eq!(flipped, rhs_field.flipped);
                        let lhs_expr = ops::FieldAccess::new_by_index(lhs, field_index).to_expr();
                        let rhs_expr = ops::FieldAccess::new_by_index(rhs, field_index).to_expr();
                        if flipped {
                            // swap lhs/rhs
                            self.compile_connect(
                                rhs_instantiated_module,
                                rhs_conditions,
                                rhs_expr,
                                lhs_instantiated_module,
                                lhs_conditions,
                                lhs_expr,
                                source_location,
                            );
                        } else {
                            self.compile_connect(
                                lhs_instantiated_module,
                                lhs_conditions,
                                lhs_expr,
                                rhs_instantiated_module,
                                rhs_conditions,
                                rhs_expr,
                                source_location,
                            );
                        }
                    }
                    return;
                }
                CanonicalType::AsyncReset(_) => unreachable!(),
                CanonicalType::SyncReset(_) => unreachable!(),
                CanonicalType::Reset(_) => unreachable!(),
                CanonicalType::Clock(_) => unreachable!(),
                CanonicalType::PhantomConst(_) => unreachable!("PhantomConst mismatch"),
            }
        }
        let Some(target) = lhs.target() else {
            unreachable!("connect lhs must have target");
        };
        let lhs_decl_conditions = self.decl_conditions[&TargetInInstantiatedModule {
            instantiated_module: lhs_instantiated_module,
            target: target.base().into(),
        }];
        let lhs = self.compile_expr(lhs_instantiated_module, lhs);
        let rhs = self.compile_expr(rhs_instantiated_module, rhs);
        let rhs = self.compiled_expr_to_value(rhs, source_location);
        self.compile_simple_connect(
            lhs_conditions[lhs_decl_conditions.len()..].intern(),
            lhs,
            rhs,
            source_location,
        );
    }
    fn compile_clock(
        &mut self,
        clk: CompiledValue<Clock>,
        source_location: SourceLocation,
    ) -> ClockTrigger {
        if let Some(&retval) = self.compiled_value_to_clock_trigger_map.get(&clk) {
            return retval;
        }
        let mut alloc_small_slot = |part_name: &str| {
            self.insns
                .state_layout
                .ty
                .small_slots
                .allocate(&StatePartLayout::scalar(
                    SlotDebugData {
                        name: Interned::default(),
                        ty: Bool.canonical(),
                    },
                    (),
                ))
                .start
        };
        let last_clk_was_low = alloc_small_slot("last_clk_was_low");
        let clk_triggered = alloc_small_slot("clk_triggered");
        let retval = ClockTrigger {
            last_clk_was_low,
            clk: self.compiled_value_bool_dest_is_small(
                clk.map_ty(CanonicalType::Clock),
                source_location,
            ),
            clk_triggered,
            source_location,
        };
        self.add_assignment(
            Interned::default(),
            [Insn::AndSmall {
                dest: clk_triggered,
                lhs: retval.clk,
                rhs: last_clk_was_low,
            }],
            source_location,
        );
        self.clock_triggers.push(retval);
        self.compiled_value_to_clock_trigger_map.insert(clk, retval);
        retval
    }
    fn compile_enum_discriminant(
        &mut self,
        enum_value: CompiledValue<Enum>,
        source_location: SourceLocation,
    ) -> StatePartIndex<StatePartKindSmallSlots> {
        if let Some(&retval) = self.enum_discriminants.get(&enum_value) {
            return retval;
        }
        let retval_ty = Enum::new(
            enum_value
                .layout
                .ty
                .variants()
                .iter()
                .map(|variant| EnumVariant {
                    name: variant.name,
                    ty: None,
                })
                .collect(),
        );
        let retval = if retval_ty == enum_value.layout.ty
            && enum_value.range.len() == TypeLen::A_SMALL_SLOT
        {
            enum_value.range.small_slots.start
        } else {
            let retval = self
                .insns
                .state_layout
                .ty
                .small_slots
                .allocate(&StatePartLayout::scalar(
                    SlotDebugData {
                        name: Interned::default(),
                        ty: retval_ty.canonical(),
                    },
                    (),
                ))
                .start;
            let discriminant_bit_width = enum_value.layout.ty.discriminant_bit_width();
            let discriminant_mask = !(!0u64 << discriminant_bit_width);
            let insn = match enum_value.range.len() {
                TypeLen::A_BIG_SLOT => Insn::AndBigWithSmallImmediate {
                    dest: retval,
                    lhs: enum_value.range.big_slots.start,
                    rhs: discriminant_mask,
                },
                TypeLen::A_SMALL_SLOT => {
                    if discriminant_bit_width == enum_value.layout.ty.type_properties().bit_width {
                        Insn::CopySmall {
                            dest: retval,
                            src: enum_value.range.small_slots.start,
                        }
                    } else {
                        Insn::AndSmallImmediate {
                            dest: retval,
                            lhs: enum_value.range.small_slots.start,
                            rhs: discriminant_mask,
                        }
                    }
                }
                _ => unreachable!(),
            };
            self.add_assignment(Interned::default(), [insn], source_location);
            retval
        };
        self.enum_discriminants.insert(enum_value, retval);
        retval
    }
    fn compile_stmt_reg<R: ResetType>(
        &mut self,
        stmt_reg: StmtReg<R>,
        instantiated_module: InstantiatedModule,
        value: CompiledValue<CanonicalType>,
    ) {
        let StmtReg { annotations, reg } = stmt_reg;
        let clk = self.compile_expr(instantiated_module, Expr::canonical(reg.clock_domain().clk));
        let clk = self
            .compiled_expr_to_value(clk, reg.source_location())
            .map_ty(Clock::from_canonical);
        let clk = self.compile_clock(clk, reg.source_location());
        struct Dispatch;
        impl ResetTypeDispatch for Dispatch {
            type Input<T: ResetType> = ();

            type Output<T: ResetType> = bool;

            fn reset(self, _input: Self::Input<Reset>) -> Self::Output<Reset> {
                unreachable!()
            }

            fn sync_reset(self, _input: Self::Input<SyncReset>) -> Self::Output<SyncReset> {
                false
            }

            fn async_reset(self, _input: Self::Input<AsyncReset>) -> Self::Output<AsyncReset> {
                true
            }
        }
        let reset = if let Some(init) = reg.init() {
            let init = self.compile_expr(instantiated_module, init);
            let init = self.compiled_expr_to_value(init, reg.source_location());
            let rst =
                self.compile_expr(instantiated_module, Expr::canonical(reg.clock_domain().rst));
            let rst = self.compiled_expr_to_value(rst, reg.source_location());
            let rst = self.compiled_value_bool_dest_is_small(rst, reg.source_location());
            let is_async = R::dispatch((), Dispatch);
            if is_async {
                let cond = Expr::canonical(reg.clock_domain().rst.cast_to(Bool));
                let cond = self.compile_expr(instantiated_module, cond);
                let cond = self.compiled_expr_to_value(cond, reg.source_location());
                let cond = cond.map_ty(Bool::from_canonical);
                // write to the register's current value since asynchronous reset is combinational
                let lhs = CompiledValue {
                    layout: value.layout,
                    range: value.range,
                    write: None,
                }
                .into();
                self.compile_simple_connect(
                    [Cond {
                        body: CondBody::IfTrue { cond },
                        source_location: reg.source_location(),
                    }][..]
                        .intern(),
                    lhs,
                    init,
                    reg.source_location(),
                );
            }
            Some(RegisterReset {
                is_async,
                init,
                rst,
            })
        } else {
            None
        };
        self.registers.push(Register {
            value,
            clk_triggered: clk.clk_triggered,
            reset,
            source_location: reg.source_location(),
        });
    }
    fn compile_declaration(
        &mut self,
        declaration: StmtDeclaration,
        parent_module: Interned<InstantiatedModule>,
        conditions: Interned<[Cond]>,
    ) -> TraceDecl {
        let target_base: TargetBase = match &declaration {
            StmtDeclaration::Wire(v) => v.wire.into(),
            StmtDeclaration::Reg(v) => v.reg.into(),
            StmtDeclaration::RegSync(v) => v.reg.into(),
            StmtDeclaration::RegAsync(v) => v.reg.into(),
            StmtDeclaration::Instance(v) => v.instance.into(),
        };
        let target = TargetInInstantiatedModule {
            instantiated_module: *parent_module,
            target: target_base.into(),
        };
        self.decl_conditions.insert(target, conditions);
        let compiled_value = self.compile_value(target);
        match declaration {
            StmtDeclaration::Wire(StmtWire { annotations, wire }) => {}
            StmtDeclaration::Reg(_) => {
                unreachable!("Reset types were already replaced by SyncReset or AsyncReset");
            }
            StmtDeclaration::RegSync(stmt_reg) => {
                self.compile_stmt_reg(stmt_reg, *parent_module, compiled_value)
            }
            StmtDeclaration::RegAsync(stmt_reg) => {
                self.compile_stmt_reg(stmt_reg, *parent_module, compiled_value)
            }
            StmtDeclaration::Instance(StmtInstance {
                annotations,
                instance,
            }) => {
                let inner_instantiated_module = InstantiatedModule::Child {
                    parent: parent_module,
                    instance: instance.intern_sized(),
                }
                .intern_sized();
                let instance_expr = instance.to_expr();
                self.compile_module(inner_instantiated_module);
                for (field_index, module_io) in
                    instance.instantiated().module_io().into_iter().enumerate()
                {
                    let instance_field =
                        ops::FieldAccess::new_by_index(instance_expr, field_index).to_expr();
                    match Expr::flow(instance_field) {
                        Flow::Source => {
                            // we need to supply the value to the instance since the
                            // parent module expects to read from the instance
                            self.compile_connect(
                                *parent_module,
                                conditions,
                                instance_field,
                                *inner_instantiated_module,
                                Interned::default(),
                                module_io.module_io.to_expr(),
                                instance.source_location(),
                            );
                        }
                        Flow::Sink => {
                            // we need to take the value from the instance since the
                            // parent module expects to write to the instance
                            self.compile_connect(
                                *inner_instantiated_module,
                                Interned::default(),
                                module_io.module_io.to_expr(),
                                *parent_module,
                                conditions,
                                instance_field,
                                instance.source_location(),
                            );
                        }
                        Flow::Duplex => unreachable!(),
                    }
                }
            }
        }
        self.make_trace_decl(*parent_module, target_base)
    }
    fn allocate_delay_chain<T, BK: InsnsBuildingKind>(
        &mut self,
        len: usize,
        layout: &TypeLayout<BK>,
        first: Option<T>,
        last: Option<T>,
        mut from_allocation: impl FnMut(TypeIndexRange) -> T,
    ) -> Vec<T> {
        match (len, first, last) {
            (0, _, _) => Vec::new(),
            (1, Some(v), _) | (1, None, Some(v)) => vec![v],
            (2, Some(first), Some(last)) => vec![first, last],
            (len, first, last) => {
                let inner_len = len - first.is_some() as usize - last.is_some() as usize;
                first
                    .into_iter()
                    .chain(
                        (0..inner_len)
                            .map(|_| from_allocation(self.insns.allocate_variable(layout))),
                    )
                    .chain(last)
                    .collect()
            }
        }
    }
    fn allocate_delay_chain_small(
        &mut self,
        len: usize,
        ty: CanonicalType,
        first: Option<StatePartIndex<StatePartKindSmallSlots>>,
        last: Option<StatePartIndex<StatePartKindSmallSlots>>,
    ) -> Vec<StatePartIndex<StatePartKindSmallSlots>> {
        self.allocate_delay_chain(
            len,
            &TypeLayout {
                small_slots: StatePartLayout::scalar(
                    SlotDebugData {
                        name: Interned::default(),
                        ty,
                    },
                    (),
                ),
                big_slots: StatePartLayout::empty(),
            },
            first,
            last,
            |range| range.small_slots.start,
        )
    }
    fn compile_memory_port_rw_helper(
        &mut self,
        memory: StatePartIndex<StatePartKindMemories>,
        stride: usize,
        mut start: usize,
        data_layout: CompiledTypeLayout<CanonicalType>,
        mask_layout: CompiledTypeLayout<CanonicalType>,
        mut read: Option<MemoryPortReadInsns<'_>>,
        mut write: Option<MemoryPortWriteInsns<'_>>,
    ) {
        match data_layout.body {
            CompiledTypeLayoutBody::Scalar => {
                let CompiledTypeLayoutBody::Scalar = mask_layout.body else {
                    unreachable!();
                };
                let signed = match data_layout.ty {
                    CanonicalType::UInt(_) => false,
                    CanonicalType::SInt(_) => true,
                    CanonicalType::Bool(_) => false,
                    CanonicalType::Array(_) => unreachable!(),
                    CanonicalType::Enum(_) => false,
                    CanonicalType::Bundle(_) => unreachable!(),
                    CanonicalType::AsyncReset(_) => false,
                    CanonicalType::SyncReset(_) => false,
                    CanonicalType::Reset(_) => false,
                    CanonicalType::Clock(_) => false,
                    CanonicalType::PhantomConst(_) => unreachable!(),
                };
                let width = data_layout.ty.bit_width();
                if let Some(MemoryPortReadInsns {
                    addr,
                    en: _,
                    write_mode: _,
                    data,
                    insns,
                }) = read
                {
                    insns.push(
                        match data.len() {
                            TypeLen::A_BIG_SLOT => {
                                let dest = data.big_slots.start;
                                if signed {
                                    Insn::MemoryReadSInt {
                                        dest,
                                        memory,
                                        addr,
                                        stride,
                                        start,
                                        width,
                                    }
                                } else {
                                    Insn::MemoryReadUInt {
                                        dest,
                                        memory,
                                        addr,
                                        stride,
                                        start,
                                        width,
                                    }
                                }
                            }
                            TypeLen::A_SMALL_SLOT => {
                                let _dest = data.small_slots.start;
                                todo!("memory ports' data are always big for now");
                            }
                            _ => unreachable!(),
                        }
                        .into(),
                    );
                }
                if let Some(MemoryPortWriteInsns {
                    addr,
                    en: _,
                    write_mode: _,
                    data,
                    mask,
                    insns,
                }) = write
                {
                    let end_label = self.insns.new_label();
                    insns.push(
                        match mask.len() {
                            TypeLen::A_BIG_SLOT => Insn::BranchIfZero {
                                target: end_label.0,
                                value: mask.big_slots.start,
                            },
                            TypeLen::A_SMALL_SLOT => Insn::BranchIfSmallZero {
                                target: end_label.0,
                                value: mask.small_slots.start,
                            },
                            _ => unreachable!(),
                        }
                        .into(),
                    );
                    insns.push(
                        match data.len() {
                            TypeLen::A_BIG_SLOT => {
                                let value = data.big_slots.start;
                                if signed {
                                    Insn::MemoryWriteSInt {
                                        value,
                                        memory,
                                        addr,
                                        stride,
                                        start,
                                        width,
                                    }
                                } else {
                                    Insn::MemoryWriteUInt {
                                        value,
                                        memory,
                                        addr,
                                        stride,
                                        start,
                                        width,
                                    }
                                }
                            }
                            TypeLen::A_SMALL_SLOT => {
                                let _value = data.small_slots.start;
                                todo!("memory ports' data are always big for now");
                            }
                            _ => unreachable!(),
                        }
                        .into(),
                    );
                    insns.push(end_label.into());
                }
            }
            CompiledTypeLayoutBody::Array { element } => {
                let CompiledTypeLayoutBody::Array {
                    element: mask_element,
                } = mask_layout.body
                else {
                    unreachable!();
                };
                let ty = <Array>::from_canonical(data_layout.ty);
                let element_bit_width = ty.element().bit_width();
                let element_size = element.layout.len();
                let mask_element_size = mask_element.layout.len();
                for element_index in 0..ty.len() {
                    self.compile_memory_port_rw_helper(
                        memory,
                        stride,
                        start,
                        *element,
                        *mask_element,
                        read.as_mut().map(
                            |MemoryPortReadInsns {
                                 addr,
                                 en,
                                 write_mode,
                                 data,
                                 insns,
                             }| MemoryPortReadInsns {
                                addr: *addr,
                                en: *en,
                                write_mode: *write_mode,
                                data: data.index_array(element_size, element_index),
                                insns,
                            },
                        ),
                        write.as_mut().map(
                            |MemoryPortWriteInsns {
                                 addr,
                                 en,
                                 write_mode,
                                 data,
                                 mask,
                                 insns,
                             }| {
                                MemoryPortWriteInsns {
                                    addr: *addr,
                                    en: *en,
                                    write_mode: *write_mode,
                                    data: data.index_array(element_size, element_index),
                                    mask: mask.index_array(mask_element_size, element_index),
                                    insns,
                                }
                            },
                        ),
                    );
                    start += element_bit_width;
                }
            }
            CompiledTypeLayoutBody::Bundle { fields } => {
                let CompiledTypeLayoutBody::Bundle {
                    fields: mask_fields,
                } = mask_layout.body
                else {
                    unreachable!();
                };
                assert_eq!(fields.len(), mask_fields.len());
                for (field, mask_field) in fields.into_iter().zip(mask_fields) {
                    let field_index_range =
                        TypeIndexRange::new(field.offset, field.ty.layout.len());
                    let mask_field_index_range =
                        TypeIndexRange::new(mask_field.offset, mask_field.ty.layout.len());
                    self.compile_memory_port_rw_helper(
                        memory,
                        stride,
                        start,
                        field.ty,
                        mask_field.ty,
                        read.as_mut().map(
                            |MemoryPortReadInsns {
                                 addr,
                                 en,
                                 write_mode,
                                 data,
                                 insns,
                             }| MemoryPortReadInsns {
                                addr: *addr,
                                en: *en,
                                write_mode: *write_mode,
                                data: data.slice(field_index_range),
                                insns,
                            },
                        ),
                        write.as_mut().map(
                            |MemoryPortWriteInsns {
                                 addr,
                                 en,
                                 write_mode,
                                 data,
                                 mask,
                                 insns,
                             }| {
                                MemoryPortWriteInsns {
                                    addr: *addr,
                                    en: *en,
                                    write_mode: *write_mode,
                                    data: data.slice(field_index_range),
                                    mask: mask.slice(mask_field_index_range),
                                    insns,
                                }
                            },
                        ),
                    );
                    start = start + field.ty.ty.bit_width();
                }
            }
        }
    }
    fn compile_memory_port_rw(
        &mut self,
        memory: StatePartIndex<StatePartKindMemories>,
        data_layout: CompiledTypeLayout<CanonicalType>,
        mask_layout: CompiledTypeLayout<CanonicalType>,
        mut read: Option<MemoryPortReadInsns<'_>>,
        mut write: Option<MemoryPortWriteInsns<'_>>,
    ) {
        let read_else_label = read.as_mut().map(
            |MemoryPortReadInsns {
                 addr: _,
                 en,
                 write_mode,
                 data: _,
                 insns,
             }| {
                let else_label = self.insns.new_label();
                insns.push(
                    Insn::BranchIfSmallZero {
                        target: else_label.0,
                        value: *en,
                    }
                    .into(),
                );
                if let Some(write_mode) = *write_mode {
                    insns.push(
                        Insn::BranchIfSmallNonZero {
                            target: else_label.0,
                            value: write_mode,
                        }
                        .into(),
                    );
                }
                else_label
            },
        );
        let write_end_label = write.as_mut().map(
            |MemoryPortWriteInsns {
                 addr: _,
                 en,
                 write_mode,
                 data: _,
                 mask: _,
                 insns,
             }| {
                let end_label = self.insns.new_label();
                insns.push(
                    Insn::BranchIfSmallZero {
                        target: end_label.0,
                        value: *en,
                    }
                    .into(),
                );
                if let Some(write_mode) = *write_mode {
                    insns.push(
                        Insn::BranchIfSmallZero {
                            target: end_label.0,
                            value: write_mode,
                        }
                        .into(),
                    );
                }
                end_label
            },
        );
        self.compile_memory_port_rw_helper(
            memory,
            data_layout.ty.bit_width(),
            0,
            data_layout,
            mask_layout,
            read.as_mut().map(
                |MemoryPortReadInsns {
                     addr,
                     en,
                     write_mode,
                     data,
                     insns,
                 }| MemoryPortReadInsns {
                    addr: *addr,
                    en: *en,
                    write_mode: *write_mode,
                    data: *data,
                    insns: *insns,
                },
            ),
            write.as_mut().map(
                |MemoryPortWriteInsns {
                     addr,
                     en,
                     write_mode,
                     data,
                     mask,
                     insns,
                 }| MemoryPortWriteInsns {
                    addr: *addr,
                    en: *en,
                    write_mode: *write_mode,
                    data: *data,
                    mask: *mask,
                    insns: *insns,
                },
            ),
        );
        if let (
            Some(else_label),
            Some(MemoryPortReadInsns {
                addr: _,
                en: _,
                write_mode: _,
                data,
                insns,
            }),
        ) = (read_else_label, read)
        {
            let end_label = self.insns.new_label();
            insns.push(
                Insn::Branch {
                    target: end_label.0,
                }
                .into(),
            );
            insns.push(else_label.into());
            let TypeIndexRange {
                small_slots,
                big_slots,
            } = data;
            for dest in small_slots.iter() {
                insns.push(Insn::ConstSmall { dest, value: 0 }.into());
            }
            for dest in big_slots.iter() {
                insns.push(
                    Insn::Const {
                        dest,
                        value: BigInt::ZERO.intern_sized(),
                    }
                    .into(),
                );
            }
            insns.push(end_label.into());
        }
        if let (Some(end_label), Some(write)) = (write_end_label, write) {
            write.insns.push(end_label.into());
        }
    }
    fn compile_memory(
        &mut self,
        mem: Mem,
        instantiated_module: InstantiatedModule,
        conditions: Interned<[Cond]>,
        trace_decls: &mut Vec<TraceDecl>,
    ) {
        let data_layout = CompiledTypeLayout::get(mem.array_type().element());
        let mask_layout = CompiledTypeLayout::get(mem.array_type().element().mask_type());
        let read_latency_plus_1 = mem
            .read_latency()
            .checked_add(1)
            .expect("read latency too big");
        let write_latency_plus_1 = mem
            .write_latency()
            .get()
            .checked_add(1)
            .expect("write latency too big");
        let read_cycle = match mem.read_under_write() {
            ReadUnderWrite::Old => 0,
            ReadUnderWrite::New => mem.read_latency(),
            ReadUnderWrite::Undefined => mem.read_latency() / 2, // something other than Old or New
        };
        let memory = self
            .insns
            .state_layout
            .memories
            .allocate(&StatePartLayout::scalar(
                (),
                MemoryData {
                    array_type: mem.array_type(),
                    data: mem.initial_value().unwrap_or_else(|| {
                        Intern::intern_owned(BitVec::repeat(
                            false,
                            mem.array_type().type_properties().bit_width,
                        ))
                    }),
                },
            ))
            .start;
        let (ports, trace_ports) = mem
            .ports()
            .iter()
            .map(|&port| {
                let target_base = TargetBase::MemPort(port);
                let target = TargetInInstantiatedModule {
                    instantiated_module,
                    target: target_base.into(),
                };
                self.decl_conditions.insert(target, conditions);
                let TraceDecl::Scope(TraceScope::MemPort(trace_port)) =
                    self.make_trace_decl(instantiated_module, target_base)
                else {
                    unreachable!();
                };
                let clk = Expr::field(port.to_expr(), "clk");
                let clk = self.compile_expr(instantiated_module, clk);
                let clk = self.compiled_expr_to_value(clk, mem.source_location());
                let clk_triggered = self
                    .compile_clock(clk.map_ty(Clock::from_canonical), mem.source_location())
                    .clk_triggered;
                let en = Expr::field(port.to_expr(), "en");
                let en = self.compile_expr(instantiated_module, en);
                let en = self.compiled_expr_to_value(en, mem.source_location());
                let en = self.compiled_value_bool_dest_is_small(en, mem.source_location());
                let addr = Expr::field(port.to_expr(), "addr");
                let addr = self.compile_expr(instantiated_module, addr);
                let addr = self.compiled_expr_to_value(addr, mem.source_location());
                let addr_ty = addr.layout.ty;
                let addr = self.compiled_value_to_dyn_array_index(
                    addr.map_ty(UInt::from_canonical),
                    mem.source_location(),
                );
                let read_data = port.port_kind().rdata_name().map(|name| {
                    let read_data =
                        self.compile_expr(instantiated_module, Expr::field(port.to_expr(), name));
                    let read_data = self.compiled_expr_to_value(read_data, mem.source_location());
                    read_data.range
                });
                let write_data = port.port_kind().wdata_name().map(|name| {
                    let write_data =
                        self.compile_expr(instantiated_module, Expr::field(port.to_expr(), name));
                    let write_data = self.compiled_expr_to_value(write_data, mem.source_location());
                    write_data.range
                });
                let write_mask = port.port_kind().wmask_name().map(|name| {
                    let write_mask =
                        self.compile_expr(instantiated_module, Expr::field(port.to_expr(), name));
                    let write_mask = self.compiled_expr_to_value(write_mask, mem.source_location());
                    write_mask.range
                });
                let write_mode = port.port_kind().wmode_name().map(|name| {
                    let write_mode =
                        self.compile_expr(instantiated_module, Expr::field(port.to_expr(), name));
                    let write_mode = self.compiled_expr_to_value(write_mode, mem.source_location());
                    self.compiled_value_bool_dest_is_small(write_mode, mem.source_location())
                });
                struct PortParts {
                    en_delayed_len: usize,
                    addr_delayed_len: usize,
                    read_data_delayed_len: usize,
                    write_data_delayed_len: usize,
                    write_mask_delayed_len: usize,
                    write_mode_delayed_len: usize,
                    read_cycle: Option<usize>,
                    write_cycle: Option<usize>,
                }
                let PortParts {
                    en_delayed_len,
                    addr_delayed_len,
                    read_data_delayed_len,
                    write_data_delayed_len,
                    write_mask_delayed_len,
                    write_mode_delayed_len,
                    read_cycle,
                    write_cycle,
                } = match port.port_kind() {
                    PortKind::ReadOnly => PortParts {
                        en_delayed_len: read_cycle + 1,
                        addr_delayed_len: read_cycle + 1,
                        read_data_delayed_len: read_latency_plus_1 - read_cycle,
                        write_data_delayed_len: 0,
                        write_mask_delayed_len: 0,
                        write_mode_delayed_len: 0,
                        read_cycle: Some(read_cycle),
                        write_cycle: None,
                    },
                    PortKind::WriteOnly => PortParts {
                        en_delayed_len: write_latency_plus_1,
                        addr_delayed_len: write_latency_plus_1,
                        read_data_delayed_len: 0,
                        write_data_delayed_len: write_latency_plus_1,
                        write_mask_delayed_len: write_latency_plus_1,
                        write_mode_delayed_len: 0,
                        read_cycle: None,
                        write_cycle: Some(mem.write_latency().get()),
                    },
                    PortKind::ReadWrite => {
                        let can_rw_at_end = match mem.read_under_write() {
                            ReadUnderWrite::Old => false,
                            ReadUnderWrite::New | ReadUnderWrite::Undefined => true,
                        };
                        let latency_plus_1 = read_latency_plus_1;
                        if latency_plus_1 != write_latency_plus_1 || !can_rw_at_end {
                            todo!(
                                "not sure what to do, issue: \
                                https://github.com/chipsalliance/firrtl-spec/issues/263"
                            );
                        }
                        PortParts {
                            en_delayed_len: latency_plus_1,
                            addr_delayed_len: latency_plus_1,
                            read_data_delayed_len: 1,
                            write_data_delayed_len: latency_plus_1,
                            write_mask_delayed_len: latency_plus_1,
                            write_mode_delayed_len: latency_plus_1,
                            read_cycle: Some(latency_plus_1 - 1),
                            write_cycle: Some(latency_plus_1 - 1),
                        }
                    }
                };
                let addr_delayed = self.allocate_delay_chain_small(
                    addr_delayed_len,
                    addr_ty.canonical(),
                    Some(addr),
                    None,
                );
                let en_delayed = self.allocate_delay_chain_small(
                    en_delayed_len,
                    Bool.canonical(),
                    Some(en),
                    None,
                );
                let read_data_delayed = self.allocate_delay_chain(
                    read_data_delayed_len,
                    &data_layout.layout,
                    None,
                    read_data,
                    |v| v,
                );
                let write_data_delayed = self.allocate_delay_chain(
                    write_data_delayed_len,
                    &data_layout.layout,
                    write_data,
                    None,
                    |v| v,
                );
                let write_mask_delayed = self.allocate_delay_chain(
                    write_mask_delayed_len,
                    &mask_layout.layout,
                    write_mask,
                    None,
                    |v| v,
                );
                let write_mode_delayed = self.allocate_delay_chain_small(
                    write_mode_delayed_len,
                    Bool.canonical(),
                    write_mode,
                    None,
                );
                let mut read_insns = Vec::new();
                let mut write_insns = Vec::new();
                self.compile_memory_port_rw(
                    memory,
                    data_layout,
                    mask_layout,
                    read_cycle.map(|read_cycle| MemoryPortReadInsns {
                        addr: addr_delayed[read_cycle],
                        en: en_delayed[read_cycle],
                        write_mode: write_mode_delayed.get(read_cycle).copied(),
                        data: read_data_delayed[0],
                        insns: &mut read_insns,
                    }),
                    write_cycle.map(|write_cycle| MemoryPortWriteInsns {
                        addr: addr_delayed[write_cycle],
                        en: en_delayed[write_cycle],
                        write_mode: write_mode_delayed.get(write_cycle).copied(),
                        data: write_data_delayed[write_cycle],
                        mask: write_mask_delayed[write_cycle],
                        insns: &mut write_insns,
                    }),
                );
                self.add_assignment(Interned::default(), read_insns, mem.source_location());
                (
                    MemoryPort {
                        clk_triggered,
                        addr_delayed,
                        en_delayed,
                        data_layout,
                        read_data_delayed,
                        write_data_delayed,
                        write_mask_delayed,
                        write_mode_delayed,
                        write_insns,
                    },
                    trace_port,
                )
            })
            .unzip();
        let name = mem.scoped_name().1 .0;
        let id = TraceMemoryId(self.memories.len());
        let stride = mem.array_type().element().bit_width();
        let trace = TraceMem {
            id,
            name,
            stride,
            element_type: self
                .make_trace_decl_child(
                    instantiated_module,
                    MakeTraceDeclTarget::Memory {
                        id,
                        depth: mem.array_type().len(),
                        stride,
                        start: 0,
                        ty: mem.array_type().element(),
                    },
                    name,
                    mem.source_location(),
                )
                .intern_sized(),
            ports: Intern::intern_owned(trace_ports),
            array_type: mem.array_type(),
        };
        trace_decls.push(trace.into());
        self.memories.push(Memory {
            mem,
            memory,
            trace,
            ports,
        });
    }
    fn compile_block(
        &mut self,
        parent_module: Interned<InstantiatedModule>,
        block: Block,
        conditions: Interned<[Cond]>,
        trace_decls: &mut Vec<TraceDecl>,
    ) {
        let Block { memories, stmts } = block;
        for memory in memories {
            self.compile_memory(memory, *parent_module, conditions, trace_decls);
        }
        for stmt in stmts {
            match stmt {
                Stmt::Connect(StmtConnect {
                    lhs,
                    rhs,
                    source_location,
                }) => self.compile_connect(
                    *parent_module,
                    conditions,
                    lhs,
                    *parent_module,
                    conditions,
                    rhs,
                    source_location,
                ),
                Stmt::Formal(StmtFormal { .. }) => todo!("implement simulating formal statements"),
                Stmt::If(StmtIf {
                    cond,
                    source_location,
                    blocks: [then_block, else_block],
                }) => {
                    let cond = self.compile_expr(*parent_module, Expr::canonical(cond));
                    let cond = self.compiled_expr_to_value(cond, source_location);
                    let cond = cond.map_ty(Bool::from_canonical);
                    self.compile_block(
                        parent_module,
                        then_block,
                        Interned::from_iter(conditions.iter().copied().chain([Cond {
                            body: CondBody::IfTrue { cond },
                            source_location,
                        }])),
                        trace_decls,
                    );
                    self.compile_block(
                        parent_module,
                        else_block,
                        Interned::from_iter(conditions.iter().copied().chain([Cond {
                            body: CondBody::IfFalse { cond },
                            source_location,
                        }])),
                        trace_decls,
                    );
                }
                Stmt::Match(StmtMatch {
                    expr,
                    source_location,
                    blocks,
                }) => {
                    let enum_expr = self.compile_expr(*parent_module, Expr::canonical(expr));
                    let enum_expr = self.compiled_expr_to_value(enum_expr, source_location);
                    let enum_expr = enum_expr.map_ty(Enum::from_canonical);
                    let discriminant = self.compile_enum_discriminant(enum_expr, source_location);
                    for (variant_index, block) in blocks.into_iter().enumerate() {
                        self.compile_block(
                            parent_module,
                            block,
                            Interned::from_iter(conditions.iter().copied().chain([Cond {
                                body: CondBody::MatchArm {
                                    discriminant,
                                    variant_index,
                                },
                                source_location,
                            }])),
                            trace_decls,
                        );
                    }
                }
                Stmt::Declaration(declaration) => {
                    trace_decls.push(self.compile_declaration(
                        declaration,
                        parent_module,
                        conditions,
                    ));
                }
            }
        }
    }
    fn compile_module(&mut self, module: Interned<InstantiatedModule>) -> &CompiledModule {
        let mut trace_decls = Vec::new();
        let module_io = module
            .leaf_module()
            .module_io()
            .iter()
            .map(
                |&AnnotatedModuleIO {
                     annotations: _,
                     module_io,
                 }| {
                    let target = TargetInInstantiatedModule {
                        instantiated_module: *module,
                        target: Target::from(module_io),
                    };
                    self.decl_conditions.insert(target, Interned::default());
                    trace_decls.push(self.make_trace_decl(*module, module_io.into()));
                    self.compile_value(target)
                },
            )
            .collect();
        match module.leaf_module().body() {
            ModuleBody::Normal(NormalModuleBody { body }) => {
                self.compile_block(module, body, Interned::default(), &mut trace_decls);
            }
            ModuleBody::Extern(ExternModuleBody {
                verilog_name: _,
                parameters: _,
                simulation,
            }) => {
                let Some(simulation) = simulation else {
                    panic!(
                        "can't simulate extern module without extern_module_simulation: {}",
                        module.leaf_module().source_location()
                    );
                };
                self.extern_modules.push(CompiledExternModule {
                    module_io_targets: module
                        .leaf_module()
                        .module_io()
                        .iter()
                        .map(|v| Target::from(v.module_io))
                        .collect(),
                    module_io,
                    simulation,
                });
            }
        }
        let hashbrown::hash_map::Entry::Vacant(entry) = self.modules.entry(*module) else {
            unreachable!("compiled same instantiated module twice");
        };
        entry.insert(CompiledModule {
            module_io,
            trace_decls: TraceModule {
                name: module.leaf_module().name(),
                children: Intern::intern_owned(trace_decls),
            },
        })
    }
    fn process_assignments(&mut self) {
        self.assignments
            .finalize(self.insns.state_layout().ty.clone().into());
        if let Some(DebugOpaque(dump_assignments_dot)) = &self.dump_assignments_dot {
            let graph =
                petgraph::graph::DiGraph::<_, _, usize>::from_elements(self.assignments.elements());
            dump_assignments_dot(&petgraph::dot::Dot::new(&graph));
        }
        let assignments_order: Vec<_> = match petgraph::algo::toposort(&self.assignments, None) {
            Ok(nodes) => nodes
                .into_iter()
                .filter_map(|n| match n {
                    AssignmentOrSlotIndex::AssignmentIndex(v) => Some(v),
                    _ => None,
                })
                .collect(),
            Err(e) => match e.node_id() {
                AssignmentOrSlotIndex::AssignmentIndex(assignment_index) => panic!(
                    "combinatorial logic cycle detected at: {}",
                    self.assignments.assignments()[assignment_index].source_location,
                ),
                AssignmentOrSlotIndex::SmallSlot(slot) => panic!(
                    "combinatorial logic cycle detected through: {}",
                    self.insns.state_layout().ty.small_slots.debug_data[slot.as_usize()].name,
                ),
                AssignmentOrSlotIndex::BigSlot(slot) => panic!(
                    "combinatorial logic cycle detected through: {}",
                    self.insns.state_layout().ty.big_slots.debug_data[slot.as_usize()].name,
                ),
            },
        };
        struct CondStackEntry<'a> {
            cond: &'a Cond,
            end_label: Label,
        }
        let mut cond_stack = Vec::<CondStackEntry<'_>>::new();
        for assignment_index in assignments_order {
            let Assignment {
                inputs: _,
                outputs: _,
                conditions,
                insns,
                source_location,
            } = &self.assignments.assignments()[assignment_index];
            let mut same_len = 0;
            for (index, (entry, cond)) in cond_stack.iter().zip(conditions).enumerate() {
                if entry.cond != cond {
                    break;
                }
                same_len = index + 1;
            }
            while cond_stack.len() > same_len {
                let CondStackEntry { cond: _, end_label } =
                    cond_stack.pop().expect("just checked len");
                self.insns.define_label_at_next_insn(end_label);
            }
            for cond in &conditions[cond_stack.len()..] {
                let end_label = self.insns.new_label();
                match cond.body {
                    CondBody::IfTrue { cond: cond_value }
                    | CondBody::IfFalse { cond: cond_value } => {
                        let (branch_if_zero, branch_if_non_zero) = match cond_value.range.len() {
                            TypeLen::A_SMALL_SLOT => (
                                Insn::BranchIfSmallZero {
                                    target: end_label.0,
                                    value: cond_value.range.small_slots.start,
                                },
                                Insn::BranchIfSmallNonZero {
                                    target: end_label.0,
                                    value: cond_value.range.small_slots.start,
                                },
                            ),
                            TypeLen::A_BIG_SLOT => (
                                Insn::BranchIfZero {
                                    target: end_label.0,
                                    value: cond_value.range.big_slots.start,
                                },
                                Insn::BranchIfNonZero {
                                    target: end_label.0,
                                    value: cond_value.range.big_slots.start,
                                },
                            ),
                            _ => unreachable!(),
                        };
                        self.insns.push(
                            if let CondBody::IfTrue { .. } = cond.body {
                                branch_if_zero
                            } else {
                                branch_if_non_zero
                            },
                            cond.source_location,
                        );
                    }
                    CondBody::MatchArm {
                        discriminant,
                        variant_index,
                    } => {
                        self.insns.push(
                            Insn::BranchIfSmallNeImmediate {
                                target: end_label.0,
                                lhs: discriminant,
                                rhs: variant_index as _,
                            },
                            cond.source_location,
                        );
                    }
                }
                cond_stack.push(CondStackEntry { cond, end_label });
            }
            self.insns.extend(insns.iter().copied(), *source_location);
        }
        for CondStackEntry { cond: _, end_label } in cond_stack {
            self.insns.define_label_at_next_insn(end_label);
        }
    }
    fn process_clocks(&mut self) -> Interned<[StatePartIndex<StatePartKindSmallSlots>]> {
        mem::take(&mut self.clock_triggers)
            .into_iter()
            .map(
                |ClockTrigger {
                     last_clk_was_low,
                     clk,
                     clk_triggered,
                     source_location,
                 }| {
                    self.insns.push(
                        Insn::XorSmallImmediate {
                            dest: last_clk_was_low,
                            lhs: clk,
                            rhs: 1,
                        },
                        source_location,
                    );
                    clk_triggered
                },
            )
            .collect()
    }
    fn process_registers(&mut self) {
        for Register {
            value,
            clk_triggered,
            reset,
            source_location,
        } in mem::take(&mut self.registers)
        {
            match reset {
                Some(RegisterReset {
                    is_async,
                    init,
                    rst,
                }) => {
                    let reg_end = self.insns.new_label();
                    let reg_reset = self.insns.new_label();
                    let branch_if_reset = Insn::BranchIfSmallNonZero {
                        target: reg_reset.0,
                        value: rst,
                    };
                    let branch_if_not_triggered = Insn::BranchIfSmallZero {
                        target: reg_end.0,
                        value: clk_triggered,
                    };
                    if is_async {
                        self.insns.push(branch_if_reset, source_location);
                        self.insns.push(branch_if_not_triggered, source_location);
                    } else {
                        self.insns.push(branch_if_not_triggered, source_location);
                        self.insns.push(branch_if_reset, source_location);
                    }
                    self.insns.extend(
                        value.range.insns_for_copy_from(value.write_value().range),
                        source_location,
                    );
                    self.insns
                        .push(Insn::Branch { target: reg_end.0 }, source_location);
                    self.insns.define_label_at_next_insn(reg_reset);
                    self.insns
                        .extend(value.range.insns_for_copy_from(init.range), source_location);
                    self.insns.define_label_at_next_insn(reg_end);
                }
                None => {
                    let reg_end = self.insns.new_label();
                    self.insns.push(
                        Insn::BranchIfSmallZero {
                            target: reg_end.0,
                            value: clk_triggered,
                        },
                        source_location,
                    );
                    self.insns.extend(
                        value.range.insns_for_copy_from(value.write_value().range),
                        source_location,
                    );
                    self.insns.define_label_at_next_insn(reg_end);
                }
            }
        }
    }
    fn process_memories(&mut self) {
        for memory_index in 0..self.memories.len() {
            let Memory {
                mem,
                memory: _,
                trace: _,
                ref mut ports,
            } = self.memories[memory_index];
            for MemoryPort {
                clk_triggered,
                addr_delayed,
                en_delayed,
                data_layout: _,
                read_data_delayed,
                write_data_delayed,
                write_mask_delayed,
                write_mode_delayed,
                write_insns,
            } in mem::take(ports)
            {
                let port_end = self.insns.new_label();
                let small_shift_reg =
                    |this: &mut Self, values: &[StatePartIndex<StatePartKindSmallSlots>]| {
                        for pair in values.windows(2).rev() {
                            this.insns.push(
                                Insn::CopySmall {
                                    dest: pair[1],
                                    src: pair[0],
                                },
                                mem.source_location(),
                            );
                        }
                    };
                let shift_reg = |this: &mut Self, values: &[TypeIndexRange]| {
                    for pair in values.windows(2).rev() {
                        this.insns
                            .extend(pair[0].insns_for_copy_to(pair[1]), mem.source_location());
                    }
                };
                self.insns.push(
                    Insn::BranchIfSmallZero {
                        target: port_end.0,
                        value: clk_triggered,
                    },
                    mem.source_location(),
                );
                small_shift_reg(self, &addr_delayed);
                small_shift_reg(self, &en_delayed);
                shift_reg(self, &write_data_delayed);
                shift_reg(self, &write_mask_delayed);
                small_shift_reg(self, &write_mode_delayed);
                shift_reg(self, &read_data_delayed);
                self.insns.extend(write_insns, mem.source_location());
                self.insns.define_label_at_next_insn(port_end);
            }
        }
    }
    pub fn compile(mut self) -> Compiled<Bundle> {
        let base_module =
            *self.compile_module(InstantiatedModule::Base(self.base_module).intern_sized());
        self.process_assignments();
        self.process_registers();
        self.process_memories();
        let clocks_triggered = self.process_clocks();
        self.insns
            .push(Insn::Return, self.base_module.source_location());
        Compiled {
            insns: Insns::from(self.insns).intern_sized(),
            base_module,
            extern_modules: Intern::intern_owned(self.extern_modules),
            io: Instance::new_unchecked(
                ScopedNameId(
                    NameId("<simulator>".intern(), Id::new()),
                    self.original_base_module.name_id(),
                ),
                self.original_base_module,
                self.original_base_module.source_location(),
            ),
            traces: SimTraces(Intern::intern_owned(self.traces.0)),
            trace_memories: Interned::from_iter(self.memories.iter().map(
                |&Memory {
                     mem: _,
                     memory,
                     trace,
                     ports: _,
                 }| (memory, trace),
            )),
            clocks_triggered,
        }
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
struct CompiledModule {
    module_io: Interned<[CompiledValue<CanonicalType>]>,
    trace_decls: TraceModule,
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct Compiled<T: BundleType> {
    insns: Interned<Insns<InsnsBuildingDone>>,
    base_module: CompiledModule,
    extern_modules: Interned<[CompiledExternModule]>,
    io: Instance<T>,
    traces: SimTraces<Interned<[SimTrace<SimTraceKind, ()>]>>,
    trace_memories: Interned<[(StatePartIndex<StatePartKindMemories>, TraceMem)]>,
    clocks_triggered: Interned<[StatePartIndex<StatePartKindSmallSlots>]>,
}

impl<T: BundleType> Compiled<T> {
    pub fn new(module: Interned<Module<T>>) -> Self {
        Self::from_canonical(Compiler::new(module.canonical().intern()).compile())
    }
    pub fn canonical(self) -> Compiled<Bundle> {
        let Self {
            insns,
            base_module,
            extern_modules,
            io,
            traces,
            trace_memories,
            clocks_triggered,
        } = self;
        Compiled {
            insns,
            base_module,
            extern_modules,
            io: Instance::from_canonical(io.canonical()),
            traces,
            trace_memories,
            clocks_triggered,
        }
    }
    pub fn from_canonical(canonical: Compiled<Bundle>) -> Self {
        let Compiled {
            insns,
            base_module,
            extern_modules,
            io,
            traces,
            trace_memories,
            clocks_triggered,
        } = canonical;
        Self {
            insns,
            base_module,
            extern_modules,
            io: Instance::from_canonical(io.canonical()),
            traces,
            trace_memories,
            clocks_triggered,
        }
    }
}

#[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub struct TraceScalarId(usize);

impl fmt::Debug for TraceScalarId {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "TraceScalarId({})", self.0)
    }
}

impl TraceScalarId {
    pub fn as_usize(self) -> usize {
        self.0
    }
}

#[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub struct TraceMemoryId(usize);

impl fmt::Debug for TraceMemoryId {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "TraceMemoryId({})", self.0)
    }
}

impl TraceMemoryId {
    pub fn as_usize(self) -> usize {
        self.0
    }
}

#[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord, Hash, Debug)]
pub struct TraceMemoryLocation {
    pub id: TraceMemoryId,
    pub depth: usize,
    pub stride: usize,
    pub start: usize,
    pub len: usize,
}

#[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum TraceLocation {
    Scalar(TraceScalarId),
    Memory(TraceMemoryLocation),
}

impl fmt::Debug for TraceLocation {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::Scalar(v) => v.fmt(f),
            Self::Memory(v) => v.fmt(f),
        }
    }
}

macro_rules! impl_trace_decl {
    (
        $(
            #[kind = $category_kind:ident]
            $(#[$category_meta:meta])*
            $category_variant:ident($category_enum:ident {
                fn $category_property_fn:ident(self) -> $category_property_fn_ret_ty:ty;
                $(
                    $(#[$meta:meta])*
                    $variant:ident($struct:ident {
                        fn $property_fn:ident($property_fn_self:ident) -> _ $property_fn_block:block
                        $($(#[$field_meta:meta])*
                        $field_name:ident: $field_ty:ty,)*
                    }),
                )*
            }),
        )*
    ) => {
        $(
            #[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord, Hash, Debug)]
            #[non_exhaustive]
            $(#[$category_meta])*
            pub enum $category_kind {
                $($(#[$meta])*
                $variant,)*
            }

            impl From<$category_kind> for TraceKind {
                fn from(v: $category_kind) -> Self {
                    TraceKind::$category_variant(v)
                }
            }

            #[derive(Copy, Clone, PartialEq, Eq, Hash)]
            #[non_exhaustive]
            $(#[$category_meta])*
            pub enum $category_enum {
                $($(#[$meta])*
                $variant($struct),)*
            }

            impl fmt::Debug for $category_enum {
                fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
                    match *self {
                        $(Self::$variant(v) => v.fmt(f),)*
                    }
                }
            }

            impl $category_enum {
                pub fn kind(self) -> $category_kind {
                    match self {
                        $(Self::$variant(_) => $category_kind::$variant,)*
                    }
                }
                pub fn name(self) -> Interned<str> {
                    match self {
                        $(Self::$variant(v) => v.name,)*
                    }
                }
                pub fn $category_property_fn(self) -> $category_property_fn_ret_ty {
                    match self {
                        $(Self::$variant(v) => v.$property_fn(),)*
                    }
                }
            }

            impl From<$category_enum> for TraceDecl {
                fn from(v: $category_enum) -> Self {
                    TraceDecl::$category_variant(v)
                }
            }

            $(
                #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
                #[non_exhaustive]
                $(#[$meta])*
                pub struct $struct {
                    $($(#[$field_meta])*
                    pub $field_name: $field_ty,)*
                }

                impl $struct {
                    pub fn $property_fn($property_fn_self) -> $category_property_fn_ret_ty $property_fn_block
                }

                impl From<$struct> for $category_enum {
                    fn from(v: $struct) -> Self {
                        $category_enum::$variant(v)
                    }
                }

                impl From<$struct> for TraceDecl {
                    fn from(v: $struct) -> Self {
                        TraceDecl::$category_variant($category_enum::$variant(v))
                    }
                }
            )*
        )*

        #[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
        pub enum TraceKind {
            $($(#[$category_meta])*
            $category_variant($category_kind),)*
        }

        impl fmt::Debug for TraceKind {
            fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
                match *self {
                    $(Self::$category_variant(v) => v.fmt(f),)*
                }
            }
        }

        #[derive(Copy, Clone, PartialEq, Eq, Hash)]
        pub enum TraceDecl {
            $($(#[$category_meta])*
            $category_variant($category_enum),)*
        }

        impl fmt::Debug for TraceDecl {
            fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
                match *self {
                    $(Self::$category_variant(v) => v.fmt(f),)*
                }
            }
        }
    };
}

impl_trace_decl! {
    #[kind = TraceScopeKind]
    Scope(TraceScope {
        fn children(self) -> Interned<[TraceDecl]>;
        Module(TraceModule {
            fn children(self) -> _ {
                self.children
            }
            name: Interned<str>,
            children: Interned<[TraceDecl]>,
        }),
        Instance(TraceInstance {
            fn children(self) -> _ {
                [self.instance_io.into(), self.module.into()][..].intern()
            }
            name: Interned<str>,
            instance_io: TraceBundle,
            module: TraceModule,
            ty: Bundle,
        }),
        Mem(TraceMem {
            fn children(self) -> _ {
                Interned::from_iter([*self.element_type].into_iter().chain(self.ports.iter().map(|&v| v.into())))
            }
            id: TraceMemoryId,
            name: Interned<str>,
            stride: usize,
            element_type: Interned<TraceDecl>,
            ports: Interned<[TraceMemPort]>,
            array_type: Array,
        }),
        MemPort(TraceMemPort {
            fn children(self) -> _ {
                [self.bundle.into()][..].intern()
            }
            name: Interned<str>,
            bundle: TraceBundle,
            ty: Bundle,
        }),
        Wire(TraceWire {
            fn children(self) -> _ {
                [*self.child][..].intern()
            }
            name: Interned<str>,
            child: Interned<TraceDecl>,
            ty: CanonicalType,
        }),
        Reg(TraceReg {
            fn children(self) -> _ {
                [*self.child][..].intern()
            }
            name: Interned<str>,
            child: Interned<TraceDecl>,
            ty: CanonicalType,
        }),
        ModuleIO(TraceModuleIO {
            fn children(self) -> _ {
                [*self.child][..].intern()
            }
            name: Interned<str>,
            child: Interned<TraceDecl>,
            ty: CanonicalType,
            flow: Flow,
        }),
        Bundle(TraceBundle {
            fn children(self) -> _ {
                self.fields
            }
            name: Interned<str>,
            fields: Interned<[TraceDecl]>,
            ty: Bundle,
            flow: Flow,
        }),
        Array(TraceArray {
            fn children(self) -> _ {
                self.elements
            }
            name: Interned<str>,
            elements: Interned<[TraceDecl]>,
            ty: Array,
            flow: Flow,
        }),
        EnumWithFields(TraceEnumWithFields {
            fn children(self) -> _ {
                Interned::from_iter([self.discriminant.into()].into_iter().chain(self.non_empty_fields))
            }
            name: Interned<str>,
            discriminant: TraceEnumDiscriminant,
            non_empty_fields: Interned<[TraceDecl]>,
            ty: Enum,
            flow: Flow,
        }),
    }),
    #[kind = TraceScalarKind]
    Scalar(TraceScalar {
        fn location(self) -> TraceLocation;
        UInt(TraceUInt {
            fn location(self) -> _ {
                self.location
            }
            location: TraceLocation,
            name: Interned<str>,
            ty: UInt,
            flow: Flow,
        }),
        SInt(TraceSInt {
            fn location(self) -> _ {
                self.location
            }
            location: TraceLocation,
            name: Interned<str>,
            ty: SInt,
            flow: Flow,
        }),
        Bool(TraceBool {
            fn location(self) -> _ {
                self.location
            }
            location: TraceLocation,
            name: Interned<str>,
            flow: Flow,
        }),
        FieldlessEnum(TraceFieldlessEnum {
            fn location(self) -> _ {
                self.location
            }
            location: TraceLocation,
            name: Interned<str>,
            ty: Enum,
            flow: Flow,
        }),
        EnumDiscriminant(TraceEnumDiscriminant {
            fn location(self) -> _ {
                self.location
            }
            location: TraceLocation,
            name: Interned<str>,
            ty: Enum,
            flow: Flow,
        }),
        Clock(TraceClock {
            fn location(self) -> _ {
                self.location
            }
            location: TraceLocation,
            name: Interned<str>,
            flow: Flow,
        }),
        SyncReset(TraceSyncReset {
            fn location(self) -> _ {
                self.location
            }
            location: TraceLocation,
            name: Interned<str>,
            flow: Flow,
        }),
        AsyncReset(TraceAsyncReset {
            fn location(self) -> _ {
                self.location
            }
            location: TraceLocation,
            name: Interned<str>,
            flow: Flow,
        }),
    }),
}

pub trait TraceWriterDecls: fmt::Debug + 'static + Sized {
    type Error: std::error::Error + Send + Sync + 'static;
    type TraceWriter: TraceWriter<Error = Self::Error>;
    fn write_decls(
        self,
        module: TraceModule,
        trace_scalar_id_count: usize,
        trace_memory_id_count: usize,
    ) -> Result<Self::TraceWriter, Self::Error>;
}

trait TraceWriterDeclsDynTrait: fmt::Debug {
    fn write_decls_dyn(
        self: Box<Self>,
        module: TraceModule,
        trace_scalar_id_count: usize,
        trace_memory_id_count: usize,
    ) -> std::io::Result<DynTraceWriter>;
}

fn err_into_io<E: std::error::Error + Send + Sync + 'static>(e: E) -> std::io::Error {
    match <dyn std::error::Error + Send + Sync>::downcast::<std::io::Error>(Box::new(e)) {
        Ok(retval) => *retval,
        Err(e) => std::io::Error::other(e),
    }
}

impl<T: TraceWriterDecls> TraceWriterDeclsDynTrait for T {
    fn write_decls_dyn(
        self: Box<Self>,
        module: TraceModule,
        trace_scalar_id_count: usize,
        trace_memory_id_count: usize,
    ) -> std::io::Result<DynTraceWriter> {
        Ok(DynTraceWriter(Box::new(
            TraceWriterDecls::write_decls(
                *self,
                module,
                trace_scalar_id_count,
                trace_memory_id_count,
            )
            .map_err(err_into_io)?,
        )))
    }
}

pub trait TraceWriter: fmt::Debug + 'static {
    type Error: std::error::Error + Send + Sync + 'static;
    fn finish_init(&mut self) -> Result<(), Self::Error> {
        Ok(())
    }
    fn change_time_to(&mut self, instant: SimInstant) -> Result<(), Self::Error> {
        let _ = instant;
        Ok(())
    }
    fn flush(&mut self) -> Result<(), Self::Error> {
        Ok(())
    }
    fn close(self) -> Result<(), Self::Error>
    where
        Self: Sized,
    {
        Ok(())
    }
    fn set_memory_element(
        &mut self,
        memory: TraceMemoryId,
        element_index: usize,
        element_data: &BitSlice,
    ) -> Result<(), Self::Error>;
    fn set_signal_uint(&mut self, id: TraceScalarId, value: &BitSlice) -> Result<(), Self::Error>;
    fn set_signal_sint(&mut self, id: TraceScalarId, value: &BitSlice) -> Result<(), Self::Error>;
    fn set_signal_bool(&mut self, id: TraceScalarId, value: bool) -> Result<(), Self::Error> {
        if value {
            self.set_signal_uint(id, bits![1])
        } else {
            self.set_signal_uint(id, bits![0])
        }
    }
    fn set_signal_clock(&mut self, id: TraceScalarId, value: bool) -> Result<(), Self::Error> {
        self.set_signal_bool(id, value)
    }
    fn set_signal_sync_reset(&mut self, id: TraceScalarId, value: bool) -> Result<(), Self::Error> {
        self.set_signal_bool(id, value)
    }
    fn set_signal_async_reset(
        &mut self,
        id: TraceScalarId,
        value: bool,
    ) -> Result<(), Self::Error> {
        self.set_signal_bool(id, value)
    }
    fn set_signal_enum_discriminant(
        &mut self,
        id: TraceScalarId,
        variant_index: usize,
        ty: Enum,
    ) -> Result<(), Self::Error>;
}

pub struct DynTraceWriterDecls(Box<dyn TraceWriterDeclsDynTrait>);

impl DynTraceWriterDecls {
    pub fn new<W: TraceWriterDecls>(writer: W) -> Self {
        Self(Box::new(writer))
    }
}

impl fmt::Debug for DynTraceWriterDecls {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.0.fmt(f)
    }
}

impl TraceWriterDecls for DynTraceWriterDecls {
    type Error = std::io::Error;
    type TraceWriter = DynTraceWriter;
    fn write_decls(
        self,
        module: TraceModule,
        trace_scalar_id_count: usize,
        trace_memory_id_count: usize,
    ) -> Result<Self::TraceWriter, Self::Error> {
        self.0
            .write_decls_dyn(module, trace_scalar_id_count, trace_memory_id_count)
    }
}

trait TraceWriterDynTrait: fmt::Debug + 'static {
    fn finish_init_dyn(&mut self) -> std::io::Result<()>;
    fn change_time_to_dyn(&mut self, instant: SimInstant) -> std::io::Result<()>;
    fn flush_dyn(&mut self) -> std::io::Result<()>;
    fn close_dyn(self: Box<Self>) -> std::io::Result<()>;
    fn set_memory_element_dyn(
        &mut self,
        memory: TraceMemoryId,
        element_index: usize,
        element_data: &BitSlice,
    ) -> std::io::Result<()>;
    fn set_signal_uint_dyn(&mut self, id: TraceScalarId, value: &BitSlice) -> std::io::Result<()>;
    fn set_signal_sint_dyn(&mut self, id: TraceScalarId, value: &BitSlice) -> std::io::Result<()>;
    fn set_signal_bool_dyn(&mut self, id: TraceScalarId, value: bool) -> std::io::Result<()>;
    fn set_signal_clock_dyn(&mut self, id: TraceScalarId, value: bool) -> std::io::Result<()>;
    fn set_signal_sync_reset_dyn(&mut self, id: TraceScalarId, value: bool) -> std::io::Result<()>;
    fn set_signal_async_reset_dyn(&mut self, id: TraceScalarId, value: bool)
        -> std::io::Result<()>;
    fn set_signal_enum_discriminant_dyn(
        &mut self,
        id: TraceScalarId,
        variant_index: usize,
        ty: Enum,
    ) -> std::io::Result<()>;
}

impl<T: TraceWriter> TraceWriterDynTrait for T {
    fn finish_init_dyn(&mut self) -> std::io::Result<()> {
        Ok(TraceWriter::finish_init(self).map_err(err_into_io)?)
    }
    fn change_time_to_dyn(&mut self, instant: SimInstant) -> std::io::Result<()> {
        Ok(TraceWriter::change_time_to(self, instant).map_err(err_into_io)?)
    }
    fn flush_dyn(&mut self) -> std::io::Result<()> {
        Ok(TraceWriter::flush(self).map_err(err_into_io)?)
    }
    fn close_dyn(self: Box<Self>) -> std::io::Result<()> {
        Ok(TraceWriter::close(*self).map_err(err_into_io)?)
    }
    fn set_memory_element_dyn(
        &mut self,
        memory: TraceMemoryId,
        element_index: usize,
        element_data: &BitSlice,
    ) -> std::io::Result<()> {
        Ok(
            TraceWriter::set_memory_element(self, memory, element_index, element_data)
                .map_err(err_into_io)?,
        )
    }
    fn set_signal_uint_dyn(&mut self, id: TraceScalarId, value: &BitSlice) -> std::io::Result<()> {
        Ok(TraceWriter::set_signal_uint(self, id, value).map_err(err_into_io)?)
    }
    fn set_signal_sint_dyn(&mut self, id: TraceScalarId, value: &BitSlice) -> std::io::Result<()> {
        Ok(TraceWriter::set_signal_sint(self, id, value).map_err(err_into_io)?)
    }
    fn set_signal_bool_dyn(&mut self, id: TraceScalarId, value: bool) -> std::io::Result<()> {
        Ok(TraceWriter::set_signal_bool(self, id, value).map_err(err_into_io)?)
    }
    fn set_signal_clock_dyn(&mut self, id: TraceScalarId, value: bool) -> std::io::Result<()> {
        Ok(TraceWriter::set_signal_clock(self, id, value).map_err(err_into_io)?)
    }
    fn set_signal_sync_reset_dyn(&mut self, id: TraceScalarId, value: bool) -> std::io::Result<()> {
        Ok(TraceWriter::set_signal_sync_reset(self, id, value).map_err(err_into_io)?)
    }
    fn set_signal_async_reset_dyn(
        &mut self,
        id: TraceScalarId,
        value: bool,
    ) -> std::io::Result<()> {
        Ok(TraceWriter::set_signal_async_reset(self, id, value).map_err(err_into_io)?)
    }
    fn set_signal_enum_discriminant_dyn(
        &mut self,
        id: TraceScalarId,
        variant_index: usize,
        ty: Enum,
    ) -> std::io::Result<()> {
        Ok(
            TraceWriter::set_signal_enum_discriminant(self, id, variant_index, ty)
                .map_err(err_into_io)?,
        )
    }
}

pub struct DynTraceWriter(Box<dyn TraceWriterDynTrait>);

impl fmt::Debug for DynTraceWriter {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.0.fmt(f)
    }
}

impl TraceWriter for DynTraceWriter {
    type Error = std::io::Error;
    fn finish_init(&mut self) -> Result<(), Self::Error> {
        self.0.finish_init_dyn()
    }
    fn flush(&mut self) -> Result<(), Self::Error> {
        self.0.flush_dyn()
    }
    fn close(self) -> Result<(), Self::Error> {
        self.0.close_dyn()
    }
    fn set_memory_element(
        &mut self,
        memory: TraceMemoryId,
        element_index: usize,
        element_data: &BitSlice,
    ) -> Result<(), Self::Error> {
        self.0
            .set_memory_element_dyn(memory, element_index, element_data)
    }
    fn change_time_to(&mut self, instant: SimInstant) -> Result<(), Self::Error> {
        self.0.change_time_to_dyn(instant)
    }
    fn set_signal_uint(&mut self, id: TraceScalarId, value: &BitSlice) -> Result<(), Self::Error> {
        self.0.set_signal_uint_dyn(id, value)
    }
    fn set_signal_sint(&mut self, id: TraceScalarId, value: &BitSlice) -> Result<(), Self::Error> {
        self.0.set_signal_sint_dyn(id, value)
    }
    fn set_signal_bool(&mut self, id: TraceScalarId, value: bool) -> Result<(), Self::Error> {
        self.0.set_signal_bool_dyn(id, value)
    }
    fn set_signal_clock(&mut self, id: TraceScalarId, value: bool) -> Result<(), Self::Error> {
        self.0.set_signal_clock_dyn(id, value)
    }
    fn set_signal_sync_reset(&mut self, id: TraceScalarId, value: bool) -> Result<(), Self::Error> {
        self.0.set_signal_sync_reset_dyn(id, value)
    }
    fn set_signal_async_reset(
        &mut self,
        id: TraceScalarId,
        value: bool,
    ) -> Result<(), Self::Error> {
        self.0.set_signal_async_reset_dyn(id, value)
    }
    fn set_signal_enum_discriminant(
        &mut self,
        id: TraceScalarId,
        variant_index: usize,
        ty: Enum,
    ) -> Result<(), Self::Error> {
        self.0
            .set_signal_enum_discriminant_dyn(id, variant_index, ty)
    }
}

#[derive(Debug)]
enum TraceWriterState<T: TraceWriterDecls> {
    Decls(T),
    Init(T::TraceWriter),
    Running(T::TraceWriter),
    Errored(Option<T::Error>),
}

trait SimTraceDebug<I> {
    fn fmt(&self, id: I, f: &mut fmt::Formatter<'_>) -> fmt::Result;
}

struct SimTraceDebugAsDebug<T, I>(T, I);

impl<T: ?Sized, I: Copy> fmt::Debug for SimTraceDebugAsDebug<&'_ T, I>
where
    T: SimTraceDebug<I>,
{
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.0.fmt(self.1, f)
    }
}

impl<T, I> SimTraceDebug<I> for Vec<T>
where
    [T]: SimTraceDebug<I>,
{
    fn fmt(&self, id: I, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        <[T]>::fmt(&**self, id, f)
    }
}

impl<T: ?Sized + 'static + Send + Sync, I> SimTraceDebug<I> for Interned<T>
where
    T: SimTraceDebug<I>,
{
    fn fmt(&self, id: I, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        T::fmt(&**self, id, f)
    }
}

impl<T: ?Sized + 'static + Send + Sync, I> SimTraceDebug<I> for Box<T>
where
    T: SimTraceDebug<I>,
{
    fn fmt(&self, id: I, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        T::fmt(&**self, id, f)
    }
}

impl<T> SimTraceDebug<()> for [T]
where
    T: SimTraceDebug<TraceScalarId>,
{
    fn fmt(&self, _id: (), f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.debug_list()
            .entries(
                self.iter()
                    .enumerate()
                    .map(|(id, v)| SimTraceDebugAsDebug(v, TraceScalarId(id))),
            )
            .finish()
    }
}

#[derive(Clone, PartialEq, Eq, Hash, Debug)]
struct SimTrace<K, S> {
    kind: K,
    state: S,
    last_state: S,
}

#[derive(Copy, Clone, PartialEq, Eq, Hash)]
struct SimTraces<T>(T);

impl<T> fmt::Debug for SimTraces<T>
where
    T: SimTraceDebug<()>,
{
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.0.fmt((), f)
    }
}

impl<K: fmt::Debug> SimTraceDebug<TraceScalarId> for SimTrace<K, ()> {
    fn fmt(&self, id: TraceScalarId, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self {
            kind,
            state,
            last_state,
        } = self;
        f.debug_struct("SimTrace")
            .field("id", &id)
            .field("kind", kind)
            .field("state", state)
            .field("last_state", last_state)
            .finish()
    }
}

impl<K: fmt::Debug> SimTraceDebug<TraceScalarId> for SimTrace<K, BitVec> {
    fn fmt(&self, id: TraceScalarId, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self {
            kind,
            state,
            last_state,
        } = self;
        f.debug_struct("SimTrace")
            .field("id", &id)
            .field("kind", kind)
            .field("state", &BitSliceWriteWithBase(state))
            .field("last_state", &BitSliceWriteWithBase(last_state))
            .finish()
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
enum SimTraceKind {
    BigUInt {
        index: StatePartIndex<StatePartKindBigSlots>,
        ty: UInt,
    },
    BigSInt {
        index: StatePartIndex<StatePartKindBigSlots>,
        ty: SInt,
    },
    BigBool {
        index: StatePartIndex<StatePartKindBigSlots>,
    },
    BigAsyncReset {
        index: StatePartIndex<StatePartKindBigSlots>,
    },
    BigSyncReset {
        index: StatePartIndex<StatePartKindBigSlots>,
    },
    BigClock {
        index: StatePartIndex<StatePartKindBigSlots>,
    },
    SmallUInt {
        index: StatePartIndex<StatePartKindSmallSlots>,
        ty: UInt,
    },
    SmallSInt {
        index: StatePartIndex<StatePartKindSmallSlots>,
        ty: SInt,
    },
    SmallBool {
        index: StatePartIndex<StatePartKindSmallSlots>,
    },
    SmallAsyncReset {
        index: StatePartIndex<StatePartKindSmallSlots>,
    },
    SmallSyncReset {
        index: StatePartIndex<StatePartKindSmallSlots>,
    },
    SmallClock {
        index: StatePartIndex<StatePartKindSmallSlots>,
    },
    EnumDiscriminant {
        index: StatePartIndex<StatePartKindSmallSlots>,
        ty: Enum,
    },
}

impl SimTraceKind {
    fn make_state(self) -> BitVec {
        match self {
            SimTraceKind::BigUInt { index: _, ty } | SimTraceKind::SmallUInt { index: _, ty } => {
                BitVec::repeat(false, ty.width)
            }
            SimTraceKind::BigSInt { index: _, ty } | SimTraceKind::SmallSInt { index: _, ty } => {
                BitVec::repeat(false, ty.width)
            }
            SimTraceKind::BigBool { index: _ }
            | SimTraceKind::BigAsyncReset { index: _ }
            | SimTraceKind::BigSyncReset { index: _ }
            | SimTraceKind::BigClock { index: _ }
            | SimTraceKind::SmallBool { index: _ }
            | SimTraceKind::SmallAsyncReset { index: _ }
            | SimTraceKind::SmallSyncReset { index: _ }
            | SimTraceKind::SmallClock { index: _ } => BitVec::repeat(false, 1),
            SimTraceKind::EnumDiscriminant { index: _, ty } => {
                BitVec::repeat(false, ty.discriminant_bit_width())
            }
        }
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
enum MaybeNeedsSettle<S, N = S> {
    NeedsSettle(S),
    NoSettleNeeded(N),
}

impl<T> MaybeNeedsSettle<T> {
    fn map<U>(self, f: impl FnOnce(T) -> U) -> MaybeNeedsSettle<U> {
        match self {
            MaybeNeedsSettle::NeedsSettle(v) => MaybeNeedsSettle::NeedsSettle(f(v)),
            MaybeNeedsSettle::NoSettleNeeded(v) => MaybeNeedsSettle::NoSettleNeeded(f(v)),
        }
    }
}

// workaround implementing FnOnce not being stable
trait MaybeNeedsSettleFn<A> {
    type Output;

    fn call(self, arg: A) -> Self::Output;
}

impl<T: FnOnce(A) -> O, A, O> MaybeNeedsSettleFn<A> for T {
    type Output = O;

    fn call(self, arg: A) -> Self::Output {
        self(arg)
    }
}

impl<S, N> MaybeNeedsSettle<S, N> {
    fn apply_no_settle<T>(self, arg: T) -> MaybeNeedsSettle<S, N::Output>
    where
        N: MaybeNeedsSettleFn<T>,
    {
        match self {
            MaybeNeedsSettle::NeedsSettle(v) => MaybeNeedsSettle::NeedsSettle(v),
            MaybeNeedsSettle::NoSettleNeeded(v) => MaybeNeedsSettle::NoSettleNeeded(v.call(arg)),
        }
    }
}

struct SimulationModuleState {
    base_targets: Vec<Target>,
    uninitialized_ios: HashMap<Target, Vec<Target>>,
    io_targets: HashMap<Target, CompiledValue<CanonicalType>>,
    did_initial_settle: bool,
}

impl fmt::Debug for SimulationModuleState {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self {
            base_targets,
            uninitialized_ios,
            io_targets,
            did_initial_settle,
        } = self;
        f.debug_struct("SimulationModuleState")
            .field("base_targets", base_targets)
            .field("uninitialized_ios", &SortedSetDebug(uninitialized_ios))
            .field("io_targets", &SortedSetDebug(io_targets))
            .field("did_initial_settle", did_initial_settle)
            .finish()
    }
}

impl SimulationModuleState {
    fn new(base_targets: impl IntoIterator<Item = (Target, CompiledValue<CanonicalType>)>) -> Self {
        let mut retval = Self {
            base_targets: Vec::new(),
            uninitialized_ios: HashMap::default(),
            io_targets: HashMap::default(),
            did_initial_settle: false,
        };
        for (base_target, value) in base_targets {
            retval.base_targets.push(base_target);
            retval.parse_io(base_target, value);
        }
        retval
    }
    /// returns `true` if `target` or any sub-targets are uninitialized inputs
    fn parse_io(&mut self, target: Target, value: CompiledValue<CanonicalType>) -> bool {
        self.io_targets.insert(target, value);
        match value.layout.body {
            CompiledTypeLayoutBody::Scalar => match target.flow() {
                Flow::Source => false,
                Flow::Sink => {
                    self.uninitialized_ios.insert(target, vec![]);
                    true
                }
                Flow::Duplex => unreachable!(),
            },
            CompiledTypeLayoutBody::Array { .. } => {
                let value = value.map_ty(Array::from_canonical);
                let mut sub_targets = Vec::new();
                for index in 0..value.layout.ty.len() {
                    let sub_target = target.join(
                        TargetPathElement::from(TargetPathArrayElement { index }).intern_sized(),
                    );
                    if self.parse_io(sub_target, value.element(index)) {
                        sub_targets.push(sub_target);
                    }
                }
                if sub_targets.is_empty() {
                    false
                } else {
                    self.uninitialized_ios.insert(target, sub_targets);
                    true
                }
            }
            CompiledTypeLayoutBody::Bundle { .. } => {
                let value = value.map_ty(Bundle::from_canonical);
                let mut sub_targets = Vec::new();
                for BundleField { name, .. } in value.layout.ty.fields() {
                    let sub_target = target.join(
                        TargetPathElement::from(TargetPathBundleField { name }).intern_sized(),
                    );
                    if self.parse_io(sub_target, value.field_by_name(name)) {
                        sub_targets.push(sub_target);
                    }
                }
                if sub_targets.is_empty() {
                    false
                } else {
                    self.uninitialized_ios.insert(target, sub_targets);
                    true
                }
            }
        }
    }
    fn mark_target_as_initialized(&mut self, mut target: Target) {
        fn remove_target_and_children(
            uninitialized_ios: &mut HashMap<Target, Vec<Target>>,
            target: Target,
        ) {
            let Some(children) = uninitialized_ios.remove(&target) else {
                return;
            };
            for child in children {
                remove_target_and_children(uninitialized_ios, child);
            }
        }
        remove_target_and_children(&mut self.uninitialized_ios, target);
        while let Some(target_child) = target.child() {
            let parent = target_child.parent();
            for child in self
                .uninitialized_ios
                .get(&*parent)
                .map(|v| &**v)
                .unwrap_or(&[])
            {
                if self.uninitialized_ios.contains_key(child) {
                    return;
                }
            }
            target = *parent;
            self.uninitialized_ios.remove(&target);
        }
    }
    #[track_caller]
    fn get_io(
        &self,
        mut target: Target,
        which_module: WhichModule,
    ) -> CompiledValue<CanonicalType> {
        assert!(
            target.canonical_ty().is_passive(),
            "simulator read/write expression must have a passive type \
            (recursively contains no fields with `#[hdl(flip)]`)"
        );
        if let Some(&retval) = self.io_targets.get(&target) {
            return retval;
        }
        loop {
            target = match target {
                Target::Base(_) => break,
                Target::Child(child) => {
                    match *child.path_element() {
                        TargetPathElement::BundleField(_) | TargetPathElement::ArrayElement(_) => {}
                        TargetPathElement::DynArrayElement(_) => panic!(
                            "simulator read/write expression must not have dynamic array indexes"
                        ),
                    }
                    *child.parent()
                }
            };
        }
        match which_module {
            WhichModule::Main => panic!(
                "simulator read/write expression must be \
                an array element/field of `Simulation::io()`"
            ),
            WhichModule::Extern { .. } => panic!(
                "simulator read/write expression must be \
                one of this module's inputs/outputs or an \
                array element/field of one of this module's inputs/outputs"
            ),
        }
    }
    #[track_caller]
    fn read_helper(
        &self,
        io: Expr<CanonicalType>,
        which_module: WhichModule,
    ) -> MaybeNeedsSettle<CompiledValue<CanonicalType>> {
        let Some(target) = io.target() else {
            match which_module {
                WhichModule::Main => panic!(
                    "can't read from an expression that's not a field/element of `Simulation::io()`"
                ),
                WhichModule::Extern { .. } => panic!(
                    "can't read from an expression that's not based on one of this module's inputs/outputs"
                ),
            }
        };
        let compiled_value = self.get_io(*target, which_module);
        match target.flow() {
            Flow::Source => {
                if !self.uninitialized_ios.is_empty() {
                    match which_module {
                        WhichModule::Main => {
                            panic!("can't read from an output before initializing all inputs");
                        }
                        WhichModule::Extern { .. } => {
                            panic!("can't read from an input before initializing all outputs");
                        }
                    }
                }
                MaybeNeedsSettle::NeedsSettle(compiled_value)
            }
            Flow::Sink => {
                if self.uninitialized_ios.contains_key(&*target) {
                    match which_module {
                        WhichModule::Main => panic!("can't read from an uninitialized input"),
                        WhichModule::Extern { .. } => {
                            panic!("can't read from an uninitialized output");
                        }
                    }
                }
                MaybeNeedsSettle::NoSettleNeeded(compiled_value)
            }
            Flow::Duplex => unreachable!(),
        }
    }
    #[track_caller]
    fn write_helper(
        &mut self,
        io: Expr<CanonicalType>,
        which_module: WhichModule,
    ) -> CompiledValue<CanonicalType> {
        let Some(target) = io.target() else {
            match which_module {
                WhichModule::Main => panic!(
                    "can't write to an expression that's not a field/element of `Simulation::io()`"
                ),
                WhichModule::Extern { .. } => panic!(
                    "can't write to an expression that's not based on one of this module's outputs"
                ),
            }
        };
        let compiled_value = self.get_io(*target, which_module);
        match target.flow() {
            Flow::Source => match which_module {
                WhichModule::Main => panic!("can't write to an output"),
                WhichModule::Extern { .. } => panic!("can't write to an input"),
            },
            Flow::Sink => {}
            Flow::Duplex => unreachable!(),
        }
        if !self.did_initial_settle {
            self.mark_target_as_initialized(*target);
        }
        compiled_value
    }
}

#[derive(Copy, Clone, Debug)]
enum WaitTarget<ChangeKey, ChangeValue> {
    Settle,
    Instant(SimInstant),
    Change { key: ChangeKey, value: ChangeValue },
}

#[derive(Clone)]
struct EarliestWaitTargets {
    settle: bool,
    instant: Option<SimInstant>,
    changes: HashMap<CompiledValue<CanonicalType>, SimValue<CanonicalType>>,
}

impl fmt::Debug for EarliestWaitTargets {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.debug_set().entries(self.iter()).finish()
    }
}

impl Default for EarliestWaitTargets {
    fn default() -> Self {
        Self {
            settle: false,
            instant: None,
            changes: HashMap::default(),
        }
    }
}

impl EarliestWaitTargets {
    fn settle() -> Self {
        Self {
            settle: true,
            instant: None,
            changes: HashMap::default(),
        }
    }
    fn instant(instant: SimInstant) -> Self {
        Self {
            settle: false,
            instant: Some(instant),
            changes: HashMap::default(),
        }
    }
    fn len(&self) -> usize {
        self.settle as usize + self.instant.is_some() as usize + self.changes.len()
    }
    fn is_empty(&self) -> bool {
        self.len() == 0
    }
    fn clear(&mut self) {
        let Self {
            settle,
            instant,
            changes,
        } = self;
        *settle = false;
        *instant = None;
        changes.clear();
    }
    fn insert<ChangeValue>(
        &mut self,
        value: impl std::borrow::Borrow<WaitTarget<CompiledValue<CanonicalType>, ChangeValue>>,
    ) where
        ChangeValue: std::borrow::Borrow<SimValue<CanonicalType>>,
    {
        let value = value.borrow();
        match value {
            WaitTarget::Settle => self.settle = true,
            WaitTarget::Instant(instant) => {
                if self.instant.is_none_or(|v| v > *instant) {
                    self.instant = Some(*instant);
                }
            }
            WaitTarget::Change { key, value } => {
                self.changes
                    .entry(*key)
                    .or_insert_with(|| value.borrow().clone());
            }
        }
    }
    fn convert_earlier_instants_to_settle(&mut self, instant: SimInstant) {
        if self.instant.is_some_and(|v| v <= instant) {
            self.settle = true;
            self.instant = None;
        }
    }
    fn iter<'a>(
        &'a self,
    ) -> impl Clone
           + Iterator<Item = WaitTarget<CompiledValue<CanonicalType>, &'a SimValue<CanonicalType>>>
           + 'a {
        self.settle
            .then_some(WaitTarget::Settle)
            .into_iter()
            .chain(self.instant.map(|instant| WaitTarget::Instant(instant)))
            .chain(
                self.changes
                    .iter()
                    .map(|(&key, value)| WaitTarget::Change { key, value }),
            )
    }
}

impl<ChangeValue: std::borrow::Borrow<SimValue<CanonicalType>>>
    Extend<WaitTarget<CompiledValue<CanonicalType>, ChangeValue>> for EarliestWaitTargets
{
    fn extend<T: IntoIterator<Item = WaitTarget<CompiledValue<CanonicalType>, ChangeValue>>>(
        &mut self,
        iter: T,
    ) {
        iter.into_iter().for_each(|v| self.insert(v))
    }
}

impl<'a, ChangeValue: std::borrow::Borrow<SimValue<CanonicalType>>>
    Extend<&'a WaitTarget<CompiledValue<CanonicalType>, ChangeValue>> for EarliestWaitTargets
{
    fn extend<T: IntoIterator<Item = &'a WaitTarget<CompiledValue<CanonicalType>, ChangeValue>>>(
        &mut self,
        iter: T,
    ) {
        iter.into_iter().for_each(|v| self.insert(v))
    }
}

impl<A> FromIterator<A> for EarliestWaitTargets
where
    Self: Extend<A>,
{
    fn from_iter<T: IntoIterator<Item = A>>(iter: T) -> Self {
        let mut retval = Self::default();
        retval.extend(iter);
        retval
    }
}

struct SimulationExternModuleState {
    module_state: SimulationModuleState,
    sim: ExternModuleSimulation,
    running_generator: Option<Pin<Box<dyn Future<Output = ()> + 'static>>>,
    wait_targets: EarliestWaitTargets,
}

impl fmt::Debug for SimulationExternModuleState {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self {
            module_state,
            sim,
            running_generator,
            wait_targets,
        } = self;
        f.debug_struct("SimulationExternModuleState")
            .field("module_state", module_state)
            .field("sim", sim)
            .field(
                "running_generator",
                &running_generator.as_ref().map(|_| DebugAsDisplay("...")),
            )
            .field("wait_targets", wait_targets)
            .finish()
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
enum WhichModule {
    Main,
    Extern { module_index: usize },
}

struct ReadBitFn {
    compiled_value: CompiledValue<CanonicalType>,
}

impl MaybeNeedsSettleFn<&'_ mut interpreter::State> for ReadBitFn {
    type Output = bool;

    fn call(self, state: &mut interpreter::State) -> Self::Output {
        match self.compiled_value.range.len() {
            TypeLen::A_SMALL_SLOT => {
                state.small_slots[self.compiled_value.range.small_slots.start] != 0
            }
            TypeLen::A_BIG_SLOT => !state.big_slots[self.compiled_value.range.big_slots.start]
                .clone()
                .is_zero(),
            _ => unreachable!(),
        }
    }
}

struct ReadBoolOrIntFn<I: BoolOrIntType> {
    compiled_value: CompiledValue<CanonicalType>,
    io: Expr<I>,
}

impl<I: BoolOrIntType> MaybeNeedsSettleFn<&'_ mut interpreter::State> for ReadBoolOrIntFn<I> {
    type Output = I::Value;

    fn call(self, state: &mut interpreter::State) -> Self::Output {
        let Self { compiled_value, io } = self;
        match compiled_value.range.len() {
            TypeLen::A_SMALL_SLOT => Expr::ty(io)
                .value_from_int_wrapping(state.small_slots[compiled_value.range.small_slots.start]),
            TypeLen::A_BIG_SLOT => Expr::ty(io).value_from_int_wrapping(
                state.big_slots[compiled_value.range.big_slots.start].clone(),
            ),
            _ => unreachable!(),
        }
    }
}

struct ReadFn {
    compiled_value: CompiledValue<CanonicalType>,
    io: Expr<CanonicalType>,
}

impl MaybeNeedsSettleFn<&'_ mut interpreter::State> for ReadFn {
    type Output = SimValue<CanonicalType>;

    fn call(self, state: &mut interpreter::State) -> Self::Output {
        let Self { compiled_value, io } = self;
        SimulationImpl::read_no_settle_helper(
            state,
            io,
            compiled_value,
            UIntValue::new(Arc::new(BitVec::repeat(false, Expr::ty(io).bit_width()))),
        )
    }
}

struct GeneratorWaker;

impl std::task::Wake for GeneratorWaker {
    fn wake(self: Arc<Self>) {
        panic!("can't await other kinds of futures in function passed to ExternalModuleSimulation");
    }
}

#[derive(Default)]
struct ReadyToRunSet {
    state_ready_to_run: bool,
    extern_modules_ready_to_run: Vec<usize>,
}

impl ReadyToRunSet {
    fn clear(&mut self) {
        let Self {
            state_ready_to_run,
            extern_modules_ready_to_run,
        } = self;
        *state_ready_to_run = false;
        extern_modules_ready_to_run.clear();
    }
}

struct SimulationImpl {
    state: interpreter::State,
    io: Expr<Bundle>,
    main_module: SimulationModuleState,
    extern_modules: Box<[SimulationExternModuleState]>,
    state_ready_to_run: bool,
    trace_decls: TraceModule,
    traces: SimTraces<Box<[SimTrace<SimTraceKind, BitVec>]>>,
    trace_memories: HashMap<StatePartIndex<StatePartKindMemories>, TraceMem>,
    trace_writers: Vec<TraceWriterState<DynTraceWriterDecls>>,
    instant: SimInstant,
    clocks_triggered: Interned<[StatePartIndex<StatePartKindSmallSlots>]>,
    breakpoints: Option<BreakpointsSet>,
    generator_waker: std::task::Waker,
}

impl fmt::Debug for SimulationImpl {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.debug_fmt(None, f)
    }
}

impl SimulationImpl {
    fn debug_fmt(&self, io: Option<&dyn fmt::Debug>, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self {
            state,
            io: self_io,
            main_module,
            extern_modules,
            state_ready_to_run,
            trace_decls,
            traces,
            trace_memories,
            trace_writers,
            instant,
            clocks_triggered,
            breakpoints: _,
            generator_waker: _,
        } = self;
        f.debug_struct("Simulation")
            .field("state", state)
            .field("io", io.unwrap_or(self_io))
            .field("main_module", main_module)
            .field("extern_modules", extern_modules)
            .field("state_ready_to_run", state_ready_to_run)
            .field("trace_decls", trace_decls)
            .field("traces", traces)
            .field("trace_memories", trace_memories)
            .field("trace_writers", trace_writers)
            .field("instant", instant)
            .field("clocks_triggered", clocks_triggered)
            .finish_non_exhaustive()
    }
    fn new(compiled: Compiled<Bundle>) -> Self {
        let io_target = Target::from(compiled.io);
        let extern_modules = Box::from_iter(compiled.extern_modules.iter().map(
            |&CompiledExternModule {
                 module_io_targets,
                 module_io,
                 simulation,
             }| {
                SimulationExternModuleState {
                    module_state: SimulationModuleState::new(
                        module_io_targets
                            .iter()
                            .copied()
                            .zip(module_io.iter().copied()),
                    ),
                    sim: simulation,
                    running_generator: None,
                    wait_targets: EarliestWaitTargets::settle(),
                }
            },
        ));
        Self {
            state: State::new(compiled.insns),
            io: compiled.io.to_expr(),
            main_module: SimulationModuleState::new(
                compiled
                    .io
                    .ty()
                    .fields()
                    .into_iter()
                    .zip(compiled.base_module.module_io)
                    .map(|(BundleField { name, .. }, value)| {
                        (
                            io_target.join(
                                TargetPathElement::from(TargetPathBundleField { name })
                                    .intern_sized(),
                            ),
                            value,
                        )
                    }),
            ),
            extern_modules,
            state_ready_to_run: true,
            trace_decls: compiled.base_module.trace_decls,
            traces: SimTraces(Box::from_iter(compiled.traces.0.iter().map(
                |&SimTrace {
                     kind,
                     state: _,
                     last_state: _,
                 }| SimTrace {
                    kind,
                    state: kind.make_state(),
                    last_state: kind.make_state(),
                },
            ))),
            trace_memories: HashMap::from_iter(compiled.trace_memories.iter().copied()),
            trace_writers: vec![],
            instant: SimInstant::START,
            clocks_triggered: compiled.clocks_triggered,
            breakpoints: None,
            generator_waker: Arc::new(GeneratorWaker).into(),
        }
    }
    fn write_traces<const ONLY_IF_CHANGED: bool>(
        &mut self,
        mut trace_writer: DynTraceWriter,
    ) -> std::io::Result<DynTraceWriter> {
        let mut set_memory_element = |memory: StatePartIndex<StatePartKindMemories>,
                                      trace_mem: &TraceMem,
                                      element_index: usize| {
            let start = trace_mem.stride * element_index;
            let end = start + trace_mem.stride;
            trace_writer.set_memory_element(
                self.trace_memories[&memory].id,
                element_index,
                &self.state.memories[memory].data[start..end],
            )
        };
        if ONLY_IF_CHANGED {
            for &(memory, element_index) in &self.state.memory_write_log {
                set_memory_element(memory, &self.trace_memories[&memory], element_index)?;
            }
        } else {
            for (&memory, trace_mem) in &self.trace_memories {
                for element_index in 0..trace_mem.array_type.len() {
                    set_memory_element(memory, trace_mem, element_index)?;
                }
            }
        }
        for (
            id,
            &SimTrace {
                kind,
                ref state,
                ref last_state,
            },
        ) in self.traces.0.iter().enumerate()
        {
            if ONLY_IF_CHANGED && state == last_state {
                continue;
            }
            let id = TraceScalarId(id);
            match kind {
                SimTraceKind::BigUInt { .. } | SimTraceKind::SmallUInt { .. } => {
                    trace_writer.set_signal_uint(id, state)?;
                }
                SimTraceKind::BigSInt { .. } | SimTraceKind::SmallSInt { .. } => {
                    trace_writer.set_signal_sint(id, state)?;
                }
                SimTraceKind::BigBool { .. } | SimTraceKind::SmallBool { .. } => {
                    trace_writer.set_signal_bool(id, state[0])?;
                }
                SimTraceKind::BigAsyncReset { .. } | SimTraceKind::SmallAsyncReset { .. } => {
                    trace_writer.set_signal_async_reset(id, state[0])?;
                }
                SimTraceKind::BigSyncReset { .. } | SimTraceKind::SmallSyncReset { .. } => {
                    trace_writer.set_signal_sync_reset(id, state[0])?;
                }
                SimTraceKind::BigClock { .. } | SimTraceKind::SmallClock { .. } => {
                    trace_writer.set_signal_clock(id, state[0])?;
                }
                SimTraceKind::EnumDiscriminant { ty, .. } => {
                    let mut variant_index = [0; mem::size_of::<usize>()];
                    variant_index.view_bits_mut::<Lsb0>()[0..state.len()]
                        .clone_from_bitslice(state);
                    trace_writer.set_signal_enum_discriminant(
                        id,
                        usize::from_le_bytes(variant_index),
                        ty,
                    )?;
                }
            }
        }
        Ok(trace_writer)
    }
    fn init_trace_writer(
        &mut self,
        trace_writer: DynTraceWriter,
    ) -> std::io::Result<DynTraceWriter> {
        let mut trace_writer = self.write_traces::<false>(trace_writer)?;
        trace_writer.finish_init()?;
        Ok(trace_writer)
    }
    fn update_trace_writer(
        &mut self,
        trace_writer: DynTraceWriter,
    ) -> std::io::Result<DynTraceWriter> {
        self.write_traces::<true>(trace_writer)
    }
    #[inline(never)]
    fn read_traces<const IS_INITIAL_STEP: bool>(&mut self) {
        for &mut SimTrace {
            kind,
            ref mut state,
            ref mut last_state,
        } in &mut self.traces.0
        {
            if !IS_INITIAL_STEP {
                mem::swap(state, last_state);
            }
            match kind {
                SimTraceKind::BigUInt { index, ty: _ } | SimTraceKind::BigSInt { index, ty: _ } => {
                    let bigint = &self.state.big_slots[index];
                    let mut bytes = bigint.to_signed_bytes_le();
                    bytes.resize(
                        state.len().div_ceil(8),
                        if bigint.is_negative() { 0xFF } else { 0 },
                    );
                    let bitslice = BitSlice::<u8, Lsb0>::from_slice(&bytes);
                    let bitslice = &bitslice[..state.len()];
                    state.clone_from_bitslice(bitslice);
                }
                SimTraceKind::BigBool { index }
                | SimTraceKind::BigAsyncReset { index }
                | SimTraceKind::BigSyncReset { index }
                | SimTraceKind::BigClock { index } => {
                    state.set(0, !self.state.big_slots[index].is_zero());
                }
                SimTraceKind::SmallUInt { index, ty: _ }
                | SimTraceKind::SmallSInt { index, ty: _ }
                | SimTraceKind::EnumDiscriminant { index, ty: _ } => {
                    let bytes = self.state.small_slots[index].to_le_bytes();
                    let bitslice = BitSlice::<u8, Lsb0>::from_slice(&bytes);
                    let bitslice = &bitslice[..state.len()];
                    state.clone_from_bitslice(bitslice);
                }
                SimTraceKind::SmallBool { index }
                | SimTraceKind::SmallAsyncReset { index }
                | SimTraceKind::SmallSyncReset { index }
                | SimTraceKind::SmallClock { index } => {
                    state.set(0, self.state.small_slots[index] != 0);
                }
            }
            if IS_INITIAL_STEP {
                last_state.copy_from_bitslice(state);
            }
        }
    }
    #[track_caller]
    fn advance_time(this_ref: &Rc<RefCell<Self>>, duration: SimDuration) {
        let run_target = this_ref.borrow().instant + duration;
        Self::run_until(this_ref, run_target);
    }
    /// clears `targets`
    #[must_use]
    fn yield_wait<'a>(
        this: Rc<RefCell<Self>>,
        module_index: usize,
        targets: &'a mut EarliestWaitTargets,
    ) -> impl Future<Output = ()> + 'a {
        struct MyGenerator<'a> {
            sim: Rc<RefCell<SimulationImpl>>,
            yielded_at_all: bool,
            module_index: usize,
            targets: &'a mut EarliestWaitTargets,
        }
        impl Future for MyGenerator<'_> {
            type Output = ();

            fn poll(
                mut self: Pin<&mut Self>,
                cx: &mut std::task::Context<'_>,
            ) -> Poll<Self::Output> {
                let this = &mut *self;
                let mut sim = this.sim.borrow_mut();
                let sim = &mut *sim;
                assert!(cx.waker().will_wake(&sim.generator_waker), "can't use ExternModuleSimulationState's methods outside of ExternModuleSimulation");
                this.targets.convert_earlier_instants_to_settle(sim.instant);
                if this.targets.is_empty() {
                    this.targets.settle = true;
                }
                if this.targets.settle {
                    if this.yielded_at_all {
                        this.targets.clear();
                        return Poll::Ready(());
                    }
                }
                sim.extern_modules[this.module_index]
                    .wait_targets
                    .extend(this.targets.iter());
                this.targets.clear();
                this.yielded_at_all = true;
                Poll::Pending
            }
        }
        MyGenerator {
            sim: this,
            yielded_at_all: false,
            module_index,
            targets,
        }
    }
    async fn yield_advance_time_or_settle(
        this: Rc<RefCell<Self>>,
        module_index: usize,
        duration: Option<SimDuration>,
    ) {
        let mut targets = duration.map_or(EarliestWaitTargets::settle(), |duration| {
            EarliestWaitTargets::instant(this.borrow().instant + duration)
        });
        Self::yield_wait(this, module_index, &mut targets).await;
    }
    fn is_extern_module_ready_to_run(&mut self, module_index: usize) -> Option<SimInstant> {
        let module = &self.extern_modules[module_index];
        let mut retval = None;
        for wait_target in module.wait_targets.iter() {
            retval = match (wait_target, retval) {
                (WaitTarget::Settle, _) => Some(self.instant),
                (WaitTarget::Instant(instant), _) if instant <= self.instant => Some(self.instant),
                (WaitTarget::Instant(instant), None) => Some(instant),
                (WaitTarget::Instant(instant), Some(retval)) => Some(instant.min(retval)),
                (WaitTarget::Change { key, value }, retval) => {
                    if Self::value_changed(&mut self.state, key, SimValue::bits(value).bits()) {
                        Some(self.instant)
                    } else {
                        retval
                    }
                }
            };
            if retval == Some(self.instant) {
                break;
            }
        }
        retval
    }
    fn get_ready_to_run_set(&mut self, ready_to_run_set: &mut ReadyToRunSet) -> Option<SimInstant> {
        ready_to_run_set.clear();
        let mut retval = None;
        if self.state_ready_to_run {
            ready_to_run_set.state_ready_to_run = true;
            retval = Some(self.instant);
        }
        for module_index in 0..self.extern_modules.len() {
            let Some(instant) = self.is_extern_module_ready_to_run(module_index) else {
                continue;
            };
            if let Some(retval) = &mut retval {
                match instant.cmp(retval) {
                    std::cmp::Ordering::Less => ready_to_run_set.clear(),
                    std::cmp::Ordering::Equal => {}
                    std::cmp::Ordering::Greater => continue,
                }
            } else {
                retval = Some(instant);
            }
            ready_to_run_set
                .extern_modules_ready_to_run
                .push(module_index);
        }
        retval
    }
    fn set_instant_no_sim(&mut self, instant: SimInstant) {
        self.instant = instant;
        self.for_each_trace_writer_storing_error(|this, mut trace_writer_state| {
            match &mut trace_writer_state {
                TraceWriterState::Decls(_) | TraceWriterState::Init(_) => unreachable!(),
                TraceWriterState::Running(trace_writer) => {
                    trace_writer.change_time_to(this.instant)?;
                }
                TraceWriterState::Errored(_) => {}
            }
            Ok(trace_writer_state)
        });
    }
    #[must_use]
    #[track_caller]
    fn run_state_settle_cycle(&mut self) -> bool {
        self.state_ready_to_run = false;
        self.state.setup_call(0);
        if self.breakpoints.is_some() {
            loop {
                match self
                    .state
                    .run(self.breakpoints.as_mut().expect("just checked"))
                {
                    RunResult::Break(break_action) => {
                        println!(
                            "hit breakpoint at:\n{:?}",
                            self.state.debug_insn_at(self.state.pc),
                        );
                        match break_action {
                            BreakAction::DumpStateAndContinue => {
                                println!("{self:#?}");
                            }
                            BreakAction::Continue => {}
                        }
                    }
                    RunResult::Return(()) => break,
                }
            }
        } else {
            let RunResult::Return(()) = self.state.run(());
        }
        if self
            .clocks_triggered
            .iter()
            .any(|i| self.state.small_slots[*i] != 0)
        {
            self.state_ready_to_run = true;
            true
        } else {
            false
        }
    }
    #[track_caller]
    fn run_extern_modules_cycle(
        this_ref: &Rc<RefCell<Self>>,
        generator_waker: &std::task::Waker,
        extern_modules_ready_to_run: &[usize],
    ) {
        let mut this = this_ref.borrow_mut();
        for module_index in extern_modules_ready_to_run.iter().copied() {
            let extern_module = &mut this.extern_modules[module_index];
            extern_module.wait_targets.clear();
            let mut generator = if !extern_module.module_state.did_initial_settle {
                let sim = extern_module.sim;
                drop(this);
                Box::into_pin(sim.run(ExternModuleSimulationState {
                    sim_impl: this_ref.clone(),
                    module_index,
                    wait_for_changes_wait_targets: EarliestWaitTargets::default(),
                }))
            } else if let Some(generator) = extern_module.running_generator.take() {
                drop(this);
                generator
            } else {
                continue;
            };
            let generator = match generator
                .as_mut()
                .poll(&mut std::task::Context::from_waker(generator_waker))
            {
                Poll::Ready(()) => None,
                Poll::Pending => Some(generator),
            };
            this = this_ref.borrow_mut();
            this.extern_modules[module_index]
                .module_state
                .did_initial_settle = true;
            if !this.extern_modules[module_index]
                .module_state
                .uninitialized_ios
                .is_empty()
            {
                panic!(
                    "extern module didn't initialize all outputs before \
                    waiting, settling, or reading any inputs: {}",
                    this.extern_modules[module_index].sim.source_location
                );
            }
            this.extern_modules[module_index].running_generator = generator;
        }
    }
    /// clears `targets`
    #[track_caller]
    fn run_until(this_ref: &Rc<RefCell<Self>>, run_target: SimInstant) {
        let mut this = this_ref.borrow_mut();
        let mut ready_to_run_set = ReadyToRunSet::default();
        let generator_waker = this.generator_waker.clone();
        assert!(
            this.main_module.uninitialized_ios.is_empty(),
            "didn't initialize all inputs",
        );
        let run_target = run_target.max(this.instant);
        let mut settle_cycle = 0;
        let mut run_extern_modules = true;
        loop {
            assert!(settle_cycle < 100000, "settle(): took too many steps");
            settle_cycle += 1;
            let next_wait_target = match this.get_ready_to_run_set(&mut ready_to_run_set) {
                Some(next_wait_target) if next_wait_target <= run_target => next_wait_target,
                _ => break,
            };
            if next_wait_target > this.instant {
                settle_cycle = 0;
                this.set_instant_no_sim(next_wait_target);
            }
            if run_extern_modules {
                drop(this);
                Self::run_extern_modules_cycle(
                    this_ref,
                    &generator_waker,
                    &ready_to_run_set.extern_modules_ready_to_run,
                );
                this = this_ref.borrow_mut();
            }
            if ready_to_run_set.state_ready_to_run {
                if this.run_state_settle_cycle() {
                    // wait for clocks to settle before running extern modules again
                    run_extern_modules = false;
                } else {
                    run_extern_modules = true;
                }
            }
            if this.main_module.did_initial_settle {
                this.read_traces::<false>();
            } else {
                this.read_traces::<true>();
            }
            this.state.memory_write_log.sort_unstable();
            this.state.memory_write_log.dedup();
            this.main_module.did_initial_settle = true;
            this.for_each_trace_writer_storing_error(|this, trace_writer_state| {
                Ok(match trace_writer_state {
                    TraceWriterState::Decls(trace_writer_decls) => TraceWriterState::Running(
                        this.init_trace_writer(trace_writer_decls.write_decls(
                            this.trace_decls,
                            this.traces.0.len(),
                            this.trace_memories.len(),
                        )?)?,
                    ),
                    TraceWriterState::Init(trace_writer) => {
                        TraceWriterState::Running(this.init_trace_writer(trace_writer)?)
                    }
                    TraceWriterState::Running(trace_writer) => {
                        TraceWriterState::Running(this.update_trace_writer(trace_writer)?)
                    }
                    TraceWriterState::Errored(e) => TraceWriterState::Errored(e),
                })
            });
            this.state.memory_write_log.clear();
        }
        if run_target > this.instant {
            this.set_instant_no_sim(run_target);
        }
    }
    #[track_caller]
    fn settle(this_ref: &Rc<RefCell<Self>>) {
        let run_target = this_ref.borrow().instant;
        Self::run_until(this_ref, run_target);
    }
    fn get_module(&self, which_module: WhichModule) -> &SimulationModuleState {
        match which_module {
            WhichModule::Main => &self.main_module,
            WhichModule::Extern { module_index } => &self.extern_modules[module_index].module_state,
        }
    }
    fn get_module_mut(&mut self, which_module: WhichModule) -> &mut SimulationModuleState {
        match which_module {
            WhichModule::Main => &mut self.main_module,
            WhichModule::Extern { module_index } => {
                &mut self.extern_modules[module_index].module_state
            }
        }
    }
    #[track_caller]
    fn read_bit(
        &mut self,
        io: Expr<CanonicalType>,
        which_module: WhichModule,
    ) -> MaybeNeedsSettle<ReadBitFn, bool> {
        self.get_module(which_module)
            .read_helper(Expr::canonical(io), which_module)
            .map(|compiled_value| ReadBitFn { compiled_value })
            .apply_no_settle(&mut self.state)
    }
    #[track_caller]
    fn write_bit(&mut self, io: Expr<CanonicalType>, value: bool, which_module: WhichModule) {
        let compiled_value = self
            .get_module_mut(which_module)
            .write_helper(io, which_module);
        self.state_ready_to_run = true;
        match compiled_value.range.len() {
            TypeLen::A_SMALL_SLOT => {
                self.state.small_slots[compiled_value.range.small_slots.start] = value as _;
            }
            TypeLen::A_BIG_SLOT => {
                self.state.big_slots[compiled_value.range.big_slots.start] = value.into()
            }
            _ => unreachable!(),
        }
    }
    #[track_caller]
    fn read_bool_or_int<I: BoolOrIntType>(
        &mut self,
        io: Expr<I>,
        which_module: WhichModule,
    ) -> MaybeNeedsSettle<ReadBoolOrIntFn<I>, I::Value> {
        self.get_module(which_module)
            .read_helper(Expr::canonical(io), which_module)
            .map(|compiled_value| ReadBoolOrIntFn { compiled_value, io })
            .apply_no_settle(&mut self.state)
    }
    #[track_caller]
    fn write_bool_or_int<I: BoolOrIntType>(
        &mut self,
        io: Expr<I>,
        value: I::Value,
        which_module: WhichModule,
    ) {
        let compiled_value = self
            .get_module_mut(which_module)
            .write_helper(Expr::canonical(io), which_module);
        self.state_ready_to_run = true;
        let value: BigInt = value.into();
        match compiled_value.range.len() {
            TypeLen::A_SMALL_SLOT => {
                let mut small_value = value.iter_u64_digits().next().unwrap_or(0);
                if value.is_negative() {
                    small_value = small_value.wrapping_neg();
                }
                self.state.small_slots[compiled_value.range.small_slots.start] = small_value;
            }
            TypeLen::A_BIG_SLOT => {
                self.state.big_slots[compiled_value.range.big_slots.start] = value
            }
            _ => unreachable!(),
        }
    }
    #[track_caller]
    fn read_write_sim_value_helper<Bits: ?Sized>(
        state: &mut interpreter::State,
        compiled_value: CompiledValue<CanonicalType>,
        start_bit_index: usize,
        bits: &mut Bits,
        read_write_big_scalar: impl Fn(bool, std::ops::Range<usize>, &mut Bits, &mut BigInt) + Copy,
        read_write_small_scalar: impl Fn(bool, std::ops::Range<usize>, &mut Bits, &mut SmallUInt) + Copy,
    ) {
        match compiled_value.layout.body {
            CompiledTypeLayoutBody::Scalar => {
                let signed = match compiled_value.layout.ty {
                    CanonicalType::UInt(_) => false,
                    CanonicalType::SInt(_) => true,
                    CanonicalType::Bool(_) => false,
                    CanonicalType::Array(_) => unreachable!(),
                    CanonicalType::Enum(_) => false,
                    CanonicalType::Bundle(_) => unreachable!(),
                    CanonicalType::AsyncReset(_) => false,
                    CanonicalType::SyncReset(_) => false,
                    CanonicalType::Reset(_) => false,
                    CanonicalType::Clock(_) => false,
                    CanonicalType::PhantomConst(_) => unreachable!(),
                };
                let bit_indexes =
                    start_bit_index..start_bit_index + compiled_value.layout.ty.bit_width();
                match compiled_value.range.len() {
                    TypeLen::A_SMALL_SLOT => read_write_small_scalar(
                        signed,
                        bit_indexes,
                        bits,
                        &mut state.small_slots[compiled_value.range.small_slots.start],
                    ),
                    TypeLen::A_BIG_SLOT => read_write_big_scalar(
                        signed,
                        bit_indexes,
                        bits,
                        &mut state.big_slots[compiled_value.range.big_slots.start],
                    ),
                    _ => unreachable!(),
                }
            }
            CompiledTypeLayoutBody::Array { element } => {
                let ty = <Array>::from_canonical(compiled_value.layout.ty);
                let element_bit_width = ty.element().bit_width();
                for element_index in 0..ty.len() {
                    Self::read_write_sim_value_helper(
                        state,
                        CompiledValue {
                            layout: *element,
                            range: compiled_value
                                .range
                                .index_array(element.layout.len(), element_index),
                            write: None,
                        },
                        start_bit_index + element_index * element_bit_width,
                        bits,
                        read_write_big_scalar,
                        read_write_small_scalar,
                    );
                }
            }
            CompiledTypeLayoutBody::Bundle { fields } => {
                let ty = Bundle::from_canonical(compiled_value.layout.ty);
                for (
                    (_field, offset),
                    CompiledBundleField {
                        offset: layout_offset,
                        ty: field_layout,
                    },
                ) in ty.fields().iter().zip(ty.field_offsets()).zip(fields)
                {
                    Self::read_write_sim_value_helper(
                        state,
                        CompiledValue {
                            layout: field_layout,
                            range: compiled_value.range.slice(TypeIndexRange::new(
                                layout_offset,
                                field_layout.layout.len(),
                            )),
                            write: None,
                        },
                        start_bit_index + offset,
                        bits,
                        read_write_big_scalar,
                        read_write_small_scalar,
                    );
                }
            }
        }
    }
    #[track_caller]
    fn read_no_settle_helper(
        state: &mut interpreter::State,
        io: Expr<CanonicalType>,
        compiled_value: CompiledValue<CanonicalType>,
        mut bits: UIntValue,
    ) -> SimValue<CanonicalType> {
        SimulationImpl::read_write_sim_value_helper(
            state,
            compiled_value,
            0,
            bits.bits_mut(),
            |_signed, bit_range, bits, value| {
                <UInt>::copy_bits_from_bigint_wrapping(value, &mut bits[bit_range]);
            },
            |_signed, bit_range, bits, value| {
                let bytes = value.to_le_bytes();
                let bitslice = BitSlice::<u8, Lsb0>::from_slice(&bytes);
                let bitslice = &bitslice[..bit_range.len()];
                bits[bit_range].clone_from_bitslice(bitslice);
            },
        );
        SimValue::from_bits(Expr::ty(io), bits)
    }
    /// doesn't modify `bits`
    fn value_changed(
        state: &mut interpreter::State,
        compiled_value: CompiledValue<CanonicalType>,
        mut bits: &BitSlice,
    ) -> bool {
        assert_eq!(bits.len(), compiled_value.layout.ty.bit_width());
        let any_change = std::cell::Cell::new(false);
        SimulationImpl::read_write_sim_value_helper(
            state,
            compiled_value,
            0,
            &mut bits,
            |_signed, bit_range, bits, value| {
                if !<UInt>::bits_equal_bigint_wrapping(value, &bits[bit_range]) {
                    any_change.set(true);
                }
            },
            |_signed, bit_range, bits, value| {
                let bytes = value.to_le_bytes();
                let bitslice = BitSlice::<u8, Lsb0>::from_slice(&bytes);
                let bitslice = &bitslice[..bit_range.len()];
                if bits[bit_range] != *bitslice {
                    any_change.set(true);
                }
            },
        );
        any_change.get()
    }
    #[track_caller]
    fn read(
        &mut self,
        io: Expr<CanonicalType>,
        which_module: WhichModule,
    ) -> (
        CompiledValue<CanonicalType>,
        MaybeNeedsSettle<ReadFn, SimValue<CanonicalType>>,
    ) {
        let compiled_value = self.get_module(which_module).read_helper(io, which_module);
        let value = compiled_value
            .map(|compiled_value| ReadFn { compiled_value, io })
            .apply_no_settle(&mut self.state);
        let (MaybeNeedsSettle::NeedsSettle(compiled_value)
        | MaybeNeedsSettle::NoSettleNeeded(compiled_value)) = compiled_value;
        (compiled_value, value)
    }
    #[track_caller]
    fn write(
        &mut self,
        io: Expr<CanonicalType>,
        value: &SimValue<CanonicalType>,
        which_module: WhichModule,
    ) {
        let compiled_value = self
            .get_module_mut(which_module)
            .write_helper(io, which_module);
        self.state_ready_to_run = true;
        assert_eq!(Expr::ty(io), SimValue::ty(value));
        Self::read_write_sim_value_helper(
            &mut self.state,
            compiled_value,
            0,
            &mut value.bits().bits(),
            |signed, bit_range, bits, value| {
                if signed {
                    *value = SInt::bits_to_bigint(&bits[bit_range]);
                } else {
                    *value = UInt::bits_to_bigint(&bits[bit_range]);
                }
            },
            |signed, bit_range, bits, value| {
                let mut small_value = [0; mem::size_of::<SmallUInt>()];
                if signed && bits[bit_range.clone()].last().as_deref().copied() == Some(true) {
                    small_value.fill(u8::MAX);
                }
                small_value.view_bits_mut::<Lsb0>()[0..bit_range.len()]
                    .clone_from_bitslice(&bits[bit_range]);
                *value = SmallUInt::from_le_bytes(small_value);
            },
        );
    }
    #[track_caller]
    fn settle_if_needed<F, O>(this_ref: &Rc<RefCell<Self>>, v: MaybeNeedsSettle<F, O>) -> O
    where
        for<'a> F: MaybeNeedsSettleFn<&'a mut interpreter::State, Output = O>,
    {
        match v {
            MaybeNeedsSettle::NeedsSettle(v) => {
                Self::settle(this_ref);
                v.call(&mut this_ref.borrow_mut().state)
            }
            MaybeNeedsSettle::NoSettleNeeded(v) => v,
        }
    }
    async fn yield_settle_if_needed<F, O>(
        this_ref: &Rc<RefCell<Self>>,
        module_index: usize,
        v: MaybeNeedsSettle<F, O>,
    ) -> O
    where
        for<'a> F: MaybeNeedsSettleFn<&'a mut interpreter::State, Output = O>,
    {
        match v {
            MaybeNeedsSettle::NeedsSettle(v) => {
                Self::yield_advance_time_or_settle(this_ref.clone(), module_index, None).await;
                v.call(&mut this_ref.borrow_mut().state)
            }
            MaybeNeedsSettle::NoSettleNeeded(v) => v,
        }
    }
    fn close_all_trace_writers(&mut self) -> std::io::Result<()> {
        let trace_writers = mem::take(&mut self.trace_writers);
        let mut retval = Ok(());
        let close_trace_writer =
            |trace_writer: TraceWriterState<DynTraceWriterDecls>| match trace_writer {
                TraceWriterState::Decls(v) => v
                    .write_decls(
                        self.trace_decls,
                        self.traces.0.len(),
                        self.trace_memories.len(),
                    )?
                    .close(),
                TraceWriterState::Init(v) => v.close(),
                TraceWriterState::Running(v) => v.close(),
                TraceWriterState::Errored(Some(e)) => Err(e),
                TraceWriterState::Errored(None) => Ok(()),
            };
        for trace_writer in trace_writers {
            retval = retval.and(close_trace_writer(trace_writer));
        }
        retval
    }
    fn for_each_trace_writer_storing_error(
        &mut self,
        mut f: impl FnMut(
            &mut Self,
            TraceWriterState<DynTraceWriterDecls>,
        ) -> std::io::Result<TraceWriterState<DynTraceWriterDecls>>,
    ) {
        let mut trace_writers = mem::take(&mut self.trace_writers);
        for trace_writer in &mut trace_writers {
            *trace_writer = match f(
                self,
                mem::replace(trace_writer, TraceWriterState::Errored(None)),
            ) {
                Ok(v) => v,
                Err(e) => TraceWriterState::Errored(Some(e)),
            };
        }
        self.trace_writers = trace_writers;
    }
    fn for_each_trace_writer_getting_error(
        &mut self,
        mut f: impl FnMut(
            &mut Self,
            TraceWriterState<DynTraceWriterDecls>,
        ) -> std::io::Result<TraceWriterState<DynTraceWriterDecls>>,
    ) -> std::io::Result<()> {
        let mut retval = Ok(());
        let mut trace_writers = mem::take(&mut self.trace_writers);
        for trace_writer in &mut trace_writers {
            *trace_writer = match f(
                self,
                mem::replace(trace_writer, TraceWriterState::Errored(None)),
            ) {
                Ok(v) => v,
                Err(e) => {
                    if retval.is_ok() {
                        retval = Err(e);
                        TraceWriterState::Errored(None)
                    } else {
                        TraceWriterState::Errored(Some(e))
                    }
                }
            };
        }
        self.trace_writers = trace_writers;
        retval
    }
    fn close(this: Rc<RefCell<Self>>) -> std::io::Result<()> {
        if this.borrow().main_module.did_initial_settle {
            Self::settle(&this);
        }
        this.borrow_mut().close_all_trace_writers()
    }
    fn flush_traces(this_ref: &Rc<RefCell<Self>>) -> std::io::Result<()> {
        if this_ref.borrow().main_module.did_initial_settle {
            Self::settle(this_ref);
        }
        this_ref.borrow_mut().for_each_trace_writer_getting_error(
            |this, trace_writer: TraceWriterState<DynTraceWriterDecls>| match trace_writer {
                TraceWriterState::Decls(v) => {
                    let mut v = v.write_decls(
                        this.trace_decls,
                        this.traces.0.len(),
                        this.trace_memories.len(),
                    )?;
                    v.flush()?;
                    Ok(TraceWriterState::Init(v))
                }
                TraceWriterState::Init(mut v) => {
                    v.flush()?;
                    Ok(TraceWriterState::Init(v))
                }
                TraceWriterState::Running(mut v) => {
                    v.flush()?;
                    Ok(TraceWriterState::Running(v))
                }
                TraceWriterState::Errored(Some(e)) => Err(e),
                TraceWriterState::Errored(None) => Ok(TraceWriterState::Errored(None)),
            },
        )
    }
}

impl Drop for SimulationImpl {
    fn drop(&mut self) {
        self.close_all_trace_writers()
            .expect("error closing trace writers");
    }
}

pub struct Simulation<T: BundleType> {
    sim_impl: Rc<RefCell<SimulationImpl>>,
    io: Expr<T>,
}

struct SortedSetDebug<'a, T, V>(&'a HashMap<T, V>);

impl<T: fmt::Debug, V> fmt::Debug for SortedSetDebug<'_, T, V> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let mut entries = Vec::from_iter(self.0.iter().map(|(v, _)| {
            if f.alternate() {
                format!("{v:#?}")
            } else {
                format!("{v:?}")
            }
        }));
        entries.sort();
        f.debug_set()
            .entries(entries.iter().map(DebugAsDisplay))
            .finish()
    }
}

struct SortedMapDebug<'a, K, V>(&'a HashMap<K, V>);

impl<K: fmt::Debug, V: fmt::Debug> fmt::Debug for SortedMapDebug<'_, K, V> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let mut entries = Vec::from_iter(self.0.iter().map(|(k, v)| {
            if f.alternate() {
                (format!("{k:#?}"), format!("{v:#?}"))
            } else {
                (format!("{k:?}"), format!("{v:?}"))
            }
        }));
        entries.sort();
        f.debug_map()
            .entries(
                entries
                    .iter()
                    .map(|(k, v)| (DebugAsDisplay(k), DebugAsDisplay(v))),
            )
            .finish()
    }
}

impl<T: BundleType> fmt::Debug for Simulation<T> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self { sim_impl, io } = self;
        sim_impl.borrow().debug_fmt(Some(io), f)
    }
}

macro_rules! impl_simulation_methods {
    (
        async_await = ($($async:tt, $await:tt)?),
        track_caller = ($(#[$track_caller:tt])?),
        which_module = |$self:ident| $which_module:expr,
    ) => {
        $(#[$track_caller])?
        pub $($async)? fn read_bool_or_int<I: BoolOrIntType>(&mut $self, io: Expr<I>) -> I::Value {
            let retval = $self
                .sim_impl
                .borrow_mut()
                .read_bool_or_int(io, $which_module);
            $self.settle_if_needed(retval)$(.$await)?
        }
        $(#[$track_caller])?
        pub $($async)? fn write_bool_or_int<I: BoolOrIntType>(
            &mut $self,
            io: Expr<I>,
            value: impl ToExpr<Type = I>,
        ) {
            let value = value.to_expr();
            assert_eq!(Expr::ty(io), Expr::ty(value), "type mismatch");
            let value = value
                .to_literal_bits()
                .expect("the value that is being written to an input must be a literal");
            $self.sim_impl.borrow_mut().write_bool_or_int(
                io,
                I::bits_to_value(Cow::Borrowed(&value)),
                $which_module,
            );
        }
        $(#[$track_caller])?
        pub $($async)? fn write_clock(&mut $self, io: Expr<Clock>, value: bool) {
            $self.sim_impl
                .borrow_mut()
                .write_bit(Expr::canonical(io), value, $which_module);
        }
        $(#[$track_caller])?
        pub $($async)? fn read_clock(&mut $self, io: Expr<Clock>) -> bool {
            let retval = $self
                .sim_impl
                .borrow_mut()
                .read_bit(Expr::canonical(io), $which_module);
            $self.settle_if_needed(retval)$(.$await)?
        }
        $(#[$track_caller])?
        pub $($async)? fn write_bool(&mut $self, io: Expr<Bool>, value: bool) {
            $self.sim_impl
                .borrow_mut()
                .write_bit(Expr::canonical(io), value, $which_module);
        }
        $(#[$track_caller])?
        pub $($async)? fn read_bool(&mut $self, io: Expr<Bool>) -> bool {
            let retval = $self
                .sim_impl
                .borrow_mut()
                .read_bit(Expr::canonical(io), $which_module);
            $self.settle_if_needed(retval)$(.$await)?
        }
        $(#[$track_caller])?
        pub $($async)? fn write_reset<R: ResetType>(&mut $self, io: Expr<R>, value: bool) {
            $self.sim_impl
                .borrow_mut()
                .write_bit(Expr::canonical(io), value, $which_module);
        }
        $(#[$track_caller])?
        pub $($async)? fn read_reset<R: ResetType>(&mut $self, io: Expr<R>) -> bool {
            let retval = $self
                .sim_impl
                .borrow_mut()
                .read_bit(Expr::canonical(io), $which_module);
            $self.settle_if_needed(retval)$(.$await)?
        }
        $(#[$track_caller])?
        pub $($async)? fn read<IO: Type>(&mut $self, io: Expr<IO>) -> SimValue<IO> {
            let retval = $self
                .sim_impl
                .borrow_mut()
                .read(Expr::canonical(io), $which_module).1;
            SimValue::from_canonical($self.settle_if_needed(retval)$(.$await)?)
        }
        $(#[$track_caller])?
        pub $($async)? fn write<IO: Type, V: value::ToSimValueWithType<IO>>(&mut $self, io: Expr<IO>, value: V) {
            $self.sim_impl.borrow_mut().write(
                Expr::canonical(io),
                &SimValue::into_canonical(value.into_sim_value_with_type(Expr::ty(io))),
                $which_module,
            );
        }
    };
}

impl<T: BundleType> Simulation<T> {
    pub fn new(module: Interned<Module<T>>) -> Self {
        Self::from_compiled(Compiled::new(module))
    }
    pub fn add_trace_writer<W: TraceWriterDecls>(&mut self, writer: W) {
        self.sim_impl
            .borrow_mut()
            .trace_writers
            .push(TraceWriterState::Decls(DynTraceWriterDecls::new(writer)));
    }
    pub fn flush_traces(&mut self) -> std::io::Result<()> {
        SimulationImpl::flush_traces(&self.sim_impl)
    }
    pub fn close(self) -> std::io::Result<()> {
        SimulationImpl::close(self.sim_impl)
    }
    pub fn canonical(self) -> Simulation<Bundle> {
        let Self { sim_impl, io } = self;
        Simulation {
            sim_impl,
            io: Expr::as_bundle(io),
        }
    }
    pub fn from_canonical(canonical: Simulation<Bundle>) -> Self {
        let Simulation { sim_impl, io } = canonical;
        Self {
            sim_impl,
            io: Expr::from_bundle(io),
        }
    }
    pub fn io(&self) -> Expr<T> {
        self.io.to_expr()
    }
    pub fn from_compiled(compiled: Compiled<T>) -> Self {
        let sim_impl = SimulationImpl::new(compiled.canonical());
        Self {
            io: Expr::from_bundle(sim_impl.io),
            sim_impl: Rc::new(RefCell::new(sim_impl)),
        }
    }
    #[track_caller]
    pub fn settle(&mut self) {
        SimulationImpl::settle(&self.sim_impl);
    }
    #[track_caller]
    pub fn advance_time(&mut self, duration: SimDuration) {
        SimulationImpl::advance_time(&self.sim_impl, duration);
    }
    #[track_caller]
    fn settle_if_needed<F, O>(&mut self, v: MaybeNeedsSettle<F, O>) -> O
    where
        for<'a> F: MaybeNeedsSettleFn<&'a mut interpreter::State, Output = O>,
    {
        SimulationImpl::settle_if_needed(&self.sim_impl, v)
    }
    impl_simulation_methods!(
        async_await = (),
        track_caller = (#[track_caller]),
        which_module = |self| WhichModule::Main,
    );
    #[doc(hidden)]
    /// This is explicitly unstable and may be changed/removed at any time
    pub fn set_breakpoints_unstable(&mut self, pcs: HashSet<usize>, trace: bool) {
        self.sim_impl.borrow_mut().breakpoints = Some(BreakpointsSet {
            last_was_break: false,
            set: pcs,
            trace,
        });
    }
}

pub struct ExternModuleSimulationState {
    sim_impl: Rc<RefCell<SimulationImpl>>,
    module_index: usize,
    wait_for_changes_wait_targets: EarliestWaitTargets,
}

impl fmt::Debug for ExternModuleSimulationState {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self {
            sim_impl: _,
            module_index,
            wait_for_changes_wait_targets: _,
        } = self;
        f.debug_struct("ExternModuleSimulationState")
            .field("sim_impl", &DebugAsDisplay("..."))
            .field("module_index", module_index)
            .finish_non_exhaustive()
    }
}

impl ExternModuleSimulationState {
    pub async fn settle(&mut self) {
        SimulationImpl::yield_advance_time_or_settle(self.sim_impl.clone(), self.module_index, None)
            .await
    }
    pub async fn advance_time(&mut self, duration: SimDuration) {
        SimulationImpl::yield_advance_time_or_settle(
            self.sim_impl.clone(),
            self.module_index,
            Some(duration),
        )
        .await
    }
    pub async fn wait_for_changes<I: IntoIterator<Item: ToExpr>>(
        &mut self,
        iter: I,
        timeout: Option<SimDuration>,
    ) {
        self.wait_for_changes_wait_targets.clear();
        let which_module = WhichModule::Extern {
            module_index: self.module_index,
        };
        for io in iter {
            let io = Expr::canonical(io.to_expr());
            let (key, value) = self.sim_impl.borrow_mut().read(io, which_module);
            let value = self.settle_if_needed(value).await;
            self.wait_for_changes_wait_targets
                .insert(WaitTarget::Change { key, value });
        }
        if let Some(timeout) = timeout {
            self.wait_for_changes_wait_targets.instant =
                Some(self.sim_impl.borrow().instant + timeout);
        }
        SimulationImpl::yield_wait(
            self.sim_impl.clone(),
            self.module_index,
            &mut self.wait_for_changes_wait_targets,
        )
        .await;
    }
    pub async fn wait_for_clock_edge(&mut self, clk: impl ToExpr<Type = Clock>) {
        let clk = clk.to_expr();
        while self.read_clock(clk).await {
            self.wait_for_changes([clk], None).await;
        }
        while !self.read_clock(clk).await {
            self.wait_for_changes([clk], None).await;
        }
    }
    async fn settle_if_needed<F, O>(&mut self, v: MaybeNeedsSettle<F, O>) -> O
    where
        for<'a> F: MaybeNeedsSettleFn<&'a mut interpreter::State, Output = O>,
    {
        SimulationImpl::yield_settle_if_needed(&self.sim_impl, self.module_index, v).await
    }
    impl_simulation_methods!(
        async_await = (async, await),
        track_caller = (),
        which_module = |self| WhichModule::Extern { module_index: self.module_index },
    );
}

pub trait ExternModuleSimGenerator: Clone + Eq + Hash + Any + Send + Sync + fmt::Debug {
    fn run<'a>(&'a self, sim: ExternModuleSimulationState) -> impl IntoFuture<Output = ()> + 'a;
}

pub struct SimGeneratorFn<Args, Fut> {
    pub args: Args,
    pub f: fn(Args, ExternModuleSimulationState) -> Fut,
}

impl<Args: fmt::Debug, Fut> fmt::Debug for SimGeneratorFn<Args, Fut> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self { args, f: _ } = self;
        f.debug_struct("SimGeneratorFn")
            .field("args", args)
            .field("f", &DebugAsDisplay("..."))
            .finish()
    }
}

impl<Args: Hash, Fut> Hash for SimGeneratorFn<Args, Fut> {
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        let Self { args, f } = self;
        args.hash(state);
        f.hash(state);
    }
}

impl<Args: Eq, Fut> Eq for SimGeneratorFn<Args, Fut> {}

impl<Args: PartialEq, Fut> PartialEq for SimGeneratorFn<Args, Fut> {
    fn eq(&self, other: &Self) -> bool {
        let Self { args, f } = self;
        *args == other.args && *f == other.f
    }
}

impl<Args: Clone, Fut> Clone for SimGeneratorFn<Args, Fut> {
    fn clone(&self) -> Self {
        Self {
            args: self.args.clone(),
            f: self.f,
        }
    }
}

impl<Args: Copy, Fut> Copy for SimGeneratorFn<Args, Fut> {}

impl<
        T: fmt::Debug + Clone + Eq + Hash + Send + Sync + 'static,
        Fut: IntoFuture<Output = ()> + 'static,
    > ExternModuleSimGenerator for SimGeneratorFn<T, Fut>
{
    fn run<'a>(&'a self, sim: ExternModuleSimulationState) -> impl IntoFuture<Output = ()> + 'a {
        (self.f)(self.args.clone(), sim)
    }
}

trait DynExternModuleSimGenerator: Any + Send + Sync + SupportsPtrEqWithTypeId + fmt::Debug {
    fn dyn_run<'a>(&'a self, sim: ExternModuleSimulationState)
        -> Box<dyn Future<Output = ()> + 'a>;
}

impl<T: ExternModuleSimGenerator> DynExternModuleSimGenerator for T {
    fn dyn_run<'a>(
        &'a self,
        sim: ExternModuleSimulationState,
    ) -> Box<dyn Future<Output = ()> + 'a> {
        Box::new(self.run(sim).into_future())
    }
}

impl InternedCompare for dyn DynExternModuleSimGenerator {
    type InternedCompareKey = PtrEqWithTypeId;

    fn interned_compare_key_ref(this: &Self) -> Self::InternedCompareKey {
        this.get_ptr_eq_with_type_id()
    }
}

#[derive(Clone, Copy, PartialEq, Eq, Hash, Debug)]
pub struct ExternModuleSimulation {
    generator: Interned<dyn DynExternModuleSimGenerator>,
    source_location: SourceLocation,
}

impl ExternModuleSimulation {
    pub fn new_with_loc<G: ExternModuleSimGenerator>(
        source_location: SourceLocation,
        generator: G,
    ) -> Self {
        Self {
            generator: Interned::cast_unchecked(
                generator.intern(),
                |v| -> &dyn DynExternModuleSimGenerator { v },
            ),
            source_location,
        }
    }
    #[track_caller]
    pub fn new<G: ExternModuleSimGenerator>(generator: G) -> Self {
        Self::new_with_loc(SourceLocation::caller(), generator)
    }
    fn run(&self, sim: ExternModuleSimulationState) -> Box<dyn Future<Output = ()> + 'static> {
        Interned::into_inner(self.generator).dyn_run(sim)
    }
}
